//---------------------------------------------------------------------
#include <vcl.h>
#include <vcl\classes.hpp>
#pragma hdrstop


#include "TChildForm.h"



//---------------------------------------------------------------------
#pragma resource "*.dfm"
TChildForm*	ChildForm;	// no need for it as
                        // we use new to create it in MDIChildCreate of MainForm
//---------------------------------------------------------------------


#include "TMigrationFC.h"
#include "TReplicationFC.h"

#include "TFrequenciesDialog.h"
#include "TThresholdsDialog.h"
#include "TMigrationForm.h"
#include "TReplicationForm.h"
#include "TMainForm.h"



//---------------------------------------------------------------------
__fastcall TChildForm::TChildForm(TComponent *Owner)
	: TForm(Owner)
{
	backup_network	= NULL;
    to_be_undone	= false;
    to_be_redone	= false;

	inside_node_index		= -1;
	inside_link_index		= -1;
	inside_fragment_index	= -1;
	inside_cluster_index	= -1;
    dragging			= false;
    mouse_down			= false;
    left_mouse_button   = false;
    right_mouse_button  = false;
    mouse_dblclk		= false;

    selection_color		= clRed;

    running = false;
    auto_demands    = false;
    single_fragment = true;
    same_demands_on_clusters = true;

    so_counter = 0;
    p_counter = 0;
    m_counter = 0;
    c_counter = 0;
    r_counter = 0;
    d_counter = 0;
    running_counter = 0;

    for (int i=0; i<MAX_SAMPLES*MAX_RUNS; i++) so_activity[i] = 0;
    SO_FREQUENCY = 10;

    P_FREQUENCY = SO_FREQUENCY / 2;
    M_FREQUENCY = SO_FREQUENCY / 1;
    C_FREQUENCY = SO_FREQUENCY / 2;
    R_FREQUENCY = SO_FREQUENCY / 3;
    D_FREQUENCY = SO_FREQUENCY / 1;
    p_enabled = false;
    m_enabled = false;
    c_enabled = false;
    r_enabled = false;
    d_enabled = false;
}




//---------------------------------------------------------------------------
void __fastcall TChildForm::FormCreate(TObject *Sender)
{
	Network			= NULL;
	backup_network	= NULL;
}


//---------------------------------------------------------------------
void __fastcall TChildForm::FormClose(TObject *Sender, TCloseAction &Action)
{
	Action = caFree;
}


//---------------------------------------------------------------------------
void __fastcall TChildForm::FormDestroy(TObject *Sender)
{
    if (Network) delete Network;
    if (backup_network) delete backup_network;
}


//---------------------------------------------------------------------------
void __fastcall TChildForm::FormPaint(TObject *Sender)
{
	NetworkPaintBox->Repaint();
	FragmentsPaintBox->Repaint();
	ClustersPaintBox->Repaint();
	ResponsePaintBox->Repaint();
	ResourcePaintBox->Repaint();
}






//---------------------------------------------------------------------------
void __fastcall TChildForm::NetworkPaintBoxMouseDown(TObject *Sender, TMouseButton Button, TShiftState Shift, int X, int Y)
{
	// patch to stop dragging after dblclk's mouse_down undesirable event
	if (mouse_dblclk)
    {
		mouse_dblclk = false; 	//  clear for next mouse_down event
    }
    else	// otherwise proceed as normal
    {
        mouse_down = true;

        if (Network != NULL)
        {
            if(!Network->IsEmpty())
            {
                // deselect everything and update
				Network->DeSelectChildren();
			    inside_node_index		= -1;
	    		inside_link_index		= -1;
			    inside_fragment_index	= -1;
			    inside_cluster_index	= -1;


                for(int i=Network->Nodes->Count-1; i>=0; i--)
                {
                    TNode* current_node = ((TNode*)Network->Nodes->Items[i]);
                    int nx = current_node->get_position().x;
                    int ny = current_node->get_position().y;

            		int RadiusRadius = current_node->get_radius() * current_node->get_radius() + 2; // 2 is tolerance
	    	        int DistanceDistance = (nx-X)*(nx-X) + (ny-Y)*(ny-Y);
            		if (DistanceDistance < RadiusRadius)
		            {
		                inside_node_index = i;
                	    return;
		            }
		        }
                for(int i=Network->Links->Count-1; i>=0; i--)
                {
                    TLink* current_link = ((TLink*)Network->Links->Items[i]);
               		if (current_link != NULL)
                    {
	               		if (current_link->Contains(Point(X,Y)))
    		            {
  		    	            inside_link_index = i;
                	   	    return;
                        }
		            }
		        }
            }
        }
    }
}





//---------------------------------------------------------------------------
void __fastcall TChildForm::NetworkPaintBoxMouseMove(TObject *Sender, TShiftState Shift, int X, int Y)
{
	static int oldX;
	static int oldY;
   	static int old_found_elemet_index;
   	       int     found_elemet_index;

    if (Network != NULL)
    {
        if (!Network->IsEmpty())
        {
		    if ( (inside_node_index != -1) || (inside_link_index != -1) )
			{
               	if ( mouse_down )	  // start dragging
       	        {
		            dragging = true;
					//Screen->Cursor = crDrag;
	   			    if (inside_node_index != -1)
                   	{
                        // limit to painting area
		                if ((X > 0) && (Y > 0) && (X < NetworkPaintBox->Width) && (Y < NetworkPaintBox->Height))
		       	        {
			                TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
			       	        current_node->set_position(Point(X,Y));
							// redraw as you drag
							NetworkPaintBox->Repaint();
                		}
                        else // outside painting area
                        {
							//Screen->Cursor = crNoDrop;
                        }
					}
	                else if (inside_link_index != -1)
    	            {
		                if ((X > 0) && (Y > 0) && (X < NetworkPaintBox->Width) && (Y < NetworkPaintBox->Height))
		       	        {
			                TLink* current_link = ((TLink*)Network->Links->Items[inside_link_index]);
                            TPoint p = Point(X,Y);  int dx = X - oldX;  int dy = Y - oldY;
       				        TPoint a = current_link->Node_a->get_position();
       				        TPoint b = current_link->Node_b->get_position();

	                        // limit to painting area
			                if (
                            	(a.x+dx > 0) && (a.y+dy > 0) && (a.x+dx < NetworkPaintBox->Width) && (a.y+dy < NetworkPaintBox->Height)
								&&
                            	(b.x+dx > 0) && (b.y+dy > 0) && (b.x+dx < NetworkPaintBox->Width) && (b.y+dy < NetworkPaintBox->Height)
                            	)
			       	        {
								//Screen->Cursor = crDrag;
       					        current_link->Node_a->set_position(Point(a.x+dx, a.y+dy));
       					        current_link->Node_b->set_position(Point(b.x+dx, b.y+dy));
								// redraw as you drag
								NetworkPaintBox->Repaint();
                            }
	                        else
    	                    {
								//Screen->Cursor = crNoDrop;
        	                }
                        }
	       			}
                }
				else if ( !mouse_down )   // moving while an item is selected,
                						  // no dragging
	            {
					// keep selected element's details displayed
	   		    }
            }
            else  // moving with no selection
            {
                bool element_found = false;
	   			found_elemet_index = -1;

				// display caption when moving over element
	            if (!element_found)
                {
                    // try nodes
	                for(int i=Network->Nodes->Count-1; i>=0; i--)
	                {
                      TNode* current_node = ((TNode*)Network->Nodes->Items[i]);
                      if (current_node != NULL)
                      {
        	            int nx = current_node->get_position().x;
            	        int ny = current_node->get_position().y;

	            		int RadiusRadius = current_node->get_radius() * current_node->get_radius() + 2; // 2 is tolerance
		    	        int DistanceDistance = (nx-X)*(nx-X) + (ny-Y)*(ny-Y);
        	    		if (DistanceDistance < RadiusRadius)
			            {
							element_found = true;
				   			found_elemet_index = i;

							// don't repaint if moving inside node
						    if (old_found_elemet_index != found_elemet_index)
							{
								NetworkLabel->Caption = "   Id    Capacity     Utilization    Fragments";
		                        current_node->DrawCaption(NetworkPaintBox->Canvas);
        	                    MainForm->StatusBar->Panels->Items[2]->Text = "Node = " + IntToStr((int)current_node->get_id());
						    }

    	                    break;
			            }
                      }
                    }
		        }
                // if still not found, try links
	            if (!element_found)
                {
	                for(int i=Network->Links->Count-1; i>=0; i--)
    	            {
        	          TLink* current_link = ((TLink*)Network->Links->Items[i]);
                      if (current_link != NULL)
                      {
            	   		if (current_link != NULL)
                	    {
	               			if (current_link->Contains(Point(X,Y)))
	    		            {
								element_found = true;
					   			found_elemet_index = i;

								// don't repaint if moving inside link
							    if (old_found_elemet_index != found_elemet_index)
								{
									NetworkLabel->Caption = "   Id    Capacity     Utilization     Type     Delay";
	        		                current_link->DrawCaption(NetworkPaintBox->Canvas);
									MainForm->StatusBar->Panels->Items[2]->Text = "Link = " + IntToStr((int)current_link->get_id()) + "   (" + IntToStr((int)current_link->Node_a->get_id()) + "_" + IntToStr((int)current_link->Node_b->get_id()) + ")";
							    }

                                break;
                	        }
		            	}
                      }
			        }
				}

                // moving outside elements, with no selection
	            if (!element_found)
                {
					NetworkPaintBox->Canvas->Pen->Color = clInfoBk;
					NetworkPaintBox->Canvas->Brush->Color = clInfoBk;
					NetworkPaintBox->Canvas->Rectangle(0, 0, 800, 15);

					NetworkLabel->Caption = "   Id    Capacity     Utilization";
					MainForm->StatusBar->Panels->Items[2]->Text = "";
                }
            }
        }
    }

	oldX = X;
	oldY = Y;
   	old_found_elemet_index = found_elemet_index;
}




//---------------------------------------------------------------------------
void __fastcall TChildForm::NetworkPaintBoxMouseUp(TObject *Sender, TMouseButton Button, TShiftState Shift, int X, int Y)
{
    mouse_down = false;
    if (Network != NULL)
    {
        if(!Network->IsEmpty())
        {
		    if ( (inside_node_index != -1) || (inside_link_index != -1) )
			{
               	if ( dragging )	  // end dragging
       	        {
       			    dragging = false;
					//Screen->Cursor = crDefault;
				    inside_node_index		= -1;
	    			inside_link_index		= -1;
				    inside_fragment_index	= -1;
	    			inside_cluster_index	= -1;
                }
				else //if ( !dragging )   // selection action
	            {
	       		    if (inside_node_index != -1)
       	            {
   			        	// draw selected node in selection color
						TNode* selected_node = (TNode*)Network->Nodes->Items[inside_node_index];
       			    	selected_node->Select(selection_color);
						NetworkPaintBox->Repaint();
                        // deselect other views
					    inside_fragment_index	= -1;
		    			inside_cluster_index 	= -1;
						FragmentsPaintBox->Repaint();
						ClustersPaintBox->Repaint();
					}
	                else if (inside_link_index != -1)
	   	            {
        		        // draw selected link in selection color
       				    TLink* selected_link = ((TLink*)Network->Links->Items[inside_link_index]);
                        selected_link->Select(selection_color);
						NetworkPaintBox->Repaint();
                        // deselect other views
					    inside_fragment_index	= -1;
		    			inside_cluster_index 	= -1;
						FragmentsPaintBox->Repaint();
						ClustersPaintBox->Repaint();
       				}
	   		    }
            }
            else  // click outside elements
            {
                if(Button == mbRight)
                {
                  Network->info_type++;   if (Network->info_type > 3) Network->info_type = 0; // cycle
                }
                //else if(Button == mbRight)
                //{
                //  Network->info_type--;   if (Network->info_type < 0) Network->info_type = 3; // cycle
                //}
                if (Network->info_type == 0)
                {
                    NetworkGroupBox->Height = 189;
                }
                else
                {
                    NetworkGroupBox->Height = 473;
                }

               	Network->DeSelectChildren();
			    inside_node_index		= -1;
    			inside_link_index		= -1;
			    inside_fragment_index	= -1;
    			inside_cluster_index	= -1;
				NetworkPaintBox->Repaint();
				FragmentsPaintBox->Repaint();
				ClustersPaintBox->Repaint();
            }

			MainForm->UpdateMenuItems(NULL);
        }
	}
}


//---------------------------------------------------------------------------
void __fastcall TChildForm::NetworkPaintBoxClick(TObject *Sender)
{
//
}

//---------------------------------------------------------------------------
void __fastcall TChildForm::NetworkPaintBoxDblClick(TObject *Sender)
{
	mouse_dblclk = true;

    if (Network != NULL)
    {
        if(!Network->IsEmpty())
        {
            if ((inside_node_index != -1) || (inside_link_index != -1))
            {
       		    if (inside_node_index != -1)
       		    {
					TNode* current_node = (TNode*)Network->Nodes->Items[inside_node_index];
           		    NodeDialog->Node = current_node;
                   	// assignment done NodeDialog::FormShow
					if (NodeDialog->ShowModal() == mrOk)
                    {
	                   	// assignment done in NodeDialog::OkBitBtnClick
                    }
                }
                else // if (inside_link_index != -1)
                {
	                TLink* current_link = (TLink*)Network->Links->Items[inside_link_index];
           		    LinkDialog->Link = current_link;
			    	LinkDialog->Node_a_ComboBox->Enabled = false;
   					LinkDialog->Node_b_ComboBox->Enabled = false;
                   	// assignment done LinkDialog::FormShow
					if (LinkDialog->ShowModal() == mrOk)
                    {
                       	// assignment done in LinkDialog::OkBitBtnClick
                    }
       		    }
            }
            else  // double_click outside elements
            {
            }

            MainForm->UpdateMenuItems(NULL);

			Network->DeSelectChildren();
			inside_node_index		= -1;
			inside_link_index		= -1;
		    inside_fragment_index	= -1;
   			inside_cluster_index 	= -1;
			NetworkPaintBox->Repaint();
			FragmentsPaintBox->Repaint();
			ClustersPaintBox->Repaint();
        }
    }
}





//------------------------------------------------------------------------------
void __fastcall TChildForm::NetworkPaintBoxPaint(TObject *Sender)
{
    if(Network != NULL)
    {
        Network->Draw(NetworkPaintBox->Canvas);

		MainForm->StatusBar->Panels->Items[0]->Text = "Nodes = " + IntToStr((int)Network->Nodes->Count);
		MainForm->StatusBar->Panels->Items[1]->Text = "Links = " + IntToStr((int)Network->Links->Count);

		if (inside_node_index != -1)
        {
          TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
          if (current_node != NULL)
          {
			NetworkLabel->Caption = "   Id    Capacity     Utilization    Fragments";
	        current_node->DrawCaption(NetworkPaintBox->Canvas);
			MainForm->StatusBar->Panels->Items[2]->Text = "Node = " + IntToStr((int)current_node->get_id());
          }
        }
		else if (inside_link_index != -1)
        {
          TLink* current_link = ((TLink*)Network->Links->Items[inside_link_index]);
          if (current_link != NULL)
          {
			NetworkLabel->Caption = "   Id    Capacity     Utilization     Type     Delay";
	        current_link->DrawCaption(NetworkPaintBox->Canvas);
			MainForm->StatusBar->Panels->Items[2]->Text = "Link = " + IntToStr((int)current_link->get_id()) + "   (" + IntToStr((int)current_link->Node_a->get_id()) + "_" + IntToStr((int)current_link->Node_b->get_id()) + ")";
          }
        }
        else
        {
			NetworkLabel->Caption = "   Id    Capacity     Utilization";
			MainForm->StatusBar->Panels->Items[2]->Text = "";
        }

        NetworkPaintBox->Canvas->Refresh();
    }
}










//---------------------------------------------------------------------------
void __fastcall TChildForm::FragmentsPaintBoxMouseDown(TObject *Sender,	TMouseButton Button, TShiftState Shift, int X, int Y)
{
	// patch to stop dragging after dblclk's mouse_down undesirable event
	if (mouse_dblclk)
    {
		mouse_dblclk = false; 	//  clear for next mouse_down event
    }
    else	// otherwise proceed as normal
    {
        mouse_down = true;
		if (Button == mbLeft)
	    {
	        left_mouse_button  = true;
    	    right_mouse_button = false;
	  	}
   	    else if (Button == mbRight)
       	{
	        left_mouse_button  = false;
    	    right_mouse_button = true;
        }
        else
        {
	        left_mouse_button  = false;
    	    right_mouse_button = false;
        }

    	if (Network != NULL)
    	{
        	if(!Network->IsEmpty())
	        {
                // deselcet everything and update
				Network->DeSelectChildren();
			    inside_node_index		= -1;
	    		inside_link_index		= -1;
			    inside_fragment_index	= -1;
			    inside_cluster_index	= -1;
				//NetworkPaintBox->Repaint();
				//FragmentsPaintBox->Repaint();
				//ClustersPaintBox->Repaint();


    	        for(int i=Network->Nodes->Count-1; i>=0; i--)
        	    {
	                TNode* current_node = ((TNode*)Network->Nodes->Items[i]);
	    	        for(int j=current_node->Fragments->Count-1; j>=0; j--)
	                {
    	                TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[j]);

        	            int fx = current_fragment->get_position().x;
   	        	        int fy = current_fragment->get_position().y;
	            		int RadiusRadius = current_fragment->get_radius() * current_fragment->get_radius() + 2; // 2 is tolerance
		    	        int DistanceDistance = (fx-X)*(fx-X) + (fy-Y)*(fy-Y);
        	    		if (DistanceDistance < RadiusRadius)
		    	        {
			                inside_node_index = i;
			                inside_fragment_index = j;
		               	    return;
			            }
		    	    }
	            }
    	    }
		}
    }
}





//---------------------------------------------------------------------------
void __fastcall TChildForm::FragmentsPaintBoxMouseMove(TObject *Sender,	TShiftState Shift, int X, int Y)
{
   	static int old_found_fragment_index;
           int     found_fragment_index;

    if (Network != NULL)
    {
		// moving with a selection
	    if ( (inside_node_index != -1) && (inside_fragment_index != -1) )
		{
           	if ( mouse_down )	  // start dragging
   	        {
	            dragging = true;
				if (left_mouse_button)				// migrate
			    {
					Screen->Cursor = crDrag;
		    	}
           	    else if (right_mouse_button)         // replicate
               	{
					Screen->Cursor = crMultiDrag;
                }
                else
                {
					Screen->Cursor = crDefault;
                }


                // move inside painting area
                if ((X > 0) && (Y > 0) && (X < FragmentsPaintBox->Width) && (Y < FragmentsPaintBox->Height))
       	        {

//
// find   target_node_index
//
target_node_index = -1;

int x[4];
x[0] = 0.000 * (FragmentsPaintBox->Width);
x[1] = 0.333 * (FragmentsPaintBox->Width);
x[2] = 0.666 * (FragmentsPaintBox->Width);
x[3] = 0.999 * (FragmentsPaintBox->Width);
int y[4];
y[0] = 0.000 * (FragmentsPaintBox->Height-1) + 15;
y[1] = 0.333 * (FragmentsPaintBox->Height-1) + 10;
y[2] = 0.666 * (FragmentsPaintBox->Height-1) + 05;
y[3] = 0.999 * (FragmentsPaintBox->Height-1) + 00;

for (int i=0; i<Network->Nodes->Count; i++)
{
	int x1 = x[i%3];
	int y1 = y[i/3];

	if (  abs(X - x1) < (x[1]-x[0])
       && abs(Y - y1) < (y[1]-y[0])
       )
    {
		target_node_index = i;
    	break;
    }
}


if (target_node_index == -1)
{
	Screen->Cursor = crNoDrop;
}

					// redraw as you drag
	                //TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
   	                //TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
		       	    //current_fragment->set_position(Point(X,Y));
					//FragmentsPaintBox->Repaint();
           		}
                else // outside painting area
                {
					Screen->Cursor = crNoDrop;
                }
			}
			else if ( !mouse_down )   // moving while a fragment is selected,
               						  // no dragging
            {
				// keep selected fragment's details displayed
   		    }
        }

        else // moving with no selection
        {
            bool fragment_found = false;
   			found_fragment_index = -1;

            for(int i=Network->Nodes->Count-1; i>=0; i--)
            {
              TNode* current_node = ((TNode*)Network->Nodes->Items[i]);
	          if (current_node != NULL)
    	      {
	            for(int j=current_node->Fragments->Count-1; j>=0; j--)
                {
                  TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[j]);
		          if (current_fragment != NULL)
    		      {
                    int fx = current_fragment->get_position().x;
   	                int fy = current_fragment->get_position().y;

            		int RadiusRadius = current_fragment->get_radius() * current_fragment->get_radius() + 2; // 2 is tolerance
	    	        int DistanceDistance = (fx-X)*(fx-X) + (fy-Y)*(fy-Y);
            		if (DistanceDistance < RadiusRadius)
		            {
						// display details when moving over fragment
						fragment_found = true;
			   			found_fragment_index = j;

						// don't repaint if moving inside fragment
					    if (old_found_fragment_index != found_fragment_index)
						{
						    current_fragment->DrawCaption(FragmentsPaintBox->Canvas);
							// clear details
                            ClustersPaintBox->Repaint();
						    current_fragment->DrawDetails(ClustersPaintBox->Canvas);
							MainForm->StatusBar->Panels->Items[2]->Text = "Fragment = " + IntToStr((int)current_fragment->get_id());
					    }

    	                break;
		            }
                  }
                }
              }
            }

            // moving outside fargments, with no selection
	        if (!fragment_found)
            {
				FragmentsPaintBox->Canvas->Pen->Color = clInfoBk;
				FragmentsPaintBox->Canvas->Brush->Color = clInfoBk;
				FragmentsPaintBox->Canvas->Rectangle(0, 0, 800, 15);

                // clear details
			    ClustersPaintBox->Repaint();

				MainForm->StatusBar->Panels->Items[2]->Text = "";
            }
        }
    }

   	old_found_fragment_index = found_fragment_index;
}






//---------------------------------------------------------------------------
void __fastcall TChildForm::FragmentsPaintBoxMouseUp(TObject *Sender, TMouseButton Button, TShiftState Shift, int X, int Y)
{
    mouse_down = false;
    if (Network != NULL)
    {
        if(!Network->IsEmpty())
        {
			if (inside_node_index != -1)
		    {
				TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
				if (inside_fragment_index != -1)
			    {
	            	if ( dragging )	  // end dragging
   			    	{
					    dragging = false;
						Screen->Cursor = crDefault;

	        	        TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);

               	    	if (  (target_node_index > -1) && (target_node_index < Network->Nodes->Count)  )
                   	    {
							if (Button == mbLeft)
						    {
		                        current_fragment->Migrate((TNode*)(Network->Nodes->Items[target_node_index]));
					    	}
                    	    else if (Button == mbRight)
                        	{
		                        current_fragment->Replicate((TNode*)(Network->Nodes->Items[target_node_index]));
                            }
        		        	target_node_index 		= -1;

						    inside_node_index 		= -1;
						    inside_fragment_index 	= -1;
						    inside_cluster_index  	= -1;
                        }

                        // show result of SO action
						NetworkPaintBox->Repaint();
						FragmentsPaintBox->Repaint();
						ClustersPaintBox->Repaint();

            	    }
	                else // if ( !dragging )   // selection action
		            {
				        // draw selected fragment and its node in selection color
						current_node->Select(selection_color);
						NetworkPaintBox->Repaint();

	        	        TFragment* selected_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
			   			selected_fragment->Select(selection_color);
						//FragmentsPaintBox->Repaint();
						// clear details
                        //ClustersPaintBox->Repaint();
					    //selected_fragment->DrawCaption(FragmentsPaintBox->Canvas);
					    //selected_fragment->DrawDetails(ClustersPaintBox->Canvas);
		    			//inside_cluster_index 	= -1;

                        /**/
                        //inside_cluster_index = 0;
			    	    //TCluster* selected_cluster = ((TCluster*)selected_fragment->Clusters->Items[inside_cluster_index]);
		   				//selected_cluster->Select(selection_color);
                        /**/

						FragmentsPaintBox->Repaint();   // also shows trackbars if cluster selected
						ClustersPaintBox->Repaint();    // update view for new selected fragment

	   			    }
	            }
            }
   	        else  // click outside fragments
            {
            	if (Button == mbRight)
                {
                  ClustersGroupBox->Visible = !ClustersGroupBox->Visible;
                }

               	Network->DeSelectChildren();
		    	inside_node_index		= -1;
		    	inside_fragment_index	= -1;
			    inside_cluster_index	= -1;
				NetworkPaintBox->Repaint();
				FragmentsPaintBox->Repaint();
				ClustersPaintBox->Repaint();
   	        }

			MainForm->UpdateMenuItems(NULL);
		}
    }
}




//---------------------------------------------------------------------------
void __fastcall TChildForm::FragmentsPaintBoxClick(TObject *Sender)
{
//
}


//---------------------------------------------------------------------------
void __fastcall TChildForm::FragmentsPaintBoxDblClick(TObject *Sender)
{
	mouse_dblclk = true;

    if (Network != NULL)
    {
		if (inside_node_index != -1)
    	{
		  TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
	      if (current_node != NULL)
    	  {
			if (inside_fragment_index != -1)
		    {
		      TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
		      if (current_fragment != NULL)
    		  {
            	MigrationForm->FC = current_fragment->MigrationFC;
                MigrationForm->UpdateDisplay();
                MigrationForm->ShowModal();

                ReplicationForm->FC = current_fragment->ReplicationFC;
                ReplicationForm->UpdateDisplay();
                ReplicationForm->ShowModal();


			    /*
                Application->MessageBox
                (
	            	(
    	           	"Id = " + IntToStr((int)current_fragment->get_id()) + "     "
        	        +
            	   	"Size = " + IntToStr((int)current_fragment->get_size()) + "%" + "     "
                    +
                	"Demand = " + IntToStr((int)current_fragment->get_demand()) + "%"
	                ).c_str(),

                    "Fragment",

    	            MB_OK
	            );
                */

				current_node->DeSelectChildren();
		    	inside_node_index		= -1;
		    	inside_fragment_index	= -1;
			    inside_cluster_index	= -1;
				NetworkPaintBox->Repaint();
				FragmentsPaintBox->Repaint();
				ClustersPaintBox->Repaint();
              }
		    }
            else  // double_click outside fragments
            {
				current_node->DeSelectChildren();
		    	inside_node_index		= -1;
		    	inside_fragment_index	= -1;
			    inside_cluster_index	= -1;
				NetworkPaintBox->Repaint();
				FragmentsPaintBox->Repaint();
				ClustersPaintBox->Repaint();
            }

			MainForm->UpdateMenuItems(NULL);
          }
        }
    }
}






//---------------------------------------------------------------------------
void __fastcall TChildForm::FragmentsPaintBoxPaint(TObject *Sender)
{
    if(Network != NULL)
    {
		FragmentsPaintBox->Canvas->Pen->Color = clInfoBk;
		FragmentsPaintBox->Canvas->Brush->Color = clInfoBk;
		FragmentsPaintBox->Canvas->Rectangle(0, 0, 800, 15);

		FragmentsPaintBox->Canvas->Pen->Color = clFuchsia;
		FragmentsPaintBox->Canvas->Brush->Color = clWhite;
		for (int i=0; i<Network->Nodes->Count; i++)
        {
            TNode* current_node = ((TNode*)Network->Nodes->Items[i]);
	        current_node->DrawDetails(FragmentsPaintBox->Canvas);
        }

		if (inside_node_index != -1)
	    {
		  TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
	      if (current_node != NULL)
    	  {
			if (inside_fragment_index != -1)
		    {
	    	  TFragment* current_fragment = (TFragment*)current_node->Fragments->Items[inside_fragment_index];
		      if (current_fragment != NULL)
    		  {
	        	current_fragment->DrawCaption(FragmentsPaintBox->Canvas);
				MainForm->StatusBar->Panels->Items[2]->Text = "Fragment = " + IntToStr((int)current_fragment->get_id());

		        if (inside_cluster_index != -1)
		        {
	    		  TCluster* current_cluster = ((TCluster*)current_fragment->Clusters->Items[inside_cluster_index]);
			      if (current_cluster != NULL)
    			  {
			        current_cluster->DrawCaption(ClustersPaintBox->Canvas);
					MainForm->StatusBar->Panels->Items[2]->Text = "Cluster = " + IntToStr((int)current_cluster->get_id());

//                    if (Network->Nodes->Count > 0) { ProgressBar1->Visible = true; ProgressBar1->Position = current_fragment->MigrationFC->migrations[0]; }
//                    if (Network->Nodes->Count > 1) { ProgressBar2->Visible = true; ProgressBar2->Position = current_fragment->MigrationFC->migrations[1]; }
//                    if (Network->Nodes->Count > 2) { ProgressBar3->Visible = true; ProgressBar3->Position = current_fragment->MigrationFC->migrations[2]; }
//                    if (Network->Nodes->Count > 3) { ProgressBar4->Visible = true; ProgressBar4->Position = current_fragment->MigrationFC->migrations[3]; }
//                    if (Network->Nodes->Count > 4) { ProgressBar5->Visible = true; ProgressBar5->Position = current_fragment->MigrationFC->migrations[4]; }
//                    if (Network->Nodes->Count > 5) { ProgressBar6->Visible = true; ProgressBar6->Position = current_fragment->MigrationFC->migrations[5]; }
//                    if (Network->Nodes->Count > 6) { ProgressBar7->Visible = true; ProgressBar7->Position = current_fragment->MigrationFC->migrations[6]; }
//                    if (Network->Nodes->Count > 7) { ProgressBar8->Visible = true; ProgressBar8->Position = current_fragment->MigrationFC->migrations[7]; }
//                    if (Network->Nodes->Count > 8) { ProgressBar9->Visible = true; ProgressBar9->Position = current_fragment->MigrationFC->migrations[8]; }


		        	if (Network->Nodes->Count > 0) { TrackBar1->Visible = true; TrackBar1->Position = current_cluster->demands[0]; }
        			if (Network->Nodes->Count > 1) { TrackBar2->Visible = true; TrackBar2->Position = current_cluster->demands[1]; }
		        	if (Network->Nodes->Count > 2) { TrackBar3->Visible = true; TrackBar3->Position = current_cluster->demands[2]; }
        			if (Network->Nodes->Count > 3) { TrackBar4->Visible = true; TrackBar4->Position = current_cluster->demands[3]; }
			       	if (Network->Nodes->Count > 4) { TrackBar5->Visible = true; TrackBar5->Position = current_cluster->demands[4]; }
		        	if (Network->Nodes->Count > 5) { TrackBar6->Visible = true; TrackBar6->Position = current_cluster->demands[5]; }
        			if (Network->Nodes->Count > 6) { TrackBar7->Visible = true; TrackBar7->Position = current_cluster->demands[6]; }
		        	if (Network->Nodes->Count > 7) { TrackBar8->Visible = true; TrackBar8->Position = current_cluster->demands[7]; }
        			if (Network->Nodes->Count > 8) { TrackBar9->Visible = true; TrackBar9->Position = current_cluster->demands[8]; }

		        	if (Network->Nodes->Count > 0) { TrackBar11->Visible = true; TrackBar11->Position = current_cluster->update_ratios[0]; }
        			if (Network->Nodes->Count > 1) { TrackBar12->Visible = true; TrackBar12->Position = current_cluster->update_ratios[1]; }
		        	if (Network->Nodes->Count > 2) { TrackBar13->Visible = true; TrackBar13->Position = current_cluster->update_ratios[2]; }
        			if (Network->Nodes->Count > 3) { TrackBar14->Visible = true; TrackBar14->Position = current_cluster->update_ratios[3]; }
			       	if (Network->Nodes->Count > 4) { TrackBar15->Visible = true; TrackBar15->Position = current_cluster->update_ratios[4]; }
		        	if (Network->Nodes->Count > 5) { TrackBar16->Visible = true; TrackBar16->Position = current_cluster->update_ratios[5]; }
        			if (Network->Nodes->Count > 6) { TrackBar17->Visible = true; TrackBar17->Position = current_cluster->update_ratios[6]; }
		        	if (Network->Nodes->Count > 7) { TrackBar18->Visible = true; TrackBar18->Position = current_cluster->update_ratios[7]; }
        			if (Network->Nodes->Count > 8) { TrackBar19->Visible = true; TrackBar19->Position = current_cluster->update_ratios[8]; }
                  }
		        }
        		else
		        {
					MainForm->StatusBar->Panels->Items[2]->Text = "";

//                    ProgressBar1->Visible = false; ProgressBar1->Position = 0;
//                    ProgressBar2->Visible = false; ProgressBar2->Position = 0;
//                    ProgressBar3->Visible = false; ProgressBar3->Position = 0;
//                    ProgressBar4->Visible = false; ProgressBar4->Position = 0;
//                    ProgressBar5->Visible = false; ProgressBar5->Position = 0;
//                    ProgressBar6->Visible = false; ProgressBar6->Position = 0;
//                    ProgressBar7->Visible = false; ProgressBar7->Position = 0;
//                    ProgressBar8->Visible = false; ProgressBar8->Position = 0;
//                    ProgressBar9->Visible = false; ProgressBar9->Position = 0;

                    TrackBar1->Visible = false;  TrackBar1->Position = 0;
        			TrackBar2->Visible = false;  TrackBar2->Position = 0;
		        	TrackBar3->Visible = false;  TrackBar3->Position = 0;
        			TrackBar4->Visible = false;  TrackBar4->Position = 0;
		        	TrackBar5->Visible = false;  TrackBar5->Position = 0;
        			TrackBar6->Visible = false;  TrackBar6->Position = 0;
		        	TrackBar7->Visible = false;  TrackBar7->Position = 0;
        			TrackBar8->Visible = false;  TrackBar8->Position = 0;
		        	TrackBar9->Visible = false;  TrackBar9->Position = 0;

                    TrackBar11->Visible = false;  TrackBar11->Position = 0;
        			TrackBar12->Visible = false;  TrackBar12->Position = 0;
		        	TrackBar13->Visible = false;  TrackBar13->Position = 0;
        			TrackBar14->Visible = false;  TrackBar14->Position = 0;
		        	TrackBar15->Visible = false;  TrackBar15->Position = 0;
        			TrackBar16->Visible = false;  TrackBar16->Position = 0;
		        	TrackBar17->Visible = false;  TrackBar17->Position = 0;
        			TrackBar18->Visible = false;  TrackBar18->Position = 0;
		        	TrackBar19->Visible = false;  TrackBar19->Position = 0;
		        }
              }
            }
            else
            {
				MainForm->StatusBar->Panels->Items[2]->Text = "Node = " + IntToStr((int)current_node->get_id());

//                ProgressBar1->Visible = false; ProgressBar1->Position = 0;
//                ProgressBar2->Visible = false; ProgressBar2->Position = 0;
//                ProgressBar3->Visible = false; ProgressBar3->Position = 0;
//                ProgressBar4->Visible = false; ProgressBar4->Position = 0;
//                ProgressBar5->Visible = false; ProgressBar5->Position = 0;
//                ProgressBar6->Visible = false; ProgressBar6->Position = 0;
//                ProgressBar7->Visible = false; ProgressBar7->Position = 0;
//                ProgressBar8->Visible = false; ProgressBar8->Position = 0;
//                ProgressBar9->Visible = false; ProgressBar9->Position = 0;

	        	TrackBar1->Visible = false;  TrackBar1->Position = 0;
       			TrackBar2->Visible = false;  TrackBar2->Position = 0;
	        	TrackBar3->Visible = false;  TrackBar3->Position = 0;
       			TrackBar4->Visible = false;  TrackBar4->Position = 0;
	        	TrackBar5->Visible = false;  TrackBar5->Position = 0;
       			TrackBar6->Visible = false;  TrackBar6->Position = 0;
	        	TrackBar7->Visible = false;  TrackBar7->Position = 0;
       			TrackBar8->Visible = false;  TrackBar8->Position = 0;
	        	TrackBar9->Visible = false;  TrackBar9->Position = 0;

                TrackBar11->Visible = false;  TrackBar11->Position = 0;
       			TrackBar12->Visible = false;  TrackBar12->Position = 0;
	        	TrackBar13->Visible = false;  TrackBar13->Position = 0;
       			TrackBar14->Visible = false;  TrackBar14->Position = 0;
	        	TrackBar15->Visible = false;  TrackBar15->Position = 0;
       			TrackBar16->Visible = false;  TrackBar16->Position = 0;
	        	TrackBar17->Visible = false;  TrackBar17->Position = 0;
       			TrackBar18->Visible = false;  TrackBar18->Position = 0;
	        	TrackBar19->Visible = false;  TrackBar19->Position = 0;
            }
          }
        }
   	    else
       	{
			MainForm->StatusBar->Panels->Items[2]->Text = "";

//            ProgressBar1->Visible = false; ProgressBar1->Position = 0;
//            ProgressBar2->Visible = false; ProgressBar2->Position = 0;
//            ProgressBar3->Visible = false; ProgressBar3->Position = 0;
//            ProgressBar4->Visible = false; ProgressBar4->Position = 0;
//            ProgressBar5->Visible = false; ProgressBar5->Position = 0;
//            ProgressBar6->Visible = false; ProgressBar6->Position = 0;
//            ProgressBar7->Visible = false; ProgressBar7->Position = 0;
//            ProgressBar8->Visible = false; ProgressBar8->Position = 0;
//            ProgressBar9->Visible = false; ProgressBar9->Position = 0;

        	TrackBar1->Visible = false;  TrackBar1->Position = 0;
   			TrackBar2->Visible = false;  TrackBar2->Position = 0;
        	TrackBar3->Visible = false;  TrackBar3->Position = 0;
   			TrackBar4->Visible = false;  TrackBar4->Position = 0;
        	TrackBar5->Visible = false;  TrackBar5->Position = 0;
   			TrackBar6->Visible = false;  TrackBar6->Position = 0;
	       	TrackBar7->Visible = false;  TrackBar7->Position = 0;
    		TrackBar8->Visible = false;  TrackBar8->Position = 0;
	       	TrackBar9->Visible = false;  TrackBar9->Position = 0;

            TrackBar11->Visible = false;  TrackBar11->Position = 0;
   			TrackBar12->Visible = false;  TrackBar12->Position = 0;
        	TrackBar13->Visible = false;  TrackBar13->Position = 0;
   			TrackBar14->Visible = false;  TrackBar14->Position = 0;
        	TrackBar15->Visible = false;  TrackBar15->Position = 0;
   			TrackBar16->Visible = false;  TrackBar16->Position = 0;
        	TrackBar17->Visible = false;  TrackBar17->Position = 0;
   			TrackBar18->Visible = false;  TrackBar18->Position = 0;
        	TrackBar19->Visible = false;  TrackBar19->Position = 0;
       	}

        FragmentsPaintBox->Canvas->Refresh();
    }
}












//---------------------------------------------------------------------------
void __fastcall TChildForm::ClustersPaintBoxMouseDown(TObject *Sender, TMouseButton Button, TShiftState Shift, int X, int Y)
{
	// patch to stop dragging after dblclk's mouse_down undesirable event
	if (mouse_dblclk)
    {
		mouse_dblclk = false; 	//  clear for next mouse_down event
    }
    else	// otherwise proceed as normal
    {
        mouse_down = true;

    	if (Network != NULL)
    	{
        	if(!Network->IsEmpty())
	        {
				if(inside_node_index != -1)
			    {
			        TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
					if(inside_fragment_index != -1)
			    	{
				        TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);

			        	current_fragment->DeSelectChildren();
			            inside_cluster_index = -1;


				        for(int i=current_fragment->Clusters->Count-1; i>=0; i--)
			    	    {
					        TCluster* current_cluster = ((TCluster*)current_fragment->Clusters->Items[i]);
			    	        int cx = current_cluster->get_position().x;
			        	    int cy = current_cluster->get_position().y;
				       		int RadiusRadius = current_cluster->get_radius() * current_cluster->get_radius() + 2; // 2 is tolerance
			   		        int DistanceDistance = (cx-X)*(cx-X) + (cy-Y)*(cy-Y);
				       		if (DistanceDistance < RadiusRadius)
			    	        {
				                inside_cluster_index = i;
			    	       	    return;
				            }
			    	    }
			        }
			    }
			}
        }
    }
}



//---------------------------------------------------------------------------
void __fastcall TChildForm::ClustersPaintBoxMouseMove(TObject *Sender, TShiftState Shift, int X, int Y)
{
   	static int old_found_cluster_index;
   	       int     found_cluster_index;

    if (Network != NULL)
    {
		// a node is current
        if (inside_node_index != -1)
        {
			// a fragment is current
		    if ( inside_fragment_index != -1 )
			{
				// a cluster is current
			    if ( inside_cluster_index != -1 )
				{
		           	if ( mouse_down )	  // start dragging
   			        {
	    		        dragging = true;
						//Screen->Cursor = crDrag;

		                // move inside painting area
    		            if ((X > 0) && (Y > 0) && (X < ClustersPaintBox->Width) && (Y < ClustersPaintBox->Height))
       			        {
                            /*
			                TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
		   	                TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
		   	                TCluster* current_cluster = ((TCluster*)current_fragment->Clusters->Items[inside_cluster_index]);
				       	    current_cluster->set_position(Point(X,Y));
							// redraw as you drag
					    	ClustersPaintBox->Repaint();
                        	*/
    	    	   		}
        	    	    else // outside painting area
		                {
							//Screen->Cursor = crNoDrop;
	        	        }
					}
					else if ( !mouse_down )   // moving while a cluster is selected,
	    	           						  // no dragging
	    	        {
						// keep selected cluster's details displayed
   				    }
		        }

    		    else // moving with no selection
   	    	    {
                  TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
				  if (current_node != NULL)
    			  {
   	                TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
				    if (current_fragment != NULL)
    			    {
						// display details when moving over cluster
		        	    bool cluster_found = false;
						found_cluster_index = -1;

		    	    	for(int i=current_fragment->Clusters->Count-1; i>=0; i--)
				        {
	    				  TCluster* current_cluster = ((TCluster*)current_fragment->Clusters->Items[i]);
					      if (current_cluster != NULL)
	    				  {
			    	        int cx = current_cluster->get_position().x;
	    	   			    int cy = current_cluster->get_position().y;
		    		   		int RadiusRadius = current_cluster->get_radius() * current_cluster->get_radius() + 2; // 2 is tolerance
			   		        int DistanceDistance = (cx-X)*(cx-X) + (cy-Y)*(cy-Y);
	        				if (DistanceDistance < RadiusRadius)
		    	    	    {
								// display cluster's details when moving over it even if not current
								cluster_found = true;
								found_cluster_index = i;

								// don't repaint if moving inside cluster
							    if (old_found_cluster_index != found_cluster_index)
								{
								    current_cluster->DrawCaption(ClustersPaintBox->Canvas);
									MainForm->StatusBar->Panels->Items[2]->Text = "Cluster = " + IntToStr((int)current_cluster->get_id());
							    }
   				                break;
				            }
    	                  }
		   	            }
	    	            // moving outside clusters, with no selection
    			        if (!cluster_found)
		                {
							ClustersPaintBox->Canvas->Pen->Color = clInfoBk;
							ClustersPaintBox->Canvas->Brush->Color = clInfoBk;
							ClustersPaintBox->Canvas->Rectangle(0, 0, 800, 15);

							MainForm->StatusBar->Panels->Items[2]->Text = "";
	   			        }
                    }
                  }
	            }
            }
        }
    }

   	old_found_cluster_index = found_cluster_index;
}



//---------------------------------------------------------------------------
void __fastcall TChildForm::ClustersPaintBoxMouseUp(TObject *Sender, TMouseButton Button, TShiftState Shift, int X, int Y)
{
    mouse_down = false;
    if (Network != NULL)
    {
        if(!Network->IsEmpty())
        {
			if (inside_node_index != -1)
		    {
				TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
				if (inside_fragment_index != -1)
			    {
			        TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
					if (inside_cluster_index != -1)
				    {
		            	if ( dragging )	  // end dragging
    		   	    	{
       					    dragging = false;
							//Screen->Cursor = crDefault;
						    inside_cluster_index  = -1;
	            	    }
						else //if ( !dragging )   // selection action
			            {
				        	// draw selected cluster in selection color
				    	    TCluster* selected_cluster = ((TCluster*)current_fragment->Clusters->Items[inside_cluster_index]);
			   				selected_cluster->Select(selection_color);
							FragmentsPaintBox->Repaint();  // show trackbars
							ClustersPaintBox->Repaint();
		   			    }
		            }
			        else  // click outside clusters
		    	    {
                    	current_fragment->DeSelectChildren();
	    		        inside_cluster_index = -1;
						//ClustersPaintBox->Repaint();
						NetworkPaintBox->Repaint();
						FragmentsPaintBox->Repaint();	// hide trackbars
						ClustersPaintBox->Repaint();
			        }

					MainForm->UpdateMenuItems(NULL);
			    }
		    }
        }
    }
}



//---------------------------------------------------------------------------
void __fastcall TChildForm::ClustersPaintBoxClick(TObject *Sender)
{
//
}



//---------------------------------------------------------------------------
void __fastcall TChildForm::ClustersPaintBoxDblClick(TObject *Sender)
{
	mouse_dblclk = true;

    if (Network != NULL)
    {
		if (inside_node_index != -1)
    	{
			TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
			if (inside_fragment_index != -1)
	    	{
		        TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
				if (inside_cluster_index != -1)
			    {
		    	    TCluster* current_cluster = ((TCluster*)current_fragment->Clusters->Items[inside_cluster_index]);

                    Application->MessageBox
                    (
    	                ("Id = " + IntToStr((int)current_cluster->get_id()) + "     "
        	            +
            	        "Size = " + IntToStr((int)current_cluster->get_size()) + "%" + "     "
                	    +
                    	"Demand = " + IntToStr((int)current_cluster->get_demand()) + "%"
	                    ).c_str(),

                        "Cluster",

                        MB_OK
                    );

					current_fragment->DeSelectChildren();
    		        inside_cluster_index = -1;
					ClustersPaintBox->Repaint();
			    }
    	        else  // double_click outside clusters
        	    {
					current_fragment->DeSelectChildren();
    		        inside_cluster_index = -1;
					ClustersPaintBox->Repaint();
	            }
    	    }
        }
    }
}



//---------------------------------------------------------------------------
void __fastcall TChildForm::ClustersPaintBoxPaint(TObject *Sender)
{
    if(Network != NULL)
    {
		ClustersPaintBox->Canvas->Pen->Color = clInfoBk;
		ClustersPaintBox->Canvas->Brush->Color = clInfoBk;
		ClustersPaintBox->Canvas->Rectangle(0, 0, 800, 15);

	    // light aqua background
		ClustersPaintBox->Canvas->Pen->Color = (TColor) RGB(240, 255, 255);
		ClustersPaintBox->Canvas->Brush->Color = (TColor) RGB(240, 255, 255);
		ClustersPaintBox->Canvas->Rectangle(0, 15, 800, 400);

		if (inside_node_index != -1)
    	{
		  TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
		  if (current_node != NULL)
          {
			if (inside_fragment_index != -1)
		    {
		      TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
			  if (current_fragment != NULL)
    	      {
			    current_fragment->DrawDetails(ClustersPaintBox->Canvas);

		        if (inside_cluster_index != -1)
		        {
	    		  TCluster* current_cluster = ((TCluster*)current_fragment->Clusters->Items[inside_cluster_index]);
				  if (current_cluster != NULL)
    		      {
			        current_cluster->DrawCaption(ClustersPaintBox->Canvas);
					MainForm->StatusBar->Panels->Items[2]->Text = "Cluster = " + IntToStr((int)current_cluster->get_id());
                    /*
		        	if (Network->Nodes->Count > 0) { TrackBar1->Position = current_cluster->demands[0];  TrackBar1->Visible = true; }
        			if (Network->Nodes->Count > 1) { TrackBar2->Position = current_cluster->demands[1];  TrackBar2->Visible = true; }
		        	if (Network->Nodes->Count > 2) { TrackBar3->Position = current_cluster->demands[2];  TrackBar3->Visible = true; }
        			if (Network->Nodes->Count > 3) { TrackBar4->Position = current_cluster->demands[3];  TrackBar4->Visible = true; }
			       	if (Network->Nodes->Count > 4) { TrackBar5->Position = current_cluster->demands[4];  TrackBar5->Visible = true; }
		        	if (Network->Nodes->Count > 5) { TrackBar6->Position = current_cluster->demands[5];  TrackBar6->Visible = true; }
        			if (Network->Nodes->Count > 6) { TrackBar7->Position = current_cluster->demands[6];  TrackBar7->Visible = true; }
		        	if (Network->Nodes->Count > 7) { TrackBar8->Position = current_cluster->demands[7];  TrackBar8->Visible = true; }
        			if (Network->Nodes->Count > 8) { TrackBar9->Position = current_cluster->demands[8];  TrackBar9->Visible = true; }

		        	if (Network->Nodes->Count > 0) { TrackBar11->Visible = true; TrackBar11->Position = current_cluster->update_ratios[0]; }
        			if (Network->Nodes->Count > 1) { TrackBar12->Visible = true; TrackBar12->Position = current_cluster->update_ratios[1]; }
		        	if (Network->Nodes->Count > 2) { TrackBar13->Visible = true; TrackBar13->Position = current_cluster->update_ratios[2]; }
        			if (Network->Nodes->Count > 3) { TrackBar14->Visible = true; TrackBar14->Position = current_cluster->update_ratios[3]; }
			       	if (Network->Nodes->Count > 4) { TrackBar15->Visible = true; TrackBar15->Position = current_cluster->update_ratios[4]; }
		        	if (Network->Nodes->Count > 5) { TrackBar16->Visible = true; TrackBar16->Position = current_cluster->update_ratios[5]; }
        			if (Network->Nodes->Count > 6) { TrackBar17->Visible = true; TrackBar17->Position = current_cluster->update_ratios[6]; }
		        	if (Network->Nodes->Count > 7) { TrackBar18->Visible = true; TrackBar18->Position = current_cluster->update_ratios[7]; }
        			if (Network->Nodes->Count > 8) { TrackBar19->Visible = true; TrackBar19->Position = current_cluster->update_ratios[8]; }
                    */
                  }
		        }
        		else
		        {
					MainForm->StatusBar->Panels->Items[2]->Text = "";
                    ///*
//                    ProgressBar1->Visible = false; ProgressBar1->Position = 0;
//                    ProgressBar2->Visible = false; ProgressBar2->Position = 0;
//                    ProgressBar3->Visible = false; ProgressBar3->Position = 0;
//                    ProgressBar4->Visible = false; ProgressBar4->Position = 0;
//                    ProgressBar5->Visible = false; ProgressBar5->Position = 0;
//                    ProgressBar6->Visible = false; ProgressBar6->Position = 0;
//                    ProgressBar7->Visible = false; ProgressBar7->Position = 0;
//                    ProgressBar8->Visible = false; ProgressBar8->Position = 0;
//                    ProgressBar9->Visible = false; ProgressBar9->Position = 0;

		        	TrackBar1->Visible = false;  TrackBar1->Position = 0;
		   			TrackBar2->Visible = false;  TrackBar2->Position = 0;
		        	TrackBar3->Visible = false;  TrackBar3->Position = 0;
		   			TrackBar4->Visible = false;  TrackBar4->Position = 0;
		        	TrackBar5->Visible = false;  TrackBar5->Position = 0;
		   			TrackBar6->Visible = false;  TrackBar6->Position = 0;
			       	TrackBar7->Visible = false;  TrackBar7->Position = 0;
		    		TrackBar8->Visible = false;  TrackBar8->Position = 0;
			       	TrackBar9->Visible = false;  TrackBar9->Position = 0;

		            TrackBar11->Visible = false;  TrackBar11->Position = 0;
   					TrackBar12->Visible = false;  TrackBar12->Position = 0;
		        	TrackBar13->Visible = false;  TrackBar13->Position = 0;
   					TrackBar14->Visible = false;  TrackBar14->Position = 0;
		        	TrackBar15->Visible = false;  TrackBar15->Position = 0;
   					TrackBar16->Visible = false;  TrackBar16->Position = 0;
		        	TrackBar17->Visible = false;  TrackBar17->Position = 0;
   					TrackBar18->Visible = false;  TrackBar18->Position = 0;
		        	TrackBar19->Visible = false;  TrackBar19->Position = 0;
                    //*/
		        }
              }
            }
          }
		}

        ClustersPaintBox->Canvas->Refresh();
	}
}













//---------------------------------------------------------------------------
void __fastcall TChildForm::ResponsePaintBoxPaint(TObject *Sender)
{
    if(Network != NULL)
    {
		int xZoomScale = 1;
		int yZoomScale = 2;
		int Points = 100;
        int xGap   = 2;
		//int Width	= ResponsePaintBox->Width;
		int Height	= ResponsePaintBox->Height;
		int xShift = 20;
		int yShift = 20;
		int xMin   = 0          /xZoomScale;
		int xMax   = Points*xGap/xZoomScale;
		int yMin   = 0          /yZoomScale;
		int yMax   = 100        /yZoomScale;
		int xDivisions =  10/xZoomScale;
		int yDivisions = 10/yZoomScale;
		int xScale = (xMax-xMin)/xDivisions;
		int yScale = (yMax-yMin)/yDivisions;
        int xDivMarkHeight	= 4;
        int yDivMarkWidth	= 4;
		int xTextSize	= 7;
		int yTextSize	= 7;
		int xXTextShift = -4;
		int yXTextShift = +4;
		int xYTextShift = -18;
		int yYTextShift = -4;

		TCanvas* _Canvas = ResponsePaintBox->Canvas;
        _Canvas->Pen->Color = clGray;
        _Canvas->Brush->Color = ResponsePaintBox->Color;
  		_Canvas->Font->Name  = "Arial";
        _Canvas->Font->Color = clGray;

        _Canvas->Font->Size = xTextSize;
        _Canvas->MoveTo(xMin+xShift, Height-yShift);
        _Canvas->LineTo(xMax+xShift, Height-yShift);
        for(int i=0; i<=xDivisions; i++)
        {
    	    _Canvas->TextOut(xMin+xShift + i*xScale + xXTextShift, Height-yShift + yXTextShift, IntToStr((int)-((xDivisions-i)*(xScale/xGap)*xZoomScale)));
	        _Canvas->MoveTo (xMin+xShift + i*xScale              , Height-yShift -              0);
    	    _Canvas->LineTo (xMin+xShift + i*xScale              , Height-yShift + xDivMarkHeight);
        }

        _Canvas->Font->Size = yTextSize;
        _Canvas->MoveTo(xMin+xShift,      Height-yShift-yMin);
        _Canvas->LineTo(xMin+xShift,      Height-yShift-yMax);
        for(int i=0; i<=yDivisions; i++)
        {
    	    _Canvas->TextOut(xMin+xShift + xYTextShift,   Height-yShift - i*yScale + yYTextShift, IntToStr((int)i*yScale*yZoomScale));
	        _Canvas->MoveTo (xMin+xShift - yDivMarkWidth, Height-yShift - i*yScale);
    	    _Canvas->LineTo (xMin+xShift +             0, Height-yShift - i*yScale);
        }



        for(int i=Network->ResponseTimeHistory->Count-Points; i<Network->ResponseTimeHistory->Count; i++)
        {
            if (i<0) continue;

            int ResponseTime = *((int*)Network->ResponseTimeHistory->Items[i]);
           	_Canvas->Pen->Color = clBlue;
			int x = i - (Network->ResponseTimeHistory->Count - Points);
	        _Canvas->MoveTo(xGap*(x+1)/xZoomScale+xMin+xShift, Height-yShift);
	        _Canvas->LineTo(xGap*(x+1)/xZoomScale+xMin+xShift, Height-yShift - ResponseTime/yZoomScale);

    		_Canvas->Pen->Color = clInfoBk;
	    	_Canvas->Brush->Color = clInfoBk;
		    _Canvas->Rectangle(0,0, 800, 20);
       }


       int rt_count = 0;
       int sum = 0;
       for(int i=Network->ResponseTimeHistory->Count-Points; i<Network->ResponseTimeHistory->Count; i++)
       {
         if (i<0) continue;
         sum +=  *((int*)Network->ResponseTimeHistory->Items[i]);
         rt_count++;
       }
       if(Network->ResponseTimeHistory->Count > 0)
       {
        float AverageResponseTime = 0.0;
        if (rt_count > 0)
        {
            AverageResponseTime = sum / rt_count;
        }

       	_Canvas->Font->Color = clBlue;
    	_Canvas->Font->Size  = 10;
   		_Canvas->TextOut(30,  0, IntToStr((int)Network->minResponse));
   		_Canvas->TextOut(78, 0, IntToStr((int)AverageResponseTime));
   		_Canvas->TextOut(133, 0, IntToStr((int)Network->maxResponse));

       	_Canvas->Font->Color = clBlack;
    	_Canvas->Font->Size  = 10;
   		_Canvas->TextOut(193, 0, IntToStr((int)*((int*)Network->ResponseTimeHistory->Items[Network->ResponseTimeHistory->Count-1])));
       }

       	_Canvas->Refresh();
	}
}








//---------------------------------------------------------------------------
void __fastcall TChildForm::ResourcePaintBoxPaint(TObject *Sender)
{
    if(Network != NULL)
    {
		int xZoomScale = 1;
		int yZoomScale = 2;
		int Points = 100;
        int xGap   = 2;
		//int Width	= ResourcePaintBox->Width;
		int Height	= ResourcePaintBox->Height;
		int xShift = 20;
		int yShift = 20;
		int xMin   = 0          /xZoomScale;
		int xMax   = Points*xGap/xZoomScale;
		int yMin   = 0          /yZoomScale;
		int yMax   = 100        /yZoomScale;
		int xDivisions = 10/xZoomScale;
		int yDivisions = 10/yZoomScale;
		int xScale = (xMax-xMin)/xDivisions;
		int yScale = (yMax-yMin)/yDivisions;
        int xDivMarkHeight	= 4;
        int yDivMarkWidth	= 4;
		int xTextSize	= 7;
		int yTextSize	= 7;
		int xXTextShift = -4;
		int yXTextShift = +4;
		int xYTextShift = -18;
		int yYTextShift = -4;

		TCanvas* _Canvas = ResourcePaintBox->Canvas;
        _Canvas->Pen->Color = clGray;
        _Canvas->Brush->Color = ResourcePaintBox->Color;
  		_Canvas->Font->Name  = "Arial";
        _Canvas->Font->Color = clGray;

        _Canvas->Font->Size = xTextSize;
        _Canvas->MoveTo(xMin+xShift, Height-yShift);
        _Canvas->LineTo(xMax+xShift, Height-yShift);
        for(int i=0; i<=xDivisions; i++)
        {
    	    _Canvas->TextOut(xMin+xShift + i*xScale + xXTextShift, Height-yShift + yXTextShift, IntToStr((int)-((xDivisions-i)*(xScale/xGap)*xZoomScale)));
	        _Canvas->MoveTo (xMin+xShift + i*xScale              , Height-yShift -              0);
    	    _Canvas->LineTo (xMin+xShift + i*xScale              , Height-yShift + xDivMarkHeight);
        }

        _Canvas->Font->Size = yTextSize;
        _Canvas->MoveTo(xMin+xShift,      Height-yShift-yMin);
        _Canvas->LineTo(xMin+xShift,      Height-yShift-yMax);
        for(int i=0; i<=yDivisions; i++)
        {
    	    _Canvas->TextOut(xMin+xShift + xYTextShift,   Height-yShift - i*yScale + yYTextShift, IntToStr((int)i*yScale*yZoomScale));
	        _Canvas->MoveTo (xMin+xShift - yDivMarkWidth, Height-yShift - i*yScale);
    	    _Canvas->LineTo (xMin+xShift +             0, Height-yShift - i*yScale);
        }

        /*
        int fragments = 0;
        for(int i=0; i<Network->Nodes->Count; i++)
        {
          TNode* current_node = ((TNode*)Network->Nodes->Items[i]);
       	  fragments += current_node->Fragments->Count;
        }
        */
        for(int i=Network->ResourceUsageHistory->Count - Points; i<Network->ResourceUsageHistory->Count; i++)
        {
            if (i<0) continue;

			int x = i - (Network->ResourceUsageHistory->Count - Points);
       		if ( (SO_FREQUENCY > 0) && ((i%(MAX_SAMPLES/SO_FREQUENCY)) == 0) )
            {
				_Canvas->Font->Color = clRed;
	    		_Canvas->Brush->Color = cl3DLight;
                _Canvas->TextOut(xGap*(x+1)/xZoomScale+xMin+xShift, Height-yShift - 100/yZoomScale - 15,
                                    IntToStr((int)so_activity[i/(MAX_SAMPLES/SO_FREQUENCY)])
                                );
                //_Canvas->TextOut(xGap*(x+1)/xZoomScale+xMin+xShift, Height-yShift - 100/yZoomScale - 15,
                //                    IntToStr((int)((so_activity[i/(MAX_SAMPLES/SO_FREQUENCY)]+1) * fragments)/100.0)// + "/" + IntToStr((int)fragments)
                //                );
            }


            int ResourceUsage = *((int*)Network->ResourceUsageHistory->Items[i]);
           	_Canvas->Pen->Color = clGreen;
	        _Canvas->MoveTo(xGap*(x+1)/xZoomScale+xMin+xShift, Height-yShift);
	        _Canvas->LineTo(xGap*(x+1)/xZoomScale+xMin+xShift, Height-yShift - ResourceUsage/yZoomScale);


			_Canvas->Pen->Color = clInfoBk;
			_Canvas->Brush->Color = clInfoBk;
			_Canvas->Rectangle(0,0, 800, 20);
        }


       int ru_count = 0;
       int sum = 0;
       for(int i=Network->ResourceUsageHistory->Count-Points; i<Network->ResourceUsageHistory->Count; i++)
       {
         if (i<0) continue;
         sum +=  *((int*)Network->ResourceUsageHistory->Items[i]);
         ru_count++;
       }
       if(Network->ResourceUsageHistory->Count > 0)
       {
        float AverageResourceUsage = 0.0;
        if (ru_count > 0)
        {
            AverageResourceUsage = sum / ru_count;
        }
       	_Canvas->Font->Color = clGreen;
       	_Canvas->Font->Size  = 10;
    	_Canvas->TextOut(30,  0, IntToStr((int)Network->minResource));
	    _Canvas->TextOut(80, 0, IntToStr((int)AverageResourceUsage));
   		_Canvas->TextOut(132, 0, IntToStr((int)Network->maxResource));

       	_Canvas->Font->Color = clBlack;
    	_Canvas->Font->Size  = 10;
   		_Canvas->TextOut(193, 0, IntToStr((int)*((int*)Network->ResourceUsageHistory->Items[Network->ResourceUsageHistory->Count-1])));
       }

       _Canvas->Refresh();
	}
}
//---------------------------------------------------------------------------/*















//---------------------------------------------------------------------------
// disabled fot replication to work
void TChildForm::backup()
{
//    if (backup_network) delete backup_network;
//    backup_network = new TNetwork(Network);
//    to_be_undone = true;
//    to_be_redone = false;
}


//---------------------------------------------------------------------------
void TChildForm::undo()
{
	// swap
	TNetwork*		temp;
    temp			= Network;
	Network			= backup_network;
    backup_network	= temp;

    to_be_undone = false;
    to_be_redone = true;
}

//---------------------------------------------------------------------------
void TChildForm::redo()
{
	// swap
	TNetwork*		temp;
    temp			= Network;
	Network			= backup_network;
    backup_network	= temp;

    to_be_undone = true;
    to_be_redone = false;
}





//---------------------------------------------------------------------------
bool TChildForm::NoSelection()
{
	return (! Network->AnySelection() );
}


//---------------------------------------------------------------------------
bool TChildForm::ToBeUndone()
{
	return to_be_undone;
}

//---------------------------------------------------------------------------
bool TChildForm::ToBeRedone()
{
	return to_be_redone;
}


//---------------------------------------------------------------------------
bool TChildForm::Running()
{
	return running;
}












//---------------------------------------------------------------------------
void __fastcall TChildForm::FileNew()
{
	Network = new TNetwork();
	MainForm->UpdateMenuItems(NULL);
    Repaint();
}




//---------------------------------------------------------------------------
void __fastcall TChildForm::FileOpen()
{
	//
}




//---------------------------------------------------------------------------
void __fastcall TChildForm::FileSave()
{
	//
}




//---------------------------------------------------------------------------
void __fastcall TChildForm::FileSaveAs()
{
	//
}




//---------------------------------------------------------------------------
void __fastcall TChildForm::FilePrint()
{
	Print();
}







//---------------------------------------------------------------------------
void __fastcall TChildForm::EditUndo()
{
    if (to_be_undone)
    {
		undo();
		Network->DeSelectChildren();
        inside_node_index		= -1;
        inside_link_index		= -1;
        inside_fragment_index	= -1;
        inside_cluster_index	= -1;

		MainForm->UpdateMenuItems(NULL);
	    Repaint();
    }
}




//---------------------------------------------------------------------------
void __fastcall TChildForm::EditRedo()
{
    if (to_be_redone)
    {
		redo();
		Network->DeSelectChildren();
        inside_node_index		= -1;
        inside_link_index		= -1;
        inside_fragment_index	= -1;
        inside_cluster_index	= -1;

		MainForm->UpdateMenuItems(NULL);
	    Repaint();
    }
}




//---------------------------------------------------------------------------
void __fastcall TChildForm::EditDelete()
{
    	if (Network != NULL)
	    {
	        if (inside_cluster_index != -1)
    	    {
				backup();
				TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
		        TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
    		    current_fragment->DeleteCluster(inside_cluster_index);
                if (current_fragment->Clusters->Count == 0)
                {
					current_node->DeleteFragment(inside_fragment_index, true);
					Network->DeSelectChildren();
			        inside_node_index		= -1;
			        inside_link_index		= -1;
		    	    inside_fragment_index	= -1;
		        	inside_cluster_index	= -1;

                }
                else
                {
	                current_fragment->DeSelectChildren();
                }
        		inside_cluster_index = -1;
            }
	        else if (inside_fragment_index != -1)
    	    {
				backup();
				TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
    		    current_node->DeleteFragment(inside_fragment_index);
        		current_node->DeSelectChildren();
		        inside_fragment_index = -1;
		        inside_cluster_index = -1;
		    }
	        else if (inside_node_index != -1)
    	    {
				backup();
   	    	    Network->DeleteNodeAndItsLinks(inside_node_index);
				Network->DeSelectChildren();
    		    inside_node_index		= -1;
	    	    inside_link_index		= -1;
	    	    inside_fragment_index	= -1;
    	    	inside_cluster_index	= -1;
            }
            else if (inside_link_index != -1)
            {
				backup();
                Network->DeleteLink(inside_link_index);
				Network->DeSelectChildren();
		        inside_node_index		= -1;
		        inside_link_index		= -1;
		        inside_fragment_index	= -1;
		        inside_cluster_index	= -1;
            }

			MainForm->UpdateMenuItems(NULL);
		    Repaint();
	    }
}




//---------------------------------------------------------------------------
void __fastcall TChildForm::EditClear()
{
   	if (Network != NULL)
    {
		backup();
		Network->Clear();
		Network->DeSelectChildren();
        inside_node_index		= -1;
        inside_link_index		= -1;
        inside_fragment_index	= -1;
        inside_cluster_index	= -1;

		MainForm->UpdateMenuItems(NULL);
	    Repaint();
    }
}




//---------------------------------------------------------------------------
bool __fastcall TChildForm::NetworkAddNode()
{
	if (Network != NULL)
    {
	  if (TNode::serial_num < MAX_NODES)
      {
		backup();
       	Network->AddNode();
		Network->DeSelectChildren();
        inside_node_index		= -1;
        inside_link_index		= -1;
        inside_fragment_index	= -1;
        inside_cluster_index	= -1;

		MainForm->UpdateMenuItems(NULL);
	    Repaint();
        return true;
      }
    }
    return false;
}




//---------------------------------------------------------------------------
bool __fastcall TChildForm::NetworkAddLink()
{
    if (Network != NULL && Network->Nodes->Count >= 2)
   	{
	    LinkDialog->Link = NULL; // just used for initial values

		LinkDialog->node_a_index = -1;
		LinkDialog->node_b_index = -1;

		LinkDialog->Node_a_ComboBox->Clear();
	    for(int i=0; i<Network->Nodes->Count; i++)
    	{
			LinkDialog->Node_a_ComboBox->Items->Add(IntToStr((int) ((TNode*)Network->Nodes->Items[i])->get_id() ));
		}
		LinkDialog->Node_a_ComboBox->ItemIndex = LinkDialog->node_a_index;
    	LinkDialog->Node_a_ComboBox->Enabled = true;

		LinkDialog->Node_b_ComboBox->Clear();
   		for(int i=0; i<Network->Nodes->Count; i++)
    	{
			LinkDialog->Node_b_ComboBox->Items->Add(IntToStr((int) ((TNode*)Network->Nodes->Items[i])->get_id() ));
		}
		LinkDialog->Node_b_ComboBox->ItemIndex = LinkDialog->node_b_index;
   		LinkDialog->Node_b_ComboBox->Enabled = true;

		LinkDialog->TypeComboBox->ItemIndex = -1;

	    LinkDialog->DelayScrollBar->Position 		=  5;
    	LinkDialog->CapacityScrollBar->Position 	= 50;

		if (LinkDialog->ShowModal() == mrOk)
        {
			backup();
			int node_a_index = LinkDialog->node_a_index;
			int node_b_index = LinkDialog->node_b_index;
			TNode* node_a = (TNode*)(Network->Nodes->Items[node_a_index]);
			TNode* node_b = (TNode*)(Network->Nodes->Items[node_b_index]);

        	TLink* added_link = Network->AddLink(node_a, node_b);

        	added_link->set_type(LinkDialog->TypeComboBox->ItemIndex);
        	added_link->set_initial_delay(LinkDialog->DelayScrollBar->Position);
        	added_link->set_capacity(LinkDialog->CapacityScrollBar->Position);
        }

		Network->DeSelectChildren();
        inside_node_index		= -1;
        inside_link_index		= -1;
        inside_fragment_index	= -1;
        inside_cluster_index	= -1;

		MainForm->UpdateMenuItems(NULL);
	    Repaint();
        return true;
   	}
    return false;
}


//---------------------------------------------------------------------------
void __fastcall TChildForm::RepositoryDataTypes()
{
	//
}

//---------------------------------------------------------------------------
void __fastcall TChildForm::RepositoryStatistics()
{
	// 	StatisticsDialog->ShowModal();
}

//---------------------------------------------------------------------------
void __fastcall TChildForm::DomainProperties()
{
	//
}


//---------------------------------------------------------------------------
void __fastcall TChildForm::RunStart()
{
	running = true;
}


//---------------------------------------------------------------------------
void __fastcall TChildForm::RunStop()
{
	running = false;
}


//---------------------------------------------------------------------------
void __fastcall TChildForm::RunEnableSO()
{
    if (Network != NULL)
    {
        for(int i=0; i<Network->Nodes->Count; i++)
        {
            TNode* node = ((TNode*)Network->Nodes->Items[i]);
    	    for (int j=0; j<node->Fragments->Count; j++)
		    {
				TFragment* fragment = ((TFragment*)node->Fragments->Items[j]);
				fragment->SO_On();
		    }
        }
    }
}


//---------------------------------------------------------------------------
void __fastcall TChildForm::RunDisableSO()
{
    if (Network != NULL)
    {
        for(int i=0; i<Network->Nodes->Count; i++)
        {
            TNode* node = ((TNode*)Network->Nodes->Items[i]);
    	    for (int j=0; j<node->Fragments->Count; j++)
		    {
				TFragment* fragment = ((TFragment*)node->Fragments->Items[j]);
				fragment->SO_Off();
		    }
        }
    }
}





//---------------------------------------------------------------------------
void __fastcall TChildForm::OptionsFrequencies()
{
    FrequenciesDialog->SelfOrganizationFrequencyScrollBar->Position = SO_FREQUENCY;
    FrequenciesDialog->PartitioningFrequencyScrollBar->Position     = P_FREQUENCY;
    FrequenciesDialog->MigrationFrequencyScrollBar->Position        = M_FREQUENCY;
    FrequenciesDialog->CombiningFrequencyScrollBar->Position        = C_FREQUENCY;
    FrequenciesDialog->ReplicationFrequencyScrollBar->Position      = R_FREQUENCY;
    FrequenciesDialog->DestructionFrequencyScrollBar->Position      = D_FREQUENCY;

    if (FrequenciesDialog->ShowModal())
    {
        SO_FREQUENCY = FrequenciesDialog->SelfOrganizationFrequencyScrollBar->Position;
        P_FREQUENCY  = FrequenciesDialog->PartitioningFrequencyScrollBar->Position;
        M_FREQUENCY  = FrequenciesDialog->MigrationFrequencyScrollBar->Position;
        C_FREQUENCY  = FrequenciesDialog->CombiningFrequencyScrollBar->Position;
        R_FREQUENCY  = FrequenciesDialog->ReplicationFrequencyScrollBar->Position;
        D_FREQUENCY  = FrequenciesDialog->DestructionFrequencyScrollBar->Position;
    }
}


//---------------------------------------------------------------------------
void __fastcall TChildForm::OptionsThresholds()
{
    ThresholdsDialog->PartitioningThresholdScrollBar->Position  = TFragment::PartitioningThreshold;
    ThresholdsDialog->MigrationThresholdScrollBar->Position     = TMigrationFC::MigrationThreshold;
    ThresholdsDialog->CombiningThresholdScrollBar->Position     = TFragment::CombiningThreshold;
    TReplicationFC::ReplicationThreshold = 100.0 - MainForm->ReplicationLevelTrackBar->Position;
    ThresholdsDialog->ReplicationThresholdScrollBar->Position   = TReplicationFC::ReplicationThreshold;
    ThresholdsDialog->DestructionThresholdScrollBar->Position   = TFragment::ReplicaDestructionThreshold;


    if (ThresholdsDialog->ShowModal())
    {
      TFragment::PartitioningThreshold       = ThresholdsDialog->PartitioningThresholdScrollBar->Position;
      TMigrationFC::MigrationThreshold       = ThresholdsDialog->MigrationThresholdScrollBar->Position;
      TFragment::CombiningThreshold          = ThresholdsDialog->CombiningThresholdScrollBar->Position;
      TReplicationFC::ReplicationThreshold   = ThresholdsDialog->ReplicationThresholdScrollBar->Position;
      MainForm->ReplicationLevelTrackBar->Position = 100.0 - TReplicationFC::ReplicationThreshold;  MainForm->ReplicationLevelTrackBarChange(NULL);
      TFragment::ReplicaDestructionThreshold = ThresholdsDialog->DestructionThresholdScrollBar->Position;
    }
}




//---------------------------------------------------------------------------
void __fastcall TChildForm::OptionsMigrationRules()
{
    TNode* current_node;
    TFragment* current_fragment = NULL;

    if (inside_node_index != -1)
   	{
	  current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
      if (current_node != NULL)
   	  {
		if (inside_fragment_index != -1)
	    {
	      current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
        }
      }
    }

    if (current_fragment == NULL)
   	{
	  current_node = ((TNode*)Network->Nodes->Items[0]);
      if (current_node != NULL)
   	  {
          current_fragment = ((TFragment*)current_node->Fragments->Items[0]);
      }
    }

    if (current_fragment != NULL)
    {
      	MigrationForm->FC = current_fragment->MigrationFC;
        MigrationForm->UpdateDisplay();
        MigrationForm->ShowModal();
    }
}


//---------------------------------------------------------------------------
void __fastcall TChildForm::OptionsReplicationRules()
{
    TNode* current_node;
    TFragment* current_fragment = NULL;

    if (inside_node_index != -1)
   	{
	  current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
      if (current_node != NULL)
   	  {
		if (inside_fragment_index != -1)
	    {
	      current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
        }
      }
    }

    if (current_fragment == NULL)
   	{
	  current_node = ((TNode*)Network->Nodes->Items[0]);
      if (current_node != NULL)
   	  {
          current_fragment = ((TFragment*)current_node->Fragments->Items[0]);
      }
    }

    if (current_fragment != NULL)
    {
        ReplicationForm->FC = current_fragment->ReplicationFC;
        ReplicationForm->UpdateDisplay();
        ReplicationForm->ShowModal();
    }
}



////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//---------------------------------------------------------------------------
void __fastcall TChildForm::TrackBar1Change(TObject *Sender)
{
		TrackBar1->Hint = "Demand from this Node = " + IntToStr((int)TrackBar1->Position) + "%";
		TrackBar11->Hint= "Update Ratio = " + IntToStr((int)TrackBar11->Position) + "%";
		if (TrackBar1->Position == 0) TrackBar11->Position = 0;

        if (inside_node_index != -1)
    	{
			TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
			if (inside_fragment_index != -1)
		    {
		        TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
/*
                for (int c=0; c<current_fragment->Clusters->Count; c++)
                {
        	      TCluster* current_cluster = (TCluster*)current_fragment->Clusters->Items[c];
		          if (current_cluster != NULL)
        		  {
		        	if (Network->Nodes->Count > 0) { current_cluster->demands[0]		= TrackBar1 ->Position; }
		        	if (Network->Nodes->Count > 0) { current_cluster->update_ratios[0]	= TrackBar11->Position; }
                    current_cluster->update_demand();
                    current_cluster->Select(selection_color);
                  }
                }
                Repaint();
*/
///*
				if (inside_cluster_index != -1)
				{
			        TCluster* current_cluster = ((TCluster*)current_fragment->Clusters->Items[inside_cluster_index]);
		        	if (Network->Nodes->Count > 0) { current_cluster->demands[0]		= TrackBar1 ->Position; }
		        	if (Network->Nodes->Count > 0) { current_cluster->update_ratios[0]	= TrackBar11->Position; }
                    current_cluster->update_demand();
                    current_cluster->Select(selection_color);
                    Repaint();
            	}
//*/
            }
        }
}
//---------------------------------------------------------------------------
void __fastcall TChildForm::TrackBar2Change(TObject *Sender)
{
		TrackBar2->Hint = "Demand from this Node = " + IntToStr((int)TrackBar2->Position) + "%";
		TrackBar12->Hint= "Update Ratio = " + IntToStr((int)TrackBar12->Position) + "%";
		if (TrackBar2->Position == 0) TrackBar12->Position = 0;

		if (inside_node_index != -1)
    	{
			TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
			if (inside_fragment_index != -1)
		    {
		        TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
				if (inside_cluster_index != -1)
				{
			        TCluster* current_cluster = ((TCluster*)current_fragment->Clusters->Items[inside_cluster_index]);
		        	if (Network->Nodes->Count > 1) { current_cluster->demands[1]		= TrackBar2 ->Position; }
		        	if (Network->Nodes->Count > 1) { current_cluster->update_ratios[1]	= TrackBar12->Position; }
                    current_cluster->update_demand();
                    current_cluster->Select(selection_color);
                    Repaint();
            	}
            }
        }
}
//---------------------------------------------------------------------------
void __fastcall TChildForm::TrackBar3Change(TObject *Sender)
{
		TrackBar3->Hint = "Demand from this Node = " + IntToStr((int)TrackBar3->Position) + "%";
		TrackBar13->Hint= "Update Ratio = " + IntToStr((int)TrackBar13->Position) + "%";
		if (TrackBar3->Position == 0) TrackBar13->Position = 0;

		if (inside_node_index != -1)
    	{
			TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
			if (inside_fragment_index != -1)
		    {
		        TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
				if (inside_cluster_index != -1)
				{
			        TCluster* current_cluster = ((TCluster*)current_fragment->Clusters->Items[inside_cluster_index]);
		        	if (Network->Nodes->Count > 2) { current_cluster->demands[2]		= TrackBar3 ->Position; }
		        	if (Network->Nodes->Count > 2) { current_cluster->update_ratios[2]	= TrackBar13->Position; }
                    current_cluster->update_demand();
                    current_cluster->Select(selection_color);
                    Repaint();
            	}
            }
        }
}
//---------------------------------------------------------------------------
void __fastcall TChildForm::TrackBar4Change(TObject *Sender)
{
		TrackBar4->Hint = "Demand from this Node = " + IntToStr((int)TrackBar4->Position) + "%";
		TrackBar14->Hint= "Update Ratio = " + IntToStr((int)TrackBar14->Position) + "%";
		if (TrackBar4->Position == 0) TrackBar14->Position = 0;

		if (inside_node_index != -1)
    	{
			TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
			if (inside_fragment_index != -1)
		    {
		        TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
				if (inside_cluster_index != -1)
				{
			        TCluster* current_cluster = ((TCluster*)current_fragment->Clusters->Items[inside_cluster_index]);
		        	if (Network->Nodes->Count > 3) { current_cluster->demands[3]		= TrackBar4 ->Position; }
		        	if (Network->Nodes->Count > 3) { current_cluster->update_ratios[3]	= TrackBar14->Position; }
                    current_cluster->update_demand();
                    current_cluster->Select(selection_color);
                    Repaint();
            	}
            }
        }
}
//---------------------------------------------------------------------------
void __fastcall TChildForm::TrackBar5Change(TObject *Sender)
{
		TrackBar5->Hint = "Demand from this Node = " + IntToStr((int)TrackBar5->Position) + "%";
		TrackBar15->Hint= "Update Ratio = " + IntToStr((int)TrackBar15->Position) + "%";
		if (TrackBar5->Position == 0) TrackBar15->Position = 0;

		if (inside_node_index != -1)
    	{
			TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
			if (inside_fragment_index != -1)
		    {
		        TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
				if (inside_cluster_index != -1)
				{
			        TCluster* current_cluster = ((TCluster*)current_fragment->Clusters->Items[inside_cluster_index]);
		        	if (Network->Nodes->Count > 4) { current_cluster->demands[4]		= TrackBar5 ->Position; }
		        	if (Network->Nodes->Count > 4) { current_cluster->update_ratios[4]	= TrackBar15->Position; }
                    current_cluster->update_demand();
                    current_cluster->Select(selection_color);
                    Repaint();
            	}
            }
        }
}
//---------------------------------------------------------------------------
void __fastcall TChildForm::TrackBar6Change(TObject *Sender)
{
		TrackBar6->Hint = "Demand from this Node = " + IntToStr((int)TrackBar6->Position) + "%";
		TrackBar16->Hint= "Update Ratio = " + IntToStr((int)TrackBar16->Position) + "%";
		if (TrackBar6->Position == 0) TrackBar16->Position = 0;

		if (inside_node_index != -1)
    	{
			TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
			if (inside_fragment_index != -1)
		    {
		        TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
				if (inside_cluster_index != -1)
				{
			        TCluster* current_cluster = ((TCluster*)current_fragment->Clusters->Items[inside_cluster_index]);
		        	if (Network->Nodes->Count > 5) { current_cluster->demands[5]		= TrackBar6 ->Position; }
		        	if (Network->Nodes->Count > 5) { current_cluster->update_ratios[5]	= TrackBar16->Position; }
                    current_cluster->update_demand();
                    current_cluster->Select(selection_color);
                    Repaint();
            	}
            }
        }
}
//---------------------------------------------------------------------------
void __fastcall TChildForm::TrackBar7Change(TObject *Sender)
{
		TrackBar7->Hint = "Demand from this Node = " + IntToStr((int)TrackBar6->Position) + "%";
		TrackBar17->Hint= "Update Ratio = " + IntToStr((int)TrackBar17->Position) + "%";
		if (TrackBar7->Position == 0) TrackBar17->Position = 0;

		if (inside_node_index != -1)
    	{
			TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
			if (inside_fragment_index != -1)
		    {
		        TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
				if (inside_cluster_index != -1)
				{
			        TCluster* current_cluster = ((TCluster*)current_fragment->Clusters->Items[inside_cluster_index]);
		        	if (Network->Nodes->Count > 6) { current_cluster->demands[6]		= TrackBar7 ->Position; }
		        	if (Network->Nodes->Count > 6) { current_cluster->update_ratios[6]	= TrackBar17->Position; }
                    current_cluster->update_demand();
                    current_cluster->Select(selection_color);
                    Repaint();
            	}
            }
        }
}
//---------------------------------------------------------------------------
void __fastcall TChildForm::TrackBar8Change(TObject *Sender)
{
		TrackBar8->Hint = "Demand from this Node = " + IntToStr((int)TrackBar8->Position) + "%";
		TrackBar18->Hint= "Update Ratio = " + IntToStr((int)TrackBar18->Position) + "%";
		if (TrackBar8->Position == 0) TrackBar18->Position = 0;

		if (inside_node_index != -1)
    	{
			TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
			if (inside_fragment_index != -1)
		    {
		        TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
				if (inside_cluster_index != -1)
				{
			        TCluster* current_cluster = ((TCluster*)current_fragment->Clusters->Items[inside_cluster_index]);
		        	if (Network->Nodes->Count > 7) { current_cluster->demands[7]		= TrackBar8 ->Position; }
		        	if (Network->Nodes->Count > 7) { current_cluster->update_ratios[7]	= TrackBar18->Position; }
                    current_cluster->update_demand();
                    current_cluster->Select(selection_color);
                    Repaint();
            	}
            }
        }
}
//---------------------------------------------------------------------------
void __fastcall TChildForm::TrackBar9Change(TObject *Sender)
{
		TrackBar9->Hint = "Demand from this Node = " + IntToStr((int)TrackBar9->Position) + "%";
		TrackBar19->Hint= "Update Ratio = " + IntToStr((int)TrackBar19->Position) + "%";
		if (TrackBar9->Position == 0) TrackBar19->Position = 0;

		if (inside_node_index != -1)
    	{
			TNode* current_node = ((TNode*)Network->Nodes->Items[inside_node_index]);
			if (inside_fragment_index != -1)
		    {
		        TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[inside_fragment_index]);
				if (inside_cluster_index != -1)
				{
			        TCluster* current_cluster = ((TCluster*)current_fragment->Clusters->Items[inside_cluster_index]);
		        	if (Network->Nodes->Count > 8) { current_cluster->demands[8]		= TrackBar9 ->Position; }
		        	if (Network->Nodes->Count > 8) { current_cluster->update_ratios[8]	= TrackBar19->Position; }
                    current_cluster->update_demand();
                    current_cluster->Select(selection_color);
                    Repaint();
            	}
            }
        }
}
//---------------------------------------------------------------------------








//---------------------------------------------------------------------------










/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

void __fastcall TChildForm::MainTimerTimer(TObject *Sender)
{

 if (running)
 {

    if (((running_counter+1) % MAX_SAMPLES) == 0) // stop
    {
      MainForm->RunStartItemClick(Sender);    // Toggle
      MainForm->RunStartSpeedButton->Down = false;
    }




    if (Network != NULL)
    {
      if(!Network->IsEmpty())
      {

        // random demands from all nodes on the selected fargment's clusters
        if (auto_demands)
        {
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// changing-demands single_fragment
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
           if (single_fragment)     // 1 Fragment, 6 Nodes, 4 Demand Values -> 4^(1*6) = 4096
           {
            if (Network->get_selected_node_index() != -1)
            {
              TNode* node = ((TNode*)Network->Nodes->Items[Network->get_selected_node_index()]);
              if (node->get_selected_fragment_index() != -1)
              {
                TFragment* fragment = ((TFragment*)node->Fragments->Items[node->get_selected_fragment_index()]);
if(!same_demands_on_clusters)
{
                for (int n=0; n<Network->Nodes->Count; n++) // demanding nodes
                {
                  if (running_counter == 0) // full random initially
                  {
                    if ( !(random(random(Network->Nodes->Count))) ) // demand from only SOME nodes
                    {
                      for (int k=0; k<fragment->Clusters->Count; k++) // different demanding nodes on diffrent clusters
                      {
                        TCluster* cluster = ((TCluster*)fragment->Clusters->Items[k]);
                        cluster->demands[n] = 25 + random(75);
    		        	cluster->update_ratios[n] = MainForm->UpdateRatioTrackBar->Position + random(10);
                        if (cluster->update_ratios[n] > 100) cluster->update_ratios[n] = 100;
                        cluster->update_demand();
                      }
                    }
                  }
                  else // stochastic random subsequently  -+ delta %
                  {
                      for (int k=0; k<fragment->Clusters->Count; k++) // different demanding nodes on diffrent clusters
                      {
                        TCluster* cluster = ((TCluster*)fragment->Clusters->Items[k]);
                        if (cluster->demands[n] > 0)
                        {
                          //                      -1 | 0 | +1     1..9
                          //cluster->demands[n] += (random(3)-1) * (1+random(9));
                          cluster->demands[n] += ( (random(100)>42) ? +1 : -1 ) * 10;
                        }
                        if (cluster->demands[n] > 100) cluster->demands[n] = 100;
                        if (cluster->demands[n] <   0) { cluster->demands[n] = 0; cluster->update_ratios[n] = 0; }
                        cluster->update_demand();
                      }
                  }
                }
}
else // if(same_demands_on_clusters)
{
                for (int n=0; n<Network->Nodes->Count; n++) // demanding nodes
                {
                  int same_initial_demand = 25 + random(75);
                  int same_initial_update_ratio = MainForm->UpdateRatioTrackBar->Position + random(10);
                  if (same_initial_update_ratio > 100) same_initial_update_ratio = 100;
                  //int same_delta  = (random(3)-1) * (1+random(9));
                  int same_delta  = ( (random(100)>42) ? +1 : -1 ) * 10;

                  if (running_counter == 0) // full random initially
                  {
                    if ( !(random(random(Network->Nodes->Count))) ) // demand from only SOME nodes
                    {
                      for (int k=0; k<fragment->Clusters->Count; k++) // different demanding nodes on diffrent clusters
                      {
                        TCluster* cluster = ((TCluster*)fragment->Clusters->Items[k]);
                        cluster->demands[n] = same_initial_demand;
    		        	cluster->update_ratios[n] = same_initial_update_ratio;
                        cluster->update_demand();
                      }
                    }
                  }
                  else // stochastic random subsequently  -+ delta %
                  {
                      for (int k=0; k<fragment->Clusters->Count; k++) // different demanding nodes on diffrent clusters
                      {
                        TCluster* cluster = ((TCluster*)fragment->Clusters->Items[k]);
                        if (cluster->demands[n] > 0)
                        {
                          //                      -1 | 0 | +1     1..9
                          cluster->demands[n] += same_delta;
                        }
                        if (cluster->demands[n] > 100) cluster->demands[n] = 100;
                        if (cluster->demands[n] <   0) { cluster->demands[n] = 0; cluster->update_ratios[n] = 0; }
                        cluster->update_demand();
                      }
                  }
                }
}
              }
            }
           }
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// changing-demands multi-fragment
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
           //random demands from all nodes on all fragment's clusters
           else //if (multi_fragment)   // 18 Fragment, 6 Nodes, 4 Demand Values -> 4^(18*6) = 1.05e+65
           {
            for (int i=0; i<Network->Nodes->Count; i++)
            {
              TNode* node = ((TNode*)Network->Nodes->Items[i]);
              for (int j=0; j<node->Fragments->Count; j++)
              {
               TFragment* fragment = ((TFragment*)node->Fragments->Items[j]);
               if (!fragment->IsReplica())
               {
                 ////////////////////////////////////////////////////////////////
                 // undo demand distribution
                 ////////////////////////////////////////////////////////////////
                 // bring demands back from replicas and zero replicas' demands
                 ////////////////////////////////////////////////////////////////
                 for (int i=0; i<fragment->Replicas->Count; i++)
                 {
                  TFragment* replica = (TFragment*)fragment->Replicas->Items[i];
            	  if (replica != NULL)
            	  {
                   for (int c=0; c<replica->Clusters->Count; c++)
                   {
                    TCluster* replica_cluster = (TCluster*)replica->Clusters->Items[c];
                    TCluster* original_cluster = (TCluster*)fragment->Clusters->Items[c];
            		if ( (original_cluster != NULL) && (replica_cluster != NULL) )
            		{
                      for (int n=0; n<Network->Nodes->Count; n++)
                      {
                         //TNode* node_n = (TNode*)Network->Nodes->Items[n];
                         if (replica_cluster->demands[n] > 0)
                         {
                            original_cluster->demands[n] = replica_cluster->demands[n];
                            original_cluster->update_ratios[n] = replica_cluster->update_ratios[n];
                            replica_cluster->demands[n] = 0;
                            replica_cluster->update_ratios[n] = 0;
                         }
                      }
                    }
                   }
                   replica->update_demand();
                  }
                 }
                 ///////////////////////////////////////////////////////////////
if(!same_demands_on_clusters)
{
                for (int n=0; n<Network->Nodes->Count; n++) // demanding nodes
                {
                  if (running_counter == 0) // full random initially
                  {
                    if ( !(random(random(Network->Nodes->Count))) ) // demand from only SOME nodes
                    {
                      for (int k=0; k<fragment->Clusters->Count; k++) // different demanding nodes on diffrent clusters
                      {
                        TCluster* cluster = ((TCluster*)fragment->Clusters->Items[k]);
                        cluster->demands[n] = 25 + random(75);
    		        	cluster->update_ratios[n] = MainForm->UpdateRatioTrackBar->Position + random(10);
                        if (cluster->update_ratios[n] > 100) cluster->update_ratios[n] = 100;
                        cluster->update_demand();
                     }
                    }
                  }
                  else // stochastic random subsequently  -+ delta %
                  {
                      for (int k=0; k<fragment->Clusters->Count; k++) // different demanding nodes on diffrent clusters
                      {
                        TCluster* cluster = ((TCluster*)fragment->Clusters->Items[k]);
                        if (cluster->demands[n] > 0)
                        {
                          //                      -1 | 0 | +1     1..9
                          //cluster->demands[n] += (random(3)-1) * (1+random(9));
                          cluster->demands[n] += ( (random(100)>42) ? +1 : -1 ) * 10;
                        }
                        if (cluster->demands[n] > 100) cluster->demands[n] = 100;
                        if (cluster->demands[n] <   0) { cluster->demands[n] = 0; cluster->update_ratios[i] = 0; }

                        cluster->update_demand();
                      }
                  }
                }
}
else // if(same_demands_on_clusters)
{
                for (int n=0; n<Network->Nodes->Count; n++) // demanding nodes
                {
                  int same_initial_demand = 25 + random(75);
                  int same_initial_update_ratio = MainForm->UpdateRatioTrackBar->Position + random(10);
                  if (same_initial_update_ratio > 100) same_initial_update_ratio = 100;
                  //int same_delta  = (random(3)-1) * (1+random(9));
                  int same_delta  = ( (random(100)>42) ? +1 : -1 ) * 10;

                  if (running_counter == 0) // full random initially
                  {
                    if ( !(random(random(Network->Nodes->Count))) ) // demand from only SOME nodes
                    {
                      for (int k=0; k<fragment->Clusters->Count; k++) // different demanding nodes on diffrent clusters
                      {
                        TCluster* cluster = ((TCluster*)fragment->Clusters->Items[k]);
                        cluster->demands[n] = same_initial_demand;
    		        	cluster->update_ratios[n] = same_initial_update_ratio;
                        cluster->update_demand();
                      }
                    }
                  }
                  else // stochastic random subsequently  -+ delta %
                  {
                      for (int k=0; k<fragment->Clusters->Count; k++) // different demanding nodes on diffrent clusters
                      {
                        TCluster* cluster = ((TCluster*)fragment->Clusters->Items[k]);
                        if (cluster->demands[n] > 0)
                        {
                          //                      -1 | 0 | +1     1..9
                          cluster->demands[n] += same_delta;
                        }
                        if (cluster->demands[n] > 100) cluster->demands[n] = 100;
                        if (cluster->demands[n] <   0) { cluster->demands[n] = 0; cluster->update_ratios[n] = 0; }
                        cluster->update_demand();
                      }
                  }
                }
}
               }
               // distribute demands from orignal to replicas nearest to demands
               if ( (fragment->Replicas->Count > 0) && (!fragment->IsReplica()) ) // if orignal ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
               {
                  for (int c=0; c<fragment->Clusters->Count; c++)
                  {
                    TCluster* original_cluster = (TCluster*)fragment->Clusters->Items[c];
            		if (original_cluster != NULL)
            		{
                      for (int n=0; n<Network->Nodes->Count; n++)
                      {
                         TNode* node_n = (TNode*)Network->Nodes->Items[n];
                         TFragment* replica_n = fragment->find_nearest_replica_to_node(node_n); // including this
                         if ( (replica_n != fragment) && (original_cluster->demands[n] != 0) )
                         {
                            TCluster* replica_cluster = (TCluster*)replica_n->Clusters->Items[c];
                    		if ( (original_cluster != NULL) && (replica_cluster != NULL) )

                            replica_cluster->demands[n] = original_cluster->demands[n];
                            replica_cluster->update_ratios[n] = original_cluster->update_ratios[n];
                            original_cluster->demands[n] = 0;
                            original_cluster->update_ratios[n] = 0;
                         }
                         replica_n->update_demand();
                      }
                    }
                  }
               }

              }
            }
           }
        }
        // fixed demands from all nodes on the selected fargment's clusters
        else //if (!auto_demands)
        {
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// fixed-demands single_fragment
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
           if (single_fragment)     // 1 Fragment, 6 Nodes, 4 Demand Values -> 4^(1*6) = 4096
           {
            if (Network->get_selected_node_index() != -1)
            {
              TNode* node = ((TNode*)Network->Nodes->Items[Network->get_selected_node_index()]);
              if (node->get_selected_fragment_index() != -1)
              {
                //TFragment* fragment = ((TFragment*)node->Fragments->Items[node->get_selected_fragment_index()]);
                //
                // Enter Demands Manually
                //
              }
            }
           }

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// fixed-demands multi-fragment
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
           //random demands from all nodes on all fragment's clusters
           else //if (multi_fragment)   // 18 Fragment, 6 Nodes, 4 Demand Values -> 4^(18*6) = 1.05e+65
           {
            for (int i=0; i<Network->Nodes->Count; i++)
            {
              TNode* node = ((TNode*)Network->Nodes->Items[i]);
              for (int j=0; j<node->Fragments->Count; j++)
              {
               TFragment* fragment = ((TFragment*)node->Fragments->Items[j]);
               if (!fragment->IsReplica())
               {
                 ////////////////////////////////////////////////////////////////
                 // undo demand distribution
                 ////////////////////////////////////////////////////////////////
                 // bring demands back from replicas and zero replicas' demands
                 ////////////////////////////////////////////////////////////////
                 for (int i=0; i<fragment->Replicas->Count; i++)
                 {
                  TFragment* replica = (TFragment*)fragment->Replicas->Items[i];
            	  if (replica != NULL)
            	  {
                   for (int c=0; c<replica->Clusters->Count; c++)
                   {
                    TCluster* replica_cluster = (TCluster*)replica->Clusters->Items[c];
                    TCluster* original_cluster = (TCluster*)fragment->Clusters->Items[c];
            		if ( (original_cluster != NULL) && (replica_cluster != NULL) )
            		{
                      for (int n=0; n<Network->Nodes->Count; n++)
                      {
                         //TNode* node_n = (TNode*)Network->Nodes->Items[n];
                         if (replica_cluster->demands[n] > 0)
                         {
                            original_cluster->demands[n] = replica_cluster->demands[n];
                            original_cluster->update_ratios[n] = replica_cluster->update_ratios[n];
                            replica_cluster->demands[n] = 0;
                            replica_cluster->update_ratios[n] = 0;
                         }
                      }
                    }
                   }
                   replica->update_demand();
                  }
                 }
                 ///////////////////////////////////////////////////////////////

if(!same_demands_on_clusters)
{
                for (int n=0; n<Network->Nodes->Count; n++) // demanding nodes
                {
                  if (running_counter == 0) // full random initially
                  {
                    if ( !(random(random(Network->Nodes->Count))) ) // demand from only SOME nodes
                    {
                      for (int k=0; k<fragment->Clusters->Count; k++) // different demanding nodes on diffrent clusters
                      {
                        TCluster* cluster = ((TCluster*)fragment->Clusters->Items[k]);
                        cluster->demands[n] = 25 + random(75);
    		        	cluster->update_ratios[n] = MainForm->UpdateRatioTrackBar->Position + random(10);
                        if (cluster->update_ratios[n] > 100) cluster->update_ratios[n] = 100;
                        cluster->update_demand();
                      }
                    }
                  }
                  else // stochastic random subsequently  -+ delta %
                  {
                      for (int k=0; k<fragment->Clusters->Count; k++) // different demanding nodes on diffrent clusters
                      {
                        TCluster* cluster = ((TCluster*)fragment->Clusters->Items[k]);
                        if (cluster->demands[n] > 0)
                        {
                          //                      -1 | 0 | +1     1..9
                          //cluster->demands[n] += (random(3)-1) * (1+random(9));
                        }
                        if (cluster->demands[n] > 100) cluster->demands[n] = 100;
                        if (cluster->demands[n] <   0) { cluster->demands[n] = 0; cluster->update_ratios[n] = 0; }
                        cluster->update_demand();
                      }
                  }
                }
}
else // if(same_demands_on_clusters)
{
                for (int n=0; n<Network->Nodes->Count; n++) // demanding nodes
                {
                  int same_initial_demand = 25 + random(75);
                  int same_initial_update_ratio = MainForm->UpdateRatioTrackBar->Position + random(10);
                  if (same_initial_update_ratio > 100) same_initial_update_ratio = 100;
                  int same_delta  = 0;//(random(3)-1) * (1+random(9));

                  if (running_counter == 0) // full random initially
                  {
                    if ( !(random(random(Network->Nodes->Count))) ) // demand from only SOME nodes
                    {
                      for (int k=0; k<fragment->Clusters->Count; k++) // different demanding nodes on diffrent clusters
                      {
                        TCluster* cluster = ((TCluster*)fragment->Clusters->Items[k]);
                        cluster->demands[n] = same_initial_demand;
    		        	cluster->update_ratios[n] = same_initial_update_ratio;
                        cluster->update_demand();
                      }
                    }
                  }
                  else // stochastic random subsequently  -+ delta %
                  {
                      for (int k=0; k<fragment->Clusters->Count; k++) // different demanding nodes on diffrent clusters
                      {
                        TCluster* cluster = ((TCluster*)fragment->Clusters->Items[k]);
                        if (cluster->demands[n] > 0)
                        {
                          //                      -1 | 0 | +1     1..9
                          cluster->demands[n] += same_delta;
                        }
                        if (cluster->demands[n] > 100) cluster->demands[n] = 100;
                        if (cluster->demands[n] <   0) { cluster->demands[n] = 0; cluster->update_ratios[n] = 0; }
                        cluster->update_demand();
                      }
                  }
                }
}
               }
               // distribute demands from orignal to replicas nearest to demands
               if ( (fragment->Replicas->Count > 0) && (!fragment->IsReplica()) ) // if orignal ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
               {
                  for (int c=0; c<fragment->Clusters->Count; c++)
                  {
                    TCluster* original_cluster = (TCluster*)fragment->Clusters->Items[c];
            		if (original_cluster != NULL)
            		{
                      for (int n=0; n<Network->Nodes->Count; n++)
                      {
                         TNode* node_n = (TNode*)Network->Nodes->Items[n];
                         TFragment* replica_n = fragment->find_nearest_replica_to_node(node_n); // including this
                         if ( (replica_n != fragment) && (original_cluster->demands[n] != 0) )
                         {
                            TCluster* replica_cluster = (TCluster*)replica_n->Clusters->Items[c];
                    		if ( (original_cluster != NULL) && (replica_cluster != NULL) )

                            replica_cluster->demands[n] = original_cluster->demands[n];
                            replica_cluster->update_ratios[n] = original_cluster->update_ratios[n];
                            original_cluster->demands[n] = 0;
                            original_cluster->update_ratios[n] = 0;
                         }
                         replica_n->update_demand();
                      }
                    }
                  }
               }

              }
            }
           }
        }








////////////////////////////////////////////////////////////////////////////////
//   Self-Organize
////////////////////////////////////////////////////////////////////////////////

        if ( (SO_FREQUENCY > 0) && ((running_counter%(MAX_SAMPLES/SO_FREQUENCY)) == 0) )
        {
          //////////////////////////////////////////////
          if ( P_FREQUENCY == C_FREQUENCY )
          {
            TFragment::partition_or_combine = ! TFragment::partition_or_combine;
          }
          //////////////////////////////////////////////

          if (MainForm->soSpeedButton->Down)
          {
           // reset self-organization
           so_counter = 0;
            p_counter = 0;
            m_counter = 0;
            c_counter = 0;
            r_counter = 0;
            d_counter = 0;
           so_activity[running_counter/(MAX_SAMPLES/SO_FREQUENCY)] = 0;

           if (single_fragment)
           {
            if (Network->get_selected_node_index() != -1)
            {
              TNode* selected_node = ((TNode*)Network->Nodes->Items[Network->get_selected_node_index()]);
              if (selected_node->get_selected_fragment_index() != -1)
              {
                TFragment* selected_fragment = ((TFragment*)selected_node->Fragments->Items[selected_node->get_selected_fragment_index()]);

                // self-organize selected fragment
				int so_target_node_id = selected_fragment->Self_Organize();
				if (so_target_node_id != INVALID_ID)
                {
                    TNode* so_target_node       = Network->GetNode(so_target_node_id);
				    int    so_target_node_index = Network->Nodes->IndexOf(so_target_node);

                    // update network
				    Network->set_selected_node_index(so_target_node_index);

                    // reset source node
                    selected_node->DeSelect();
                    selected_node->set_selected_fragment_index(-1);

                    // update target node
                    so_target_node->Select(selection_color);
                    so_target_node->set_selected_fragment_index(so_target_node->Fragments->Count-1);

                    TFragment* selected_fragment = ((TFragment*)so_target_node->Fragments->Items[so_target_node->get_selected_fragment_index()]);
                    selected_fragment->Select(selection_color);

                    TCluster* selected_cluster = ((TCluster*)selected_fragment->Clusters->Items[0]);
                    selected_cluster->Select(selection_color);

                    // update flags
                    inside_node_index     = so_target_node_index;
                    inside_fragment_index = so_target_node->get_selected_fragment_index();
                }

                // update node utilization
                //for(int i=0; i<Network->Nodes->Count; i++)
                {
                  selected_node->update_utilization();
                }
              }
            }
           }
           else //if (multi_fragment)
           {
            // self-organize all fragments
            for(int i=Network->Nodes->Count-1; i>=0; i--)
            {
    		  Network->DeSelectChildren();
        	  inside_node_index		= -1;
        	  inside_link_index		= -1;
        	  inside_fragment_index	= -1;
        	  inside_cluster_index	= -1;

              TNode* current_node = ((TNode*)Network->Nodes->Items[i]);
   	       	  for (int j=0; j<current_node->Fragments->Count; j++)
	          {
    			TFragment* current_fragment = ((TFragment*)current_node->Fragments->Items[j]);
				int so_target_node_id = current_fragment->Self_Organize();

				if (so_target_node_id != INVALID_ID)
                {
                    // do nothing
                    // can't track multi-fargments
                }
	          }
            }
            // update node utilization
            for(int i=0; i<Network->Nodes->Count; i++)
            {
              TNode* current_node = ((TNode*)Network->Nodes->Items[i]);
              current_node->update_utilization();
            }
           }
          }




          int fragments = 0;
          for(int i=0; i<Network->Nodes->Count; i++)
          {
              TNode* current_node = ((TNode*)Network->Nodes->Items[i]);
   	       	  fragments += current_node->Fragments->Count;
          }

////?????????????????????????????????????????????????????????????????????///////
////?????????????????????????????????????????????????????????????????????///////
////?????????????????????????????????????????????????????????????????????///////
          //so_counter = p_counter + m_counter + c_counter + r_counter + d_counter;
////?????????????????????????????????????????????????????????????????????///////
////?????????????????????????????????????????????????????????????????????///////
////?????????????????????????????????????????????????????????????????????///////
          if (fragments > 0)
          {
            so_activity[running_counter/(MAX_SAMPLES/SO_FREQUENCY)] = (so_counter * 100) / fragments;
          }
          else
          {
            so_activity[running_counter/(MAX_SAMPLES/SO_FREQUENCY)] = 0;
          }

        } //end   if ((running_counter%(MAX_SAMPLES/SO_FREQUENCY)) == 0)


////////////////////////////////////////////////////////////////////////////////






////////////////////////////////////////////////////////////////////////////////
// calculate response time
////////////////////////////////////////////////////////////////////////////////
		if (Network != NULL)
        {
            float total_response = 0;
            int demanded_fragments = 0;
            int all_fragments = 0;
            for (int i=0; i<Network->Nodes->Count; i++)
            {
              TNode* node = ((TNode*)Network->Nodes->Items[i]);
              for (int j=0; j<node->Fragments->Count; j++)
              {
                TFragment* fragment = ((TFragment*)node->Fragments->Items[j]);
                //
                fragment->update_response();
                //
                all_fragments++;
                if (fragment->get_demand() > 0)
                {
                    demanded_fragments++;
                    total_response += fragment->response;
                }
                Network->minResponse = (fragment->response < Network->minResponse) ? fragment->response : Network->minResponse;
                Network->maxResponse = (fragment->response > Network->maxResponse) ? fragment->response : Network->maxResponse;
              }
            }
//
            demanded_fragments = all_fragments;
//
            float average_response = 0;
            if (demanded_fragments > 0)
            {
              average_response = total_response / demanded_fragments;
            }
            //if (average_response > 100) average_response = 100;
            //if (average_response <   0) average_response =   0;

            int* ResponseTime = new int(average_response);
            Network->ResponseTimeHistory->Add(ResponseTime);
        }
////////////////////////////////////////////////////////////////////////////////








////////////////////////////////////////////////////////////////////////////////
// calculate resource usage
////////////////////////////////////////////////////////////////////////////////
        if (Network != NULL)
        {
            if(!Network->IsEmpty())
            {
	            int  NodesResourceUsage = 0;
                for (int i=0; i<Network->Nodes->Count; i++)
                {
                    TNode* current_node = ((TNode*)Network->Nodes->Items[i]);
                    NodesResourceUsage += current_node->get_utilization();
        		}
                NodesResourceUsage /= Network->Nodes->Count;

	            int  LinksResourceUsage = 0;
                for (int i=0; i<Network->Links->Count; i++)
                {
                    TLink* current_link = ((TLink*)Network->Links->Items[i]);
                    LinksResourceUsage += current_link->get_utilization();
		        }
                LinksResourceUsage /= Network->Links->Count;

	            int* ResourceUsage = new int(NodesResourceUsage*NODES_RESOURCE_USAGE_FACTOR + LinksResourceUsage*LINKS_RESOURCE_USAGE_FACTOR);

                if ((running_counter%(MAX_SAMPLES/SO_FREQUENCY)) == 0)
                {
                     *ResourceUsage += so_activity[running_counter/(MAX_SAMPLES/SO_FREQUENCY)] / 2.0; // resource for so actions
                }
                //if (*ResourceUsage > 100) *ResourceUsage = 100;
                //if (*ResourceUsage <   0) *ResourceUsage =   0;

                Network->minResource = (*ResourceUsage < Network->minResource) ? *ResourceUsage : Network->minResource;
                Network->maxResource = (*ResourceUsage > Network->maxResource) ? *ResourceUsage : Network->maxResource;
    	        Network->ResourceUsageHistory->Add(ResourceUsage);
            }
        }
////////////////////////////////////////////////////////////////////////////////////////////////




		MainForm->UpdateMenuItems(NULL);
		NetworkPaintBox->Repaint();
		FragmentsPaintBox->Repaint();
		ClustersPaintBox->Repaint();
		ResponsePaintBox->Repaint();
		ResourcePaintBox->Repaint();



      }
    }


    running_counter++;

 } //end if (running)

}
//---------------------------------------------------------------------------

/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/


void __fastcall TChildForm::FormKeyPress(TObject *Sender, char &Key)
{
    switch (Key)
    {
        case 'P', 'p' : { MainForm->PartitioningSpeedButtonClick(Sender); MainForm->PartitioningSpeedButton->Down = !MainForm->PartitioningSpeedButton->Down; } break;
        case 'M', 'm' : { MainForm->MigrationSpeedButtonClick(Sender);    MainForm->MigrationSpeedButton->Down    = !MainForm->MigrationSpeedButton->Down;    } break;
        case 'C', 'c' : { MainForm->CombiningSpeedButtonClick(Sender);    MainForm->CombiningSpeedButton->Down    = !MainForm->CombiningSpeedButton->Down;    } break;
        case 'R', 'r' : { MainForm->ReplicationSpeedButtonClick(Sender);  MainForm->ReplicationSpeedButton->Down  = !MainForm->ReplicationSpeedButton->Down;  } break;
        case 'D', 'd' : { MainForm->DestructionSpeedButtonClick(Sender);  MainForm->DestructionSpeedButton->Down  = !MainForm->DestructionSpeedButton->Down;  } break;
    }
}
//---------------------------------------------------------------------------

