//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <math.h>
#pragma hdrstop

#include "functions.h"
#include "TCluster.h"
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
#include "TNetwork.h"
#include "TNode.h"
#include "TFragment.h"
//---------------------------------------------------------------------------



int TCluster::scale = 6;



TCluster::TCluster(TCluster* _cluster)
{
  if (_cluster != NULL)
  {
	Fragment  = _cluster->Fragment;

	set_id(_cluster->get_id());
	set_position(_cluster->get_position());
	set_size(_cluster->get_size());

	for (int i=0; i<MAX_NODES; i++) demands[i]       = _cluster->demands[i];
	for (int i=0; i<MAX_NODES; i++) update_ratios[i] = _cluster->update_ratios[i];
	update_ratio  = _cluster->update_ratio;
	update_demand();

	selected = _cluster->IsSelected();
  }
}


TCluster::TCluster(TFragment* _Fragment)
{
  if (_Fragment != NULL)
  {
	Fragment	= _Fragment;

	id  = _Fragment->get_id()*10 + _Fragment->cluster_serial_num++;
//	set_size(20+random(80));
	set_size(100);

	for (int i=0; i<MAX_NODES; i++) demands[i]       = 0;
	for (int i=0; i<MAX_NODES; i++) update_ratios[i] = 0;
	update_ratio  = 0;


//	set_size(100);
//	for (int i=0; i<MAX_NODES; i++) demands[i]       = 100;

	update_demand();

	DeSelect();
  }
}



TCluster::~TCluster()
{
	DeSelect();
	Fragment	= NULL;
}



void TCluster::set_size(int _size)
{
	size	= _size;
    radius	= size/scale;
    Fragment->update_size();
}




void TCluster::update_demand()
{
    int total_demand = 0;
    for (int i=0; i<Fragment->Node->Network->Nodes->Count; i++)
    {
// size only needed in traffic = f(demand, size)
//        total_demand += ((demands[i] * size ) / 100.0);  // demand freqency on the whole of a sized-cluster
        total_demand += demands[i];
    }
    demand = total_demand / MAX_FULLY_DEMANDING_NODES;

    int total_update = 0;
    for (int i=0; i<Fragment->Node->Network->Nodes->Count; i++)
    {
// size only needed in traffic = f(demand, size)
//        total_update += ((demands[i] * size ) / 100.0) * (update_ratios[i] / 100.0);
        total_update += (demands[i] * update_ratios[i]) / 100.0;
    }
    int update = total_update / MAX_FULLY_DEMANDING_NODES;


	update_ratio = 0;
    if (demand > 0)
    {
    	update_ratio = (update  * 100.0) / demand;
    }


    //if (demand <   0) demand =   0;
    //if (demand > 100) demand = 100;
	if (demand < 50)	//  0 .. 50
	{
        color	= (TColor) RGB(
								255,
       	                   	    255,
                       	        255 - (demand*2)*2.55
    						  );
	}
    else	//  50 .. 100
	{
         color	= (TColor) RGB(
       						    ((100-demand)*2)*2.55,
       						    ((100-demand)*2)*2.55,
 	    	        	        0
    						  );
	}

	if (!selected)
    {
       	selection_color = color;
    }

	Fragment->update_demand();
}



bool TCluster::similar_in_demand_patterns_to(TCluster* _cluster) // for fragment partitioning
{
        //      pt = P * exp(-1 * logx(n,2)/4)      n=0..16//

////////////////////////////////////////////////////////////////////////////////
//    double DifferenceThreshold = TFragment::PartitioningThreshold;
    double DifferenceThreshold = TFragment::PartitioningThreshold * exp( -1 * logx(Fragment->Node->get_num_of_processors(), 2)/16.0 );
////////////////////////////////////////////////////////////////////////////////

    //
    //      |  |  |
    //      |  |  |        |        cluster 1
    //      |  |  |  |  |  |
    //
    //
    //      |           |           cluster 2
    //      |           |
    //      |     |  |  |  |
    //
    //
    //        +  +     -  +
    //     00 99 66 00 66 33       difference
    //


    float difference;
    float max_positive_difference = 0.0;
    float min_negative_difference = 0.0;

	for (int i=0; i<Fragment->Node->Network->Nodes->Count; i++)
    {
        difference = demands[i] - _cluster->demands[i];
        if ( (difference > 0) && (difference > max_positive_difference) )
        {
            max_positive_difference = difference;
        }
        else if ( (difference < 0) && (difference < min_negative_difference) )
        {
            min_negative_difference = difference;
        }
        else // if ( difference == 0 )
        {
            // do nothing
        }
    }


    if ( (max_positive_difference >= DifferenceThreshold) || (-min_negative_difference >= DifferenceThreshold) )
    {
        return false;  // different
    }
    return true;       // similar
}





void TCluster::Draw(TCanvas* _canvas)
{
		if (selected)
        {
    		_canvas->Pen->Color	= selection_color;
        }
        else
        {
    		_canvas->Pen->Color		= color;
        }
		_canvas->Brush->Color	= color;

		_canvas->Ellipse(position.x-radius, position.y-radius, position.x+radius, position.y+radius);

		if (selected)
        {
			_canvas->Font->Color = selection_color;
        }
        else
        {
   	   	  if (demand > 50)
          {
			_canvas->Font->Color = clWhite;
   	      }
		  else
   	      {
			_canvas->Font->Color = clBlack;
   	      }
        }

		_canvas->Font->Name = "NewCareer";
		_canvas->Font->Size  = 8;
   		int left = position.x - _canvas->Font->Size * 1.0;
   		int top  = position.y - _canvas->Font->Size * 0.8;
   		_canvas->TextOut(left, top, IntToStr((int)id));

		_canvas->Pen->Color = clInfoBk;
		_canvas->Brush->Color = clInfoBk;
		_canvas->Rectangle(0, 0, 800, 15);

        _canvas->Refresh();
}



void TCluster::DrawCaption(TCanvas* _canvas)
{
		_canvas->Pen->Color = clInfoBk;
		_canvas->Brush->Color = clInfoBk;
		_canvas->Rectangle(0, 0, 800, 15);

		_canvas->Font->Name  = "MS Sans Serif";
		_canvas->Font->Size  = 8;
		_canvas->Font->Color = clOlive;
		_canvas->Brush->Color = clInfoBk;
        //_canvas->TextOut(20, 0, IntToStr((int)Fragment->get_id()));
        //_canvas->TextOut(60, 0, IntToStr((int)id));
   		_canvas->TextOut(15, 0, IntToStr((int)size));
   		_canvas->TextOut(65, 0, IntToStr((int)demand));
   		_canvas->TextOut(130, 0, IntToStr((int)update_ratio));

	    _canvas->Refresh();
}


bool TCluster::IsSelected()
{
   	return selected;
}




void TCluster::Select(TColor _selection_color)
{
	Fragment->set_selected_cluster_index(Fragment->Clusters->IndexOf(this));
   	selection_color = _selection_color;
	selected = true;
}



void TCluster::DeSelect()
{
	if (selected)
    {
		selected = false;
	   	selection_color = color;
		Fragment->set_selected_cluster_index(-1);
    }
}



