//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop

#include "functions.h"
#include "TFragment.h"
//---------------------------------------------------------------------------


#include "TNode.h"
#include "TMigrationFC.h"
#include "TReplicationFC.h"

#include "TMainForm.h"  // MainForm->((TChildForm*)ActiveMDIChild)->
#include "TChildForm.h" // P_FREQUENCY M_FREQUENCY C_FREQUENCY R_FREQUENCY D_FREQUENCY
//---------------------------------------------------------------------------




int TFragment::scale 	= 2;

bool   TFragment::partition_or_combine = false;
double TFragment::PartitioningThreshold         = 75.0;     // minmum disimilarity level to justfy partitioning
double TFragment::CombiningThreshold            = 50.0;     // minmum similarity level to justfy combining
double TFragment::ReplicaDestructionThreshold   =  5.0;     // minmum demand level below which destruct replica



// copy constructor
TFragment::TFragment(TFragment* _fragment)
{
  if (_fragment != NULL)
  {
	cluster_serial_num = _fragment->cluster_serial_num;
    partitioned_fragment_id_serial_num = _fragment->partitioned_fragment_id_serial_num;

    Node			= _fragment->Node;
    //MigrationFC 	= new TMigrationFC(this);       // replicas start thinking independently
    //ReplicationFC = new TReplicationFC(this);     // replicas start thinking independently
    MigrationFC 	= NULL;       // replicas are dummy
    ReplicationFC 	= NULL;     // replicas are dummy

	Clusters		= new TList();
   	for (int i=0; i<_fragment->Clusters->Count; i++)
    {
        TCluster* copyCluster = new TCluster((TCluster*)_fragment->Clusters->Items[i]);
		if (copyCluster != NULL)
		{
          copyCluster->Fragment = this;
		  Clusters->Add(copyCluster);
        }
    }

	set_id(_fragment->get_id());
  	set_position(_fragment->get_position());
   	update_size();


	Replicas  = new TList();
   	for (int i=0; i<_fragment->Replicas->Count; i++)
    {
        TFragment* replica = (TFragment*)_fragment->Replicas->Items[i];
		if (replica != NULL)
		{
		  Replicas->Add(replica);
          replica->Replicas->Add(this);
        }
    }


	for (int i=0; i<MAX_NODES; i++) demands[i] = _fragment->demands[i];
	for (int i=0; i<MAX_NODES; i++) total_demands[i] = _fragment->total_demands[i];
	for (int i=0; i<MAX_NODES; i++) update_ratios[i] = _fragment->update_ratios[i];
	update_ratio  = _fragment->update_ratio;

    DemandLevelTable = new TList();
   	for (int i=0; i<_fragment->DemandLevelTable->Count; i++)
    {
      TDemandLevel* new_DemandLevel = new TDemandLevel(*((TDemandLevel*)_fragment->DemandLevelTable->Items[i]));
  	  if (new_DemandLevel != NULL)
      {
	    DemandLevelTable->Add(new_DemandLevel);
      }
    }

    DemandHistoryTable = new TList();
   	for (int i=0; i<_fragment->DemandHistoryTable->Count; i++)
    {
      TDemandEntry* new_DemandEntry = new TDemandEntry(*((TDemandEntry*)_fragment->DemandHistoryTable->Items[i]));
  	  if (new_DemandEntry != NULL)
      {
	    DemandHistoryTable->Add(new_DemandEntry);
      }
    }

	selected = _fragment->IsSelected();
	set_replica(_fragment->IsReplica());

	SO_Off();

/*
    ////////////////////////////////////////////////////////////////////////////
   	for (int i=0; i<_fragment->Clusters->Count; i++)
    {
        TCluster* cluster = (TCluster*)Clusters->Items[i];
        cluster->update_demand();
    }
    this->update_demand();    // also updates collective responses[]
    ////////////////////////////////////////////////////////////////////////////
*/
  }
}




TFragment::TFragment(TNode* _Node)
{
  if (_Node != NULL)
  {
	cluster_serial_num = 0;
    partitioned_fragment_id_serial_num = 0;

    Node			= _Node;
    MigrationFC 	= new TMigrationFC(this);
    ReplicationFC 	= new TReplicationFC(this);
	Clusters		= new TList();
	Replicas        = new TList();
   	DemandLevelTable = new TList();
   	DemandHistoryTable = new TList();

	id  = _Node->get_id()*10 + _Node->fragment_serial_num++;
	selected = false;

    int NewClusters = 5;//3 + random(8);
   	for (int i=0; i<NewClusters; i++)
    {
		TCluster* Cluster = new TCluster(this);
	    AddCluster(Cluster);
    }
	DeSelect();
	DeSelectChildren();

   	update_size();

	for (int i=0; i<MAX_NODES; i++) demands[i]       = 0;
	for (int i=0; i<MAX_NODES; i++) update_ratios[i] = 0;
	for (int i=0; i<MAX_NODES; i++) total_demands[i] = 0;

	update_ratio  = 0;

    update_demand();

	set_replica(false);
	SO_Off();
  }
}




TFragment::~TFragment()
{
    // notify other replicas of self-deletion
	for(int i=0; i<Replicas->Count; i++)
    {
        TFragment* replica = (TFragment*)Replicas->Items[i];
        replica->Replicas->Remove(this);
    }

    // if the self-deleted fragment is an orignal, then one of its replicas become an orignal
    if ( (Replicas->Count > 0) &&  (!IsReplica()) )
    {
        TFragment* original = (TFragment*)Replicas->Items[0];
        original->set_replica(false);
    }


    // un-distribute own demands back to original before self-deletion
    if ( IsReplica() && (Replicas->Count > 0) )
    {
      // zero_th replica is chosen as an original
      TFragment* original_fragment = (TFragment*)Replicas->Items[0];
      if (original_fragment != NULL)
      {
        for (int c=0; c<Clusters->Count; c++)
        {
          TCluster* replica_cluster = (TCluster*)Clusters->Items[c];
          TCluster* original_cluster = (TCluster*)original_fragment->Clusters->Items[c];

          if ( (original_cluster != NULL) && (replica_cluster != NULL) )
          {
            for (int n=0; n<Node->Network->Nodes->Count; n++)
            {
              //TNode* node_n = (TNode*)Network->Nodes->Items[n];
              if (replica_cluster->demands[n] != 0)
              {
                original_cluster->demands[n] = replica_cluster->demands[n];
                original_cluster->update_ratios[n] = replica_cluster->update_ratios[n];
                replica_cluster->demands[n] = 0;
                replica_cluster->update_ratios[n] = 0;
              }
            }
          }
        }
        original_fragment->update_demand();
      }
    }



	Clusters->Clear();
	delete	Clusters;

	DemandLevelTable->Clear();
	delete	DemandLevelTable;

	DemandHistoryTable->Clear();
	delete	DemandHistoryTable;


    Node	= NULL;

    if (!IsReplica()) // only if original
    {
    	Replicas->Clear();
        delete Replicas;

        delete MigrationFC;
        delete ReplicationFC;
    }
}




TPoint TFragment::get_position()
{
	Node->calculate_area_lefttop();

   	TPoint position_;
    position_.x = position.x + Node->area_left;
    position_.y = position.y + Node->area_top;

    return position_;
}



void TFragment::set_position(TPoint _position)
{
	Node->calculate_area_lefttop();

    position.x = _position.x - Node->area_left;
    position.y = _position.y - Node->area_top;
}




int TFragment::update_size()
{
        int total_size = 0;
        for (int i=0; i<Clusters->Count; i++)
        {
	      TCluster* current_cluster = (TCluster*)Clusters->Items[i];
		  if (current_cluster != NULL)
		  {
		    total_size += current_cluster->get_size();
          }
        }

	   	size = total_size / MAX_FULL_SIZE_CLUSTERS_PER_FRAGMENT;

	    radius	= pow(size, .75)/scale;

        return size;
}




float TFragment::calculate_mobility()
{
    float mobility;
    //mobility = f(size, no of class relationships);
    mobility = 100 * (exp( -1 * ((double)size) / 100.0 ) );  // = 100 ... 38
    return mobility;
}




float TFragment::calculate_availability()
{
    int   demanding_nodes = 0;
    for (int n=0; n<Node->Network->Nodes->Count; n++)
    {
        if (demands[n] > 0)
        {
            demanding_nodes++;
        }
    }

    float availability = 0.0;      // = Replicas->Count / Demanding Nodes
    if (demanding_nodes > 0)
    {
        availability = (Replicas->Count * 100.0) / demanding_nodes;
    }
    return availability;
}




TFragment* TFragment::find_nearest_replica_to_node(TNode* _node)
{
    int demanding_node_index  = Node->Network->Nodes->IndexOf(_node);
    int this_node_index       = Node->Network->Nodes->IndexOf(this->Node);

    int        minimum_delay   = Node->Network->delays[this_node_index][demanding_node_index];
    TFragment* nearest_replica = this;
    for (int i=0; i<Replicas->Count; i++)
    {
      TFragment* replica = (TFragment*)Replicas->Items[i];
      int index = Node->Network->Nodes->IndexOf(replica->Node);

      int delay = Node->Network->delays[index][demanding_node_index];
      if (delay < minimum_delay)
      {
         minimum_delay = delay;
         nearest_replica = replica;
      }
    }

    return nearest_replica;
}



int TFragment::update_demand()
{
    // distribute demands from orignal to replicas nearest to demands
    if ( (Replicas->Count > 0) && (!IsReplica()) ) // if orignal
    {
      for (int c=0; c<Clusters->Count; c++)
      {
        TCluster* cluster = (TCluster*)Clusters->Items[c];
		if (cluster != NULL)
		{
          for (int n=0; n<Node->Network->Nodes->Count; n++)
          {
             TNode* node_n = (TNode*)Node->Network->Nodes->Items[n];
             TFragment* replica_n = find_nearest_replica_to_node(node_n); // including this
             if(replica_n != NULL)
             {
               TCluster* replica_n_cluster = (TCluster*)replica_n->Clusters->Items[c];
               if ( (replica_n_cluster != NULL) && (replica_n != this) && (cluster->demands[n] != 0) )
               {
                 replica_n_cluster->demands[n] = cluster->demands[n];
                 replica_n_cluster->update_ratios[n] = cluster->update_ratios[n];
                 cluster->demands[n] = 0;
                 cluster->update_ratios[n] = 0;
               }
             }
          }
        }
      }
    }


    for (int n=0; n<Node->Network->Nodes->Count; n++)
    {
        total_demands[n] = 0;
        for (int c=0; c<Clusters->Count; c++)
        {
	      TCluster* current_cluster = (TCluster*)Clusters->Items[c];
		  if (current_cluster != NULL)
		  {
            if (current_cluster->demands[n] != 0)
            {
		        total_demands[n] += current_cluster->demands[n];
            }
          }
        }

        demands[n] = 0;
        if (Clusters->Count > 0)
        {
            demands[n] = total_demands[n] / Clusters->Count;
        }
	}

    int total_demand = 0;
    for (int n=0; n<Node->Network->Nodes->Count; n++)
    {
        if (demands[n] != 0)
        {
            total_demand += demands[n];
        }
    }
    demand = total_demand / MAX_FULLY_DEMANDING_NODES;



    for (int n=0; n<Node->Network->Nodes->Count; n++)
    {
        total_updates[n] = 0;
        for (int c=0; c<Clusters->Count; c++)
        {
	      TCluster* current_cluster = (TCluster*)Clusters->Items[c];
		  if (current_cluster != NULL)
		  {
            if (current_cluster->demands[n] != 0)
            {
		        total_updates[n] += (current_cluster->demands[n] * current_cluster->update_ratios[n]) / 100.0;
            }
          }
        }

        updates[n] = 0;
        if (Clusters->Count > 0)
        {
            updates[n] = total_updates[n] / Clusters->Count;
        }

        if (demands[n] > 0)
        {
            update_ratios[n] = (updates[n] * 100.0) / demands[n];
        }
	}

    int total_update = 0;
    for (int n=0; n<Node->Network->Nodes->Count; n++)
    {
        if (updates[n] != 0)
        {
            total_update += updates[n];
        }
    }
    update = total_update / MAX_FULLY_DEMANDING_NODES;


	update_ratio = 0;
   	if (demand > 0)
    {
        update_ratio = (update * 100.0) / demand;
    }



/*
    //if (demand <   0) demand =   0;
    //if (demand > 100) demand = 100;
	if (demand < 50)	//  0 .. 50
	{
        color	= (TColor) RGB(
								255 - (demand*2)*2.55,
                       	        255,
       	                   	    255
    						  );
	}
    else	//  50 .. 100
	{
        color	= (TColor) RGB(
 	    	        	        0,
    							((100-demand)*2)*2.55,
 	    	        	        ((100-demand)*2)*2.55
    						  );
	}

   	if (!selected)
    {
       	selection_color = color;
    }
*/



	update_DemandLevelTable();
    update_response();

    Node->update_utilization();
    Node->Network->update_links_utilization();

    return demand;
}



////////////////////////////////////////////////////////////////////////////////////
void TFragment::update_DemandLevelTable()
{
   	DemandLevelTable->Clear();
    for (int n=0; n<Node->Network->Nodes->Count; n++)
    {
        TDemandLevel* new_DemandLevel = new TDemandLevel();
		new_DemandLevel->NodeId			=	((TNode*)(Node->Network->Nodes->Items[n]))->get_id();
		new_DemandLevel->Demand			=	demands[n];
		new_DemandLevel->UpdateRatio	=	update_ratios[n];
	    if (new_DemandLevel->Demand > 0.0)
        {
        	DemandLevelTable->Add(new_DemandLevel);
        }
    }
}
////////////////////////////////////////////////////////////////////////////////////




////////////////////////////////////////////////////////////////////////////////////
void TFragment::update_response()
{
    int to = -1;
    for (int n=0; n<Node->Network->Nodes->Count; n++)
    {
      TNode* node = ((TNode*)Node->Network->Nodes->Items[n]);
      if (Node->get_id() == node->get_id() )
      {
            to = n;
            break;
      }
    }


    for (int from=0; from<Node->Network->Nodes->Count; from++)
    {
//      responses[from] = (demands[from]/100.0) * Node->Network->delays[from][to];
///*
      //responses[from] = retrieval + update
      double response_of_retrieval_demand =   ((demands[from]/100.0)*((100.0 - update_ratios[from])/100.0)) * 1.0 * Node->Network->delays[from][to];
                                                                                                         // delay0
      // response_of_update_demand = replica-synchronization of replica along slowest path, cos in synch is done in parallel
      double response_of_update_demand    =   ((demands[from]/100.0)*(update_ratios[from]/100.0)) * 1.0 * Node->Network->delays[from][to]                                                       // delay0
          + ((Replicas->Count > 0) ? 1 : 0) * ((demands[from]/100.0)*(update_ratios[from]/100.0)) * 3.0 * Node->Network->delays[from][to] * (MainForm->ReplicaSynchronizationTrackBar->Position/100.0);   // delay of slowest path from original (say 3.0 for WAN) * up-to-dateness level

      responses[from] = response_of_retrieval_demand + response_of_update_demand;
//*/
    }


/*
    int slowest_response = 0;   // maximum delay
    for (int n=0; n<Node->Network->Nodes->Count; n++)
    {
       if (slowest_response < responses[n])
       {
         slowest_response = responses[n];
       }
    }

    response = slowest_response;
*/

    //int demanding_nodes = 0;
    int total_response = 0;
    for (int n=0; n<Node->Network->Nodes->Count; n++)
    {
      if (responses[n] > 0)
      {
        //demanding_nodes++;
        total_response += responses[n];
      }
    }

    int average_response = 0;
//    if (demanding_nodes > 0)
    if (Node->Network->Nodes->Count > 0)
    {
//        average_response = total_response / demanding_nodes;
        average_response = total_response / Node->Network->Nodes->Count;
    }

    response = average_response;
    //if (response <   0) response =   0;
    //if (response > 100) response = 100;
	if (response < 50)	//  0 .. 50
	{
        color	= (TColor) RGB(
								255 - (response*2)*2.55,
                       	        255,
       	                   	    255
    						  );
	}
    else	//  50 .. 100
	{
        color	= (TColor) RGB(
 	    	        	        0,
    							((100-response)*2)*2.55,
 	    	        	        ((100-response)*2)*2.55
    						  );
	}

   	if (!selected)
    {
       	selection_color = color;
    }

}
////////////////////////////////////////////////////////////////////////////////////





void TFragment::AddCluster(TCluster* _Cluster)
{
  if(_Cluster != NULL)
  {
    int Columns = 10; //5;
	int xGap    = 35;
    int yGap    = 35;
	int xShift  = 20;
   	int yShift  = 35;

    int Column, Row, x, y;
   	Column	= (Clusters->Count)%Columns;
	Row		= (Clusters->Count)/Columns;
	x	= xShift + xGap*Column;
   	y	= yShift + yGap*Row;


    ////////////////////////////////////////////////////////////////////////////////////////////////////////////
	_Cluster->Fragment = this;	// must do this first to calculate_area_lefttop();
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////
	_Cluster->set_position(Point(x,y));

    Clusters->Add(_Cluster);
	DeSelectChildren();

   	update_size();

    update_demand();
  }
}




void TFragment::DeleteCluster(int _index)
{
   	for (int i=Clusters->Count-1; i>_index; i--)
    {
	    ((TCluster*)Clusters->Items[i])->set_position(((TCluster*)Clusters->Items[i-1])->get_position());
    }

    Clusters->Delete(_index);
	if (_index == selected_cluster_index)
    {
       	selected_cluster_index = -1;
    }

   	update_size();

    update_demand();
}








bool TFragment::similar_in_demand_patterns_to(TFragment* _fragment) // for fragment combining
{
        //      ct = C + (100-C)*(1 - exp(-1 * logx(n,2)/4)     n=0..16 //

////////////////////////////////////////////////////////////////////////////////
//    double DifferenceThreshold = 100.0 - TFragment::CombiningThreshold;
    double DifferenceThreshold = 100.0 - ( TFragment::CombiningThreshold + (100.0 - TFragment::CombiningThreshold) * (1.0 - exp(-1 * logx(Node->get_num_of_processors(),2)/16.0)) );      //
////////////////////////////////////////////////////////////////////////////////

    //
    //      |  |  |
    //      |  |  |        |        fragment 1
    //      |  |  |  |  |  |
    //
    //
    //      |           |           fragment 2
    //      |           |
    //      |     |  |  |  |
    //
    //
    //        +  +     -  +
    //     00 99 66 00 66 33       differences
    //
    //     99                      difference = largest absolute difference
    //

    float difference;
    float max_positive_difference = 0.0;
    float min_negative_difference = 0.0;

	for (int i=0; i<Node->Network->Nodes->Count; i++)
    {
        difference = demands[i] - _fragment->demands[i];
        if ( (difference > 0) && (difference > max_positive_difference) )
        {
            max_positive_difference = difference;
        }
        else if ( (difference < 0) && (difference < min_negative_difference) )
        {
            min_negative_difference = difference;
        }
        else // if ( difference == 0 )
        {
            // do nothing
        }
    }


    if ( (max_positive_difference >= DifferenceThreshold) || (-min_negative_difference >= DifferenceThreshold) )
    {
        return false;  // different
    }
    return true;       // similar
}




void TFragment::SO_On    ()
{
	SO = true;
}



void TFragment::SO_Off   ()
{
	SO = false;
}









/*

int TFragment::Self_Organize()
{
 if (SO)
 {
  TChildForm* ChildForm = ((TChildForm*)MainForm->ActiveMDIChild);


  int        migration_node_id = INVALID_ID;
  int        replication_node_ids[MAX_NODES]; for (int i=0; i<MAX_NODES; i++) replication_node_ids[i] = INVALID_ID;
          // replicas are non-SO fragments belonging to this original Fragment




  // disable alternating if frequencies are not the same, and enable both the same
  if (ChildForm->P_FREQUENCY != ChildForm->C_FREQUENCY)
  {
    TFragment::partition_or_combine = true;
  }
  if (TFragment::partition_or_combine)
  {
   if ( ChildForm->p_enabled && (ChildForm->SO_FREQUENCY > 0)  &&  (ChildForm->P_FREQUENCY > 0)  && ((ChildForm->running_counter%(ChildForm->SO_FREQUENCY/ChildForm->P_FREQUENCY)) == 0) )
   {
    //bool gradual_partitioning = true;
    bool gradual_partitioning = false;

               ////////////////////////////////////////////////////////////////
               // undo demand distribution
               ////////////////////////////////////////////////////////////////
               // bring demands back from replicas
               ////////////////////////////////////////////////////////////////
               for (int i=0; i<Replicas->Count; i++)
               {
                  TFragment* replica = (TFragment*)Replicas->Items[i];
            	  if (replica != NULL)
            	  {
                   for (int c=0; c<replica->Clusters->Count; c++)
                   {
                    TCluster* replica_cluster = (TCluster*)replica->Clusters->Items[c];
                    TCluster* original_cluster = (TCluster*)Clusters->Items[c];
            		if ( (original_cluster != NULL) && (replica_cluster != NULL) )
            		{
                      for (int n=0; n<Node->Network->Nodes->Count; n++)
                      {
                         //TNode* node_n = (TNode*)Network->Nodes->Items[n];
                         if (replica_cluster->demands[n] > 0)
                         {
                            original_cluster->demands[n] = replica_cluster->demands[n];
                            original_cluster->update_ratios[n] = replica_cluster->update_ratios[n];
                            replica_cluster->demands[n] = 0;
                            replica_cluster->update_ratios[n] = 0;
                         }
                      }
                    }
                   }
                   replica->update_demand();
                  }
               }
               /////////////////////////////////////////////////////////////////

    TFragment* ParentFragment = this;
    TFragment* ChildFragment  = NULL;

    bool done = false;
    while (!done)
    {
      done = true; // unless you prove me wrong

      // get the first cluster, add it to the new fragment, remove from parent fragment
      TCluster* first_cluster = ((TCluster*)ParentFragment->Clusters->Items[0]);

      // find dissimilar clusters in the remaining clusters
      // and remove them from parent fragment, add to the new fragment,
      for (int i=1; i<ParentFragment->Clusters->Count; i++)
	  {
        TCluster* cluster = ((TCluster*)ParentFragment->Clusters->Items[i]);
        // PartitioningFC->Run();
        if ( ! first_cluster->similar_in_demand_patterns_to(cluster) )
        {
///////////////////////
//ChildForm->so_counter++;
///////////////////////

               /////////////////////////////////////////////////////////////////
               // delete replicas
               /////////////////////////////////////////////////////////////////
               while (Replicas->Count > 0)
               {
                  TFragment* replica = (TFragment*)Replicas->Items[Replicas->Count - 1];
            	  if (replica != NULL)
            	  {
                    replica->Node->RemoveFragment(replica);
                    Replicas->Remove(replica);
                    //delete replica;
                  }
               }
               //Replicas->Clear();
               /////////////////////////////////////////////////////////////////


            // create a new empty fragment to hole removed clusters
            if (ChildFragment == NULL)
            {
              // Partition();
              ChildFragment = new TFragment(Node);
              ChildFragment->set_id(this->get_id()*10 + ++partitioned_fragment_id_serial_num);
              ChildFragment->Clusters->Clear();
    		  Node->AddFragment(ChildFragment);
              ChildFragment->SO_On();
            }

            ParentFragment->DeleteCluster(i);
            ChildFragment->AddCluster(cluster);

            i--; // shift up
            done = false;
        }
        ParentFragment->update_demand(); // redistribute demand back to empty replicas
      }

      if (gradual_partitioning)
      {
        break;
      }

      // partition new child
      ParentFragment = ChildFragment;
      ChildFragment  = NULL;
    }
   }
  }







//  if ( (ChildForm->running_counter > 0) && (ChildForm->SO_FREQUENCY > 0)  &&  (ChildForm->M_FREQUENCY > 0)  && ((ChildForm->running_counter%(ChildForm->SO_FREQUENCY/ChildForm->M_FREQUENCY)) == 0) )
  if ( ChildForm->m_enabled && (ChildForm->SO_FREQUENCY > 0)  &&  (ChildForm->M_FREQUENCY > 0)  && ((ChildForm->running_counter%(ChildForm->SO_FREQUENCY/ChildForm->M_FREQUENCY)) == 0) )
  {
    try
    {
        migration_node_id  = MigrationFC->Run();
    }
    catch (...)
    {
        // slience errors
    }

    if (migration_node_id != INVALID_ID)
	{
	  TNode* migration_node = Node->Network->GetNode(migration_node_id);
      if (migration_node != NULL)
      {
        Migrate(migration_node);
///////////////////////
ChildForm->so_counter++;
///////////////////////
      }
    }
  }












  // disable alternating if frequencies are not the same, and enable both
  if (ChildForm->P_FREQUENCY != ChildForm->C_FREQUENCY)
  {
    TFragment::partition_or_combine = false;
  }
  if ( ! TFragment::partition_or_combine)
  {
   if ( ChildForm->c_enabled && (ChildForm->SO_FREQUENCY > 0)  &&  (ChildForm->C_FREQUENCY > 0)  && ((ChildForm->running_counter%(ChildForm->SO_FREQUENCY/ChildForm->C_FREQUENCY)) == 0) )
   {
    //bool gradual_combining = true;
    bool gradual_combining = false;

               ////////////////////////////////////////////////////////////////
               // undo demand distribution
               ////////////////////////////////////////////////////////////////
               // bring demands back from replicas
               ////////////////////////////////////////////////////////////////
               for (int i=0; i<Replicas->Count; i++)
               {
                  TFragment* replica = (TFragment*)Replicas->Items[i];
            	  if (replica != NULL)
            	  {
                   for (int c=0; c<replica->Clusters->Count; c++)
                   {
                    TCluster* replica_cluster = (TCluster*)replica->Clusters->Items[c];
                    TCluster* original_cluster = (TCluster*)Clusters->Items[c];
            		if ( (original_cluster != NULL) && (replica_cluster != NULL) )
            		{
                      for (int n=0; n<Node->Network->Nodes->Count; n++)
                      {
                         //TNode* node_n = (TNode*)Network->Nodes->Items[n];
                         if (replica_cluster->demands[n] > 0)
                         {
                            original_cluster->demands[n] = replica_cluster->demands[n];
                            original_cluster->update_ratios[n] = replica_cluster->update_ratios[n];
                            replica_cluster->demands[n] = 0;
                            replica_cluster->update_ratios[n] = 0;
                         }
                      }
                    }
                   }
                   replica->update_demand();
                  }
               }
               ////////////////////////////////////////////////////////////////


    TFragment* ThisFragment = this;
    TFragment* CombineWithFragment = NULL;

    for (int i=0; i<ThisFragment->Node->Fragments->Count; i++)
    {
      if (CombineWithFragment == NULL)
      {
        TFragment* CombineWithFragment  = (TFragment*)ThisFragment->Node->Fragments->Items[i];
        if (CombineWithFragment == ThisFragment) continue;  // skip

               ////////////////////////////////////////////////////////////////
               // undo demand distribution of CombineWithFragment
               ////////////////////////////////////////////////////////////////
               // bring demands back from replicas
               ////////////////////////////////////////////////////////////////
               for (int i=0; i<CombineWithFragment->Replicas->Count; i++)
               {
                  TFragment* replica = (TFragment*)CombineWithFragment->Replicas->Items[i];
            	  if (replica != NULL)
            	  {
                   for (int c=0; c<replica->Clusters->Count; c++)
                   {
                    TCluster* replica_cluster = (TCluster*)replica->Clusters->Items[c];
                    TCluster* original_cluster = (TCluster*)CombineWithFragment->Clusters->Items[c];
            		if ( (original_cluster != NULL) && (replica_cluster != NULL) )
            		{
                      for (int n=0; n<Node->Network->Nodes->Count; n++)
                      {
                         //TNode* node_n = (TNode*)Network->Nodes->Items[n];
                         if (replica_cluster->demands[n] > 0)
                         {
                            original_cluster->demands[n] = replica_cluster->demands[n];
                            original_cluster->update_ratios[n] = replica_cluster->update_ratios[n];
                            replica_cluster->demands[n] = 0;
                            replica_cluster->update_ratios[n] = 0;
                         }
                      }
                    }
                   }
                   replica->update_demand();
                  }
               }
               ////////////////////////////////////////////////////////////////


        // CombiningFC->Run();
        if ( CombineWithFragment->similar_in_demand_patterns_to(ThisFragment) )
        {
///////////////////////
//ChildForm->so_counter++;
///////////////////////


               /////////////////////////////////////////////////////////////////
               // delete replicas of ThisFragment
               /////////////////////////////////////////////////////////////////
               while (Replicas->Count > 0)
               {
                  TFragment* replica = (TFragment*)Replicas->Items[Replicas->Count - 1];
            	  if (replica != NULL)
            	  {
                    replica->Node->RemoveFragment(replica);
                    Replicas->Remove(replica);
                    //delete replica;
                  }
               }
               //Replicas->Clear();
               /////////////////////////////////////////////////////////////////

               /////////////////////////////////////////////////////////////////
               // delete replicas of CombineWithFragment
               /////////////////////////////////////////////////////////////////
               while (CombineWithFragment->Replicas->Count > 0)
               {
                  TFragment* replica = (TFragment*)CombineWithFragment->Replicas->Items[CombineWithFragment->Replicas->Count - 1];
            	  if (replica != NULL)
            	  {
                    replica->Node->RemoveFragment(replica);
                    CombineWithFragment->Replicas->Remove(replica);
                    //delete replica;
                  }
               }
               //CombineWithFragment->Replicas->Clear();
               /////////////////////////////////////////////////////////////////

            // Combine();
            for (int i=0; i<ThisFragment->Clusters->Count; i++)
            {
              TCluster* cluster = ((TCluster*)ThisFragment->Clusters->Items[i]);
              ThisFragment->DeleteCluster(i);
    		  CombineWithFragment->AddCluster(cluster);
              i--;
            }
            CombineWithFragment->set_id(ThisFragment->get_id());
            delete ThisFragment->Node->RemoveFragment(ThisFragment);

            if (gradual_combining)
            {
              break;
            }

            ThisFragment = CombineWithFragment;
            CombineWithFragment = NULL;

            i--; // process the new fragment shifted up in place of the deletion one.
        }
      }
    }
    if (ThisFragment != NULL)
    {
        ThisFragment->update_demand(); // redistribute demand back to empty replicas
    }
    if (CombineWithFragment != NULL)
    {
        CombineWithFragment->update_demand(); // redistribute demand back to empty replicas
    }

   }

  }









  if ( ChildForm->r_enabled && (ChildForm->running_counter > 0) && (ChildForm->SO_FREQUENCY > 0)  &&  (ChildForm->R_FREQUENCY > 0)  && ((ChildForm->running_counter%(ChildForm->SO_FREQUENCY/ChildForm->R_FREQUENCY)) == 0) )
  {
    for (int i=0; i<MAX_NODES; i++)
	{
      replication_node_ids[i] = INVALID_ID;
    }

    try
    {
        ReplicationFC->Run(replication_node_ids);
    }
    catch (...)
    {
        // slience errors
    }


    for (int i=0; i<MAX_NODES; i++)
	{
      if (replication_node_ids[i] != INVALID_ID)
      {
	    TNode* replication_node = Node->Network->GetNode(replication_node_ids[i]);
        if (replication_node != NULL)
        {
          Replicate(replication_node);
///////////////////////
ChildForm->so_counter++;
///////////////////////
        }
      }
    }
  }


  if ( ChildForm->d_enabled && (ChildForm->SO_FREQUENCY > 0)  &&  (ChildForm->D_FREQUENCY > 0)  && ((ChildForm->running_counter%(ChildForm->SO_FREQUENCY/ChildForm->D_FREQUENCY)) == 0) )
  {
    for (int i=0; i<Replicas->Count; i++)
    {
        TFragment* replica = (TFragment*)Replicas->Items[i];
        if (  (replica != NULL) && (replica->demand < TFragment::ReplicaDestructionThreshold)  )
        {
            // destroy low_demanded_replica
            if (  ((TNode*)replica->Node)->RemoveFragment(replica) )
            {
                //Replicas->Remove(replica);
                delete replica; // will also undistribute demand back to original
            }

            i--;    // pack

///////////////////////
//ChildForm->so_counter++;
///////////////////////

        }
    }
  }



  return migration_node_id;
 }

 return -1;
}

*/













void TFragment::Partition(int _fromClusterIndex, int _toClusterIndex)
{
  if (SO)
  {
  	// clusters
    // statistics tables
    //Fragment->update_demand();
  }
}




void TFragment::Migrate  (TNode* _toNode)
{
	if ( (_toNode != NULL) && (this->Node != NULL) )
    {
      // only migrate to other nodes
	  if (((TNode*)_toNode)->get_id() != ((TNode*)this->Node)->get_id())
      {
		// remove original but keep
    	TFragment* Fragment = Node->RemoveFragment(this, false);

		// send
		_toNode->AddFragment(Fragment);
//	    for (long l=0; l<1000000L; l++) ; // sending delay

		// wait for acknowledgement
//    	for (long l=0; l<1000000L; l++) ; // acknowledging delay

		// inform repository
//    	for (long l=0; l<1000000L; l++) ; // informing repository delay

        Fragment->update_demand();
	  }
    }
}





void TFragment::Combine  (TFragment* _withFragment)
{
  if (SO)
  {
  	// clusters
    // statistics tables
    //Fragment->update_demand();
  }
}






void TFragment::Replicate(TNode* _onNode)
{
	if ( (_onNode != NULL) && (this->Node != NULL) )
    {
      // only replicate on other nodes which don't aleardy have a replica of this fragment
      for (int i=0; i<_onNode->Fragments->Count; i++)
      {
        TFragment* aFragment = (TFragment*)_onNode->Fragments->Items[i];
        if (aFragment->get_id() == this->get_id())
        {
            return;
        }
      }


	  //if (((TNode*)_onNode)->get_id() != ((TNode*)this->Node)->get_id())
      {
  	    TFragment* Fragment = new TFragment(this);
        Fragment->SO_Off(); // make dummy

        /*
        // empty statistics as replicas don't have the same demand info
   	    for (int i=0; i<fragment->Clusters->Count; i++)
        {
		  TCluster* cluster = (TCluster*)Fragment->Clusters->Items[i];
          for (int n=0; n<Node->Network->Nodes->Count; n++)
          {
		    cluster->demands[n] = 0;
	        cluster->update_ratios[n] = 0;
          }
          cluster->update_demand();
        }
        */

	    // send
	    _onNode->AddFragment(Fragment);
//      for (long l=0; l<1000000L; l++) ; // sending delay

	    // wait for acknowledgement
//      for (long l=0; l<1000000L; l++) ; // acknowledging delay

	    // inform repository
//      for (long l=0; l<1000000L; l++) ; // informing repository delay


        // set replica flag
        //this->set_replica(true);    // keep "this" as orignal
        Fragment->set_replica(true);

        // let them know of each other
        this->Replicas->Add(Fragment);
        Fragment->Replicas->Add(this);

        Fragment->update_demand();
      }
	}
}




void TFragment::DestroyReplica  (TFragment* _replica)
{
  if (SO)
  {
  	// clusters
    // statistics tables
    //Fragment->update_demand();
  }
}









void TFragment::Draw(TCanvas* _canvas)
{
		if (selected)
        {
    		_canvas->Pen->Color	= selection_color;
        }
        else
        {
    		_canvas->Pen->Color	= color;
        }
		_canvas->Brush->Color	= color;


//    	if (replica)
//        {
//			_canvas->Pen->Width = 2;
//			_canvas->Pen->Color = clSilver;
//        }

        int x = get_position().x;	// modifies x,y to
        int y = get_position().y;	//                 screen location
		_canvas->Ellipse(x-radius, y-radius, x+radius, y+radius);




		if (selected)
        {
			_canvas->Font->Color = selection_color;
        }
        else
        {
   	   	  if (replica)
          {
			_canvas->Font->Color = clGray;
       		//_canvas->Font->Style = _canvas->Font->Style << fsBold;
   	      }
          else
          {
            //if (demand > 66)
       	    if (response > 66.0)
            {
			  _canvas->Font->Color = clSilver;
            }
            else
            {
			  _canvas->Font->Color = clBlack;
            }
          }
        }



		_canvas->Font->Name = "NewCareer";
		_canvas->Font->Size  = 10;
   		int left = x - (_canvas->Font->Size * 6) / 10 ;
   		int top  = y - (_canvas->Font->Size * 7) / 10;
   		_canvas->TextOut(left, top, IntToStr((int)id));

		_canvas->Pen->Width = 1;
   		_canvas->Font->Style = _canvas->Font->Style >> fsBold;
        _canvas->Refresh();
}



void TFragment::DrawCaption(TCanvas* _canvas)
{
		_canvas->Pen->Color = clInfoBk;
		_canvas->Brush->Color = clInfoBk;
		_canvas->Rectangle(0, 0, 800, 15);

		_canvas->Font->Name  = "MS Sans Serif";
		_canvas->Font->Size  = 8;
		_canvas->Font->Color = clTeal;
		_canvas->Brush->Color = clInfoBk;
        //_canvas->TextOut(180, 0, IntToStr((int)Node->get_id()));
        //_canvas->TextOut(210, 0, IntToStr((int)id));
   		_canvas->TextOut(140, 0, IntToStr((int)size));
   		_canvas->TextOut(185, 0, IntToStr((int)demand));
   		_canvas->TextOut(255, 0, IntToStr((int)update_ratio));
   		_canvas->TextOut(335, 0, IntToStr((int)response));
   		_canvas->TextOut(405, 0, IntToStr((int)Replicas->Count));


        // put debug info at Node 9 place
	    update_demand();

		_canvas->Pen->Color = clWhite;
		_canvas->Brush->Color = clWhite;
		_canvas->Rectangle(400, 300, 600, 500);

		_canvas->Font->Color = clBlack;
		_canvas->Brush->Color = clWhite;
   		_canvas->TextOut(400, 320, "Node  Demand  UpdateRatio");
		_canvas->Font->Color = clTeal;
		_canvas->Brush->Color = clWhite;
        for (int c=0; c<DemandLevelTable->Count; c++)
        {
       		_canvas->TextOut(410, 340+15*c, IntToStr((int)((TDemandLevel*)(DemandLevelTable->Items[c]))->NodeId));
   	    	_canvas->TextOut(450, 340+15*c, IntToStr((int)((TDemandLevel*)(DemandLevelTable->Items[c]))->Demand));
   		    _canvas->TextOut(500, 340+15*c, IntToStr((int)((TDemandLevel*)(DemandLevelTable->Items[c]))->UpdateRatio));
        }

	    _canvas->Refresh();
}



void TFragment::DrawDetails(TCanvas* _canvas)
{
   	for (int i=0; i<Clusters->Count; i++)
    {
		((TCluster*)(Clusters->Items[i]))->Draw(_canvas);
    }
}




bool TFragment::IsSelected()
{
   	return selected;
}





void TFragment::Select(TColor _selection_color)
{
	Node->set_selected_fragment_index(Node->Fragments->IndexOf(this));
   	selection_color = _selection_color;
	selected = true;
}



void TFragment::DeSelect()
{
	if (selected)
    {
		selected = false;
   		selection_color = color;
		DeSelectChildren();
		Node->set_selected_fragment_index(-1);
    }
}



void TFragment::DeSelectChildren()
{
   	for (int i=0; i<Clusters->Count; i++)
    {
 		((TCluster*)(Clusters->Items[i]))->DeSelect();
    }
   	selected_cluster_index = -1;
}




bool TFragment::AnySelection()
{
    return (selected_cluster_index != -1);
}









///*

int TFragment::Self_Organize()
{
 if (SO)
 {
  TChildForm* ChildForm = ((TChildForm*)MainForm->ActiveMDIChild);


  int        migration_node_id = INVALID_ID;
  int        replication_node_ids[MAX_NODES]; for (int i=0; i<MAX_NODES; i++) replication_node_ids[i] = INVALID_ID;
          // replicas are non-SO fragments belonging to this original Fragment




  // disable alternating if frequencies are not the same, and enable both the same
  if (ChildForm->P_FREQUENCY != ChildForm->C_FREQUENCY)
  {
    TFragment::partition_or_combine = true;
  }
  if (TFragment::partition_or_combine)
  {
   if ( ChildForm->p_enabled && (Replicas->Count == 0) && (ChildForm->SO_FREQUENCY > 0)  &&  (ChildForm->P_FREQUENCY > 0)  && ((ChildForm->running_counter%(ChildForm->SO_FREQUENCY/ChildForm->P_FREQUENCY)) == 0) )
   {
    //bool gradual_partitioning = true;
    bool gradual_partitioning = false;
    TFragment* ParentFragment = this;
    TFragment* ChildFragment  = NULL;

    bool done = false;
    while (!done)
    {
      done = true; // unless you prove me wrong

      // get the first cluster, add it to the new fragment, remove from parent fragment
      TCluster* first_cluster = ((TCluster*)ParentFragment->Clusters->Items[0]);

      // find dissimilar clusters in the remaining clusters
      // and remove them from parent fragment, add to the new fragment,
      for (int i=1; i<ParentFragment->Clusters->Count; i++)
	  {
        TCluster* cluster = ((TCluster*)ParentFragment->Clusters->Items[i]);
        // PartitioningFC->Run();
        if ( ! first_cluster->similar_in_demand_patterns_to(cluster) )
        {
///////////////////////
//ChildForm->so_counter++;
///////////////////////
            // create a new empty fragment to hole removed clusters
            if (ChildFragment == NULL)
            {
              // Partition();
              ChildFragment = new TFragment(Node);
              ChildFragment->set_id(this->get_id()*10 + ++partitioned_fragment_id_serial_num);
              ChildFragment->Clusters->Clear();
    		  Node->AddFragment(ChildFragment);
              ChildFragment->SO_On();
            }

            ParentFragment->DeleteCluster(i);
            ChildFragment->AddCluster(cluster);

            i--; // shift up
            done = false;
        }
      }

      if (gradual_partitioning)
      {
        break;
      }

      // partition new child
      ParentFragment = ChildFragment;
      ChildFragment  = NULL;
    }
   }
  }







//  if ( (ChildForm->running_counter > 0) && (ChildForm->SO_FREQUENCY > 0)  &&  (ChildForm->M_FREQUENCY > 0)  && ((ChildForm->running_counter%(ChildForm->SO_FREQUENCY/ChildForm->M_FREQUENCY)) == 0) )
  if ( ChildForm->m_enabled && (ChildForm->SO_FREQUENCY > 0)  &&  (ChildForm->M_FREQUENCY > 0)  && ((ChildForm->running_counter%(ChildForm->SO_FREQUENCY/ChildForm->M_FREQUENCY)) == 0) )
  {
    try
    {
        migration_node_id  = MigrationFC->Run();
    }
    catch (...)
    {
        // slience errors
    }

    if (migration_node_id != INVALID_ID)
	{
	  TNode* migration_node = Node->Network->GetNode(migration_node_id);
      if (migration_node != NULL)
      {
        Migrate(migration_node);
///////////////////////
ChildForm->so_counter++;
///////////////////////
      }
    }
  }












  // disable alternating if frequencies are not the same, and enable both
  if (ChildForm->P_FREQUENCY != ChildForm->C_FREQUENCY)
  {
    TFragment::partition_or_combine = false;
  }
  if ( ! TFragment::partition_or_combine)
  {
   if ( ChildForm->c_enabled && (Replicas->Count == 0) && (ChildForm->SO_FREQUENCY > 0)  &&  (ChildForm->C_FREQUENCY > 0)  && ((ChildForm->running_counter%(ChildForm->SO_FREQUENCY/ChildForm->C_FREQUENCY)) == 0) )
   {
    //bool gradual_combining = true;
    bool gradual_combining = false;

    TFragment* ThisFragment = this;
    TFragment* CombineWithFragment;

    for (int i=0; i<ThisFragment->Node->Fragments->Count; i++)
    {
        CombineWithFragment  = (TFragment*)ThisFragment->Node->Fragments->Items[i];

        if (CombineWithFragment == ThisFragment) continue;  // skip
        if (CombineWithFragment->Replicas->Count > 0) continue;  // skip

        // CombiningFC->Run();
        if ( CombineWithFragment->similar_in_demand_patterns_to(ThisFragment) )
        {
///////////////////////
//ChildForm->so_counter++;
///////////////////////
            // Combine();
            for (int i=0; i<ThisFragment->Clusters->Count; i++)
            {
              TCluster* cluster = ((TCluster*)ThisFragment->Clusters->Items[i]);
              ThisFragment->DeleteCluster(i);
    		  CombineWithFragment->AddCluster(cluster);
              i--;
            }
            CombineWithFragment->set_id(ThisFragment->get_id());
            delete ThisFragment->Node->RemoveFragment(ThisFragment);

            if (gradual_combining)
            {
              break;
            }

            ThisFragment = CombineWithFragment;

            i--; // process the new fragment shifted up in place of the deletion one.
        }
    }
   }

  }









  if ( ChildForm->r_enabled && (ChildForm->running_counter > 0) && (ChildForm->SO_FREQUENCY > 0)  &&  (ChildForm->R_FREQUENCY > 0)  && ((ChildForm->running_counter%(ChildForm->SO_FREQUENCY/ChildForm->R_FREQUENCY)) == 0) )
  {
    for (int i=0; i<MAX_NODES; i++)
	{
      replication_node_ids[i] = INVALID_ID;
    }

    try
    {
        ReplicationFC->Run(replication_node_ids);
    }
    catch (...)
    {
        // slience errors
    }

    for (int i=0; i<MAX_NODES; i++)
	{
      if (replication_node_ids[i] != INVALID_ID)
      {
	    TNode* replication_node = Node->Network->GetNode(replication_node_ids[i]);
        if (replication_node != NULL)
        {
          Replicate(replication_node);
///////////////////////
ChildForm->so_counter++;
///////////////////////
        }
      }
    }
  }





  if ( ChildForm->d_enabled && (ChildForm->SO_FREQUENCY > 0)  &&  (ChildForm->D_FREQUENCY > 0)  && ((ChildForm->running_counter%(ChildForm->SO_FREQUENCY/ChildForm->D_FREQUENCY)) == 0) )
  {
    for (int i=0; i<Replicas->Count; i++)
    {
        TFragment* replica = (TFragment*)Replicas->Items[i];
        if (  (replica != NULL) && (replica->demand < TFragment::ReplicaDestructionThreshold)  )
        {
            // destroy low_demanded_replica
            if (  ((TNode*)replica->Node)->RemoveFragment(replica) )
            {
                //Replicas->Remove(replica);
                delete replica; // will also undistribute demand back to original
            }

            i--;    // pack

///////////////////////
//ChildForm->so_counter++;
///////////////////////

        }
    }
 }

  return migration_node_id;
 }

 return -1;
}

//*/
