//---------------------------------------------------------------------------
#ifndef TFragmentH
#define TFragmentH
//---------------------------------------------------------------------------


#define     MAX_FULL_SIZE_CLUSTERS_PER_FRAGMENT         10


#include "TCluster.h";
class TNode;

class	TMigrationFC;
class	TReplicationFC;


struct TDemandEntry			// single entry
{
	int 		NodeId;
	bool        Update;
    TDateTime	DateTime;
};


struct TDemandLevel			// accumulative entries from single demanding_node
{
	int 		NodeId;
	double      Demand;
	double      UpdateRatio;
};



class TFragment
{
	public:
    static bool    partition_or_combine;
    static double  PartitioningThreshold; 	    // minmum disimilarity level to justfy partitioning
    static double  CombiningThreshold;		    // minmum similarity level to justfy combining
    static double  ReplicaDestructionThreshold; // minmum demand level below which destruct replica

	static int	scale;
		   int	cluster_serial_num;
           int  partitioned_fragment_id_serial_num;


		float   demand;                     // average demand from all nodes
		float   update;                     // average updating demand from all nodes
		float	update_ratio;				// average update ratio from all nodes

		float   demands[MAX_NODES];			// demand per node
		float   updates[MAX_NODES];         // updating demand per node
		float   update_ratios[MAX_NODES];	// update/demand ratio per node

		float   total_demands[MAX_NODES];   // sum
		float   total_updates[MAX_NODES];   // sum

		float   response;            		// average response_time to all nodes
		float   responses[MAX_NODES]; 		// response_time to demanding node = (demand*???size???/100) * delay


	private:        // private user declarations
		int      id;
		TPoint   position;
		int      size; 		int 	radius;
		/*float    demand;*/	TColor	color;
        bool	 selected;
        TColor   selection_color;
	    int		 selected_cluster_index;

        bool	 replica;
        bool	 SO;		// self organization switch on/off

public:         // public user declarations
    TNode*              Node;
    TMigrationFC*       MigrationFC;
    TReplicationFC*     ReplicationFC;
	TList*              Clusters;
	TList*              Replicas;

   	TList*			DemandLevelTable;		// in current SO time unit (days)
   	TList*			DemandHistoryTable;		// last H time unit        (years)

	TFragment(TFragment* _fragment);
	TFragment(TNode* _Node);
	virtual ~TFragment();

	int      get_id() 		    			{ return id;}
	TPoint   get_position();
	int		 get_size()      				{ return size;}
	int      get_radius()      				{ return radius;}
	float	 get_demand() 					{ return demand;}
	TColor   get_color()         			{ return color;}
	bool     IsReplica()	     			{ return replica;}
	int  	 get_selected_cluster_index()   { return selected_cluster_index;}

	float	 calculate_mobility();
	float	 calculate_availability();      // = Replicas->Count / Demanding Nodes
	float	 calculate_update_level()       { return update_ratio; }

	void set_id(int _id)	               			{ id        	= _id;}
	void set_position(TPoint _position);
   	int  update_size();
    TFragment* find_nearest_replica_to_node(TNode* _node);
	int  update_demand();    // also updates collective responses[]
	void update_DemandLevelTable();
    void update_response();
    void set_replica(bool _replica)					{ replica		= _replica;}
    void set_selected_cluster_index(int _index)		{ selected_cluster_index = _index; }

	void AddCluster(TCluster* _Cluster);
	void DeleteCluster(int _index);


    bool similar_in_demand_patterns_to(TFragment* _fragment); // for fragment combining

	// self-organization behaivour
	void SO_On    ();
	void SO_Off   ();
	int  Self_Organize();

	void Partition(int _fromClusterIndex, int _toClusterIndex);
	void Migrate  (TNode* _toNode);
	void Combine  (TFragment* _withFragment);
	void Replicate(TNode* _onNode);
    void DestroyReplica(TFragment* _replica);

    
	void Draw(TCanvas* _canvas);
	void DrawCaption(TCanvas* _canvas);
	void DrawDetails(TCanvas* _canvas);

	bool IsSelected();
    void Select(TColor _selection_color);
    void DeSelect();
    void DeSelectChildren();
    bool AnySelection();
};
//---------------------------------------------------------------------------
#endif
