////////////////////////////////////////////////////////////////////////////////
#include "TFuzzySet.h"
////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
//TVariable class implementation
///////////////////////////////////////////////////////////////////////////////
TVariable::TVariable(char* _name, char* _unit, double _begin, double _end)
{
	if (_name != NULL)
	{
    	name	= new char[strlen(_name)+1];
		strcpy(name, _name);
    }
    if (_unit != NULL)
    {
       	unit	= new char[strlen(_unit)+1];
        strcpy(unit, _unit);
    }
	begin	= _begin;
	end		= _end;

    for (int i=0; i<MAX_FUZZYSETS_PER_VARIABLE; i++)
    {
		fuzzysets[i] = NULL;
    }
}


TVariable::~TVariable()
{
   	if (name != NULL) delete[] name;
    name = NULL;
   	if (unit != NULL) delete[] unit;
    unit = NULL;
	clear_fuzzysets();
}


void TVariable::clear_fuzzysets()
{
  for (int i=0; i<MAX_FUZZYSETS_PER_VARIABLE; i++)
  {
		if (fuzzysets[i] != NULL) delete fuzzysets[i];
		fuzzysets[i] = NULL;
  }
}

void TVariable::add_fuzzyset(TFuzzySet* fuzzyset)
{
  for (int i=0; i<MAX_FUZZYSETS_PER_VARIABLE; i++)
  {
  	if (fuzzysets[i] == NULL)
    {
		fuzzyset->set_variable(this);
		fuzzysets[i] = fuzzyset;
  	    break;
    }
  }
}

void TVariable::remove_fuzzyset(TFuzzySet* fuzzyset)
{
  for (int i=0; i<MAX_FUZZYSETS_PER_VARIABLE; i++)
  {
  	if (*(fuzzysets[i]) == *fuzzyset)
    {
		delete fuzzysets[i];
		// shift up
		for (int j=i; j<MAX_FUZZYSETS_PER_VARIABLE-1; j++)
    	{
			fuzzysets[j] = fuzzysets[j+1];
	    }
		// invalidate last
        fuzzysets[MAX_FUZZYSETS_PER_VARIABLE-1] = NULL;
        break;
    }
  }
}


bool TVariable::operator == (TVariable& variable2)
{
	if (*name  != *(variable2.name)) return false;
	if (*unit  != *(variable2.unit)) return false;
	if (begin  != variable2.begin) 	 return false;
	if (end    != variable2.end)     return false;
	for (int i=0; i<MAX_FUZZYSETS_PER_VARIABLE; i++)
   	{
    	if (!(fuzzysets[i] == variable2.fuzzysets[i])) return false;
    }
    return true;
}
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
//TFuzzySet class implementation
///////////////////////////////////////////////////////////////////////////////
TFuzzySet::TFuzzySet(	TVariable*	_variable,
						char* 		_name,
                        int 		_shape,
                        int 		_direction,
						double 		_min,
                        double 		_mid1,
                        double 		_mid,
                        double 		_mid2,
                        double 		_max,
                        double 		_minTruth,
                        double 		_maxTruth
					)
{
	variable	= _variable;
    if (_name != NULL)
   	{
   		name	= new char[strlen(_name)+1];
        strcpy(name, _name);
   	}
	shape		= _shape;
	direction	= _direction;
    if (variable != NULL)
    {
		if (_min  < variable->begin) _min = variable->begin;
		if (_min  > variable->end)   _min = variable->end;
		if (_mid1 < variable->begin) _mid1= variable->begin;
		if (_mid1 > variable->end)   _mid1= variable->end;
		if (_mid  < variable->begin) _mid = variable->begin;
		if (_mid  > variable->end)   _mid = variable->end;
		if (_mid2 < variable->begin) _mid2= variable->begin;
		if (_mid2 > variable->end)   _mid2= variable->end;
		if (_max  < variable->begin) _max = variable->begin;
		if (_max  > variable->end)   _max = variable->end;
		min		= ((_min  - variable->begin)/(variable->end - variable->begin))*DOMAIN_WIDTH;
		mid1	= ((_mid1 - variable->begin)/(variable->end - variable->begin))*DOMAIN_WIDTH;
		mid		= ((_mid  - variable->begin)/(variable->end - variable->begin))*DOMAIN_WIDTH;
		mid2	= ((_mid2 - variable->begin)/(variable->end - variable->begin))*DOMAIN_WIDTH;
		max		= ((_max  - variable->begin)/(variable->end - variable->begin))*DOMAIN_WIDTH;
	}
	minTruth	= _minTruth;
	maxTruth	= _maxTruth;
	for(int i=0; i<DOMAIN_POINTS; i++) 	truths[i] = ZERO_TRUTH;
	for(int i=0; i<MAX_HEDGES; i++)    	hedges[i] = INVALID_HEDGE;
	interpolate_truths();
	set_alphaCut(MIN_TRUTH);
}


TFuzzySet::TFuzzySet(TFuzzySet& set2)
{
	set_variable(set2.get_variable());
    if (set2.get_name() != NULL)
   	{
   		name	= new char[strlen(set2.get_name())+1];
        strcpy(name, set2.get_name());
   	}
	set_shape(set2.get_shape());
	set_direction(set2.get_direction());
	set_min (set2.get_min());
	set_mid1(set2.get_mid1());
	set_mid (set2.get_mid());
	set_mid2(set2.get_mid2());
	set_max (set2.get_max());
	set_minTruth(set2.get_minTruth());
	set_maxTruth(set2.get_maxTruth());
	for(int i=0; i<MAX_HEDGES; i++)    	hedges[i] = set2.hedges[i];
	for(int i=0; i<DOMAIN_POINTS; i++) 	truths[i] = set2.truths[i];
	set_alphaCut(set2.get_alphaCut());
}


// convert scalar into a singleton fuzzy set
TFuzzySet::TFuzzySet(double scalar)
{
	variable	= new TVariable();
    name 		= new char[strlen("Singleton")+1];	strcpy(name, "Singleton");
	shape		= SINGLETON;
	direction	= UP;
	if (scalar  < variable->begin) scalar = variable->begin;
	if (scalar  > variable->end)   scalar = variable->end;
	min			= ((scalar - variable->begin)/(variable->end - variable->begin))*DOMAIN_WIDTH;
	mid1		= ((scalar - variable->begin)/(variable->end - variable->begin))*DOMAIN_WIDTH;
	mid			= ((scalar - variable->begin)/(variable->end - variable->begin))*DOMAIN_WIDTH;
	mid2		= ((scalar - variable->begin)/(variable->end - variable->begin))*DOMAIN_WIDTH;
	max			= ((scalar - variable->begin)/(variable->end - variable->begin))*DOMAIN_WIDTH;
	minTruth	= MIN_TRUTH;
	maxTruth	= MAX_TRUTH;
	for(int i=0; i<DOMAIN_POINTS; i++) 	truths[i] = ZERO_TRUTH;
	for(int i=0; i<MAX_HEDGES; i++)    	hedges[i] = INVALID_HEDGE;
	interpolate_truths();
	set_alphaCut(MIN_TRUTH);
}


TFuzzySet::~TFuzzySet()
{
	variable = NULL; // don't delete, not owned by this
	if (name != NULL) delete[] name;
    name = NULL;
}






void TFuzzySet::interpolate_truths()
{
	switch (shape)
    {
    	case NO_SHAPE:
        	{
				for (int x=0; x<min; x++)				{ truths[x] = ZERO_TRUTH; }
				for (int x=min; x<=max; x++)
				{
			   		truths[x] = ZERO_TRUTH;
				}
				for (int x=max+1; x<DOMAIN_POINTS; x++)	{ truths[x] = ZERO_TRUTH; }
            }
            break;
    	case SINGLETON:
        	{
				for (int x=0; x<min; x++)				{ truths[x] = ZERO_TRUTH; }
				for (int x=min; x<mid; x++)
				{
			   		truths[x] = minTruth;
				}

		   		truths[(int)mid] = maxTruth;

				for (int x=mid+1; x<=max; x++)
				{
			   		truths[x] = minTruth;
				}
				for (int x=max+1; x<DOMAIN_POINTS; x++)	{ truths[x] = ZERO_TRUTH; }
			}
            break;
    	case PLATEAU:
           	{
				for (int x=0; x<min; x++)				{ truths[x] = ZERO_TRUTH; }
				for (int x=min; x<=max; x++)
				{
			   		truths[x] = maxTruth;
				}
				for (int x=max+1; x<DOMAIN_POINTS; x++)	{ truths[x] = ZERO_TRUTH; }
            }
            break;
    	case LINEAR:
        	{
  			  for (int x=0; x<min; x++)				{ truths[x] = ZERO_TRUTH; }
			  if(direction == UP)
              {
              	for (int x=min; x<=max; x++)
				{
			   		truths[x] = minTruth + (x-min) * ((maxTruth-minTruth)/(max-min));
				}
              }
              else if(direction == DOWN)
              {
              	for (int x=min; x<=max; x++)
				{
			   		truths[x] = maxTruth - (x-min) * ((maxTruth-minTruth)/(max-min));
				}
              }
			  for (int x=max+1; x<DOMAIN_POINTS; x++)	{ truths[x] = ZERO_TRUTH; }
            }
            break;
    	case TRIANGLE:
        	{
				for (int x=0; x<min; x++)				{ truths[x] = ZERO_TRUTH; }
				for (int x=min; x<mid; x++)
				{
			   		truths[x] = minTruth + (x-min) * ((maxTruth-minTruth)/(mid-min));
				}
				for (int x=mid; x<=max; x++)
				{
			   		truths[x] = maxTruth - (x-mid) * ((maxTruth-minTruth)/(max-mid));
				}
				for (int x=max+1; x<DOMAIN_POINTS; x++) { truths[x] = ZERO_TRUTH; }
            }
            break;
    	case TRAPEZOIDAL:
        	{
				for (int x=0; x<min; x++)				{ truths[x] = ZERO_TRUTH; }
				for (int x=min; x<mid1; x++)
				{
			   		truths[x] = minTruth + (x-min) * ((maxTruth-minTruth)/(mid1-min));
				}
				for (int x=mid1; x<mid2; x++)
				{
			   		truths[x] = maxTruth;
				}
				for (int x=mid2; x<=max; x++)
				{
			   		truths[x] = maxTruth - (x-mid2) * ((maxTruth-minTruth)/(max-mid2));
				}
				for (int x=max+1; x<DOMAIN_POINTS; x++) { truths[x] = ZERO_TRUTH; }
            }
            break;
    	case SHOULDER:
        	{
			  for (int x=0; x<min; x++)				{ truths[x] = ZERO_TRUTH; }
              if(direction == LEFT)
              {
				for (int x=min; x<mid; x++)
				{
			   		truths[x] = maxTruth;
				}
				for (int x=mid; x<=max; x++)
				{
			   		truths[x] = maxTruth - (x-mid) * ((maxTruth-minTruth)/(max-mid));
				}
              }
			  else if(direction == RIGHT)
              {
				for (int x=min; x<mid; x++)
				{
			   		truths[x] = minTruth + (x-min) * ((maxTruth-minTruth)/(mid-min));
				}
				for (int x=mid; x<=max; x++)
				{
			   		truths[x] = maxTruth;
				}
              }
			  for (int x=max+1; x<DOMAIN_POINTS; x++) 	{ truths[x] = ZERO_TRUTH; }
            }
            break;
    	case S_CURVE:
        	{
			  for (int x=0; x<min; x++)				{ truths[x] = ZERO_TRUTH; }
			  if(direction == UP)
              {
              	for (int x=min; x<mid; x++)
				{
			   		truths[x] = minTruth+(maxTruth-minTruth)*(2* ( ((x-min)/((mid-min)*2)) * ((x-min)/((mid-min)*2)) ));
				}
              	for (int x=mid; x<=max; x++)
				{
			   		truths[x] = minTruth+(maxTruth-minTruth)*(1 - 2* ( ((max-x)/((max-mid)*2)) * ((max-x)/((max-mid)*2)) ));
				}
              }
              else if(direction == DOWN)
              {
              	for (int x=min; x<mid; x++)
				{
			   		truths[x] = minTruth+(maxTruth-minTruth)*(1 - (2* ( ((x-min)/((mid-min)*2)) * ((x-min)/((mid-min)*2)) )));
				}
              	for (int x=mid; x<=max; x++)
				{
			   		truths[x] = minTruth+(maxTruth-minTruth)*(1 - (1 - 2* ( ((max-x)/((max-mid)*2)) * ((max-x)/((max-mid)*2)) )));
				}
              }
			  for (int x=max+1; x<DOMAIN_POINTS; x++)	{ truths[x] = ZERO_TRUTH; }
            }
            break;
    	case PI_CURVE:
        	{
				for (int x=0; x<min; x++)				{ truths[x] = ZERO_TRUTH; }
              	for (int x=min; x<mid1; x++)
				{
			   		truths[x] = minTruth+(maxTruth-minTruth)*(2* ( ((x-min)/((mid1-min)*2)) * ((x-min)/((mid1-min)*2)) ));
				}
              	for (int x=mid1; x<mid; x++)
				{
			   		truths[x] = minTruth+(maxTruth-minTruth)*(1 - 2* ( ((mid-x)/((mid-mid1)*2)) * ((mid-x)/((mid-mid1)*2)) ));
				}
              	for (int x=mid; x<mid2; x++)
				{
			   		truths[x] = minTruth+(maxTruth-minTruth)*(1 - (2* ( ((x-mid)/((mid2-mid)*2)) * ((x-mid)/((mid2-mid)*2)) )));
				}
              	for (int x=mid2; x<=max; x++)
				{
			   		truths[x] = minTruth+(maxTruth-minTruth)*(1 - (1 - 2* ( ((max-x)/((max-mid2)*2)) * ((max-x)/((max-mid2)*2)) )));
				}
				for (int x=max+1; x<DOMAIN_POINTS; x++)	{ truths[x] = ZERO_TRUTH; }
            }
            break;
    	case BETA_CURVE:
        	{
				for (int x=0; x<min; x++)				{ truths[x] = ZERO_TRUTH; }
              	for (int x=min; x<mid; x++)
				{
			   		truths[x] = 1 / ( 1 + ( ((mid-x)/(mid1-min)) * ((mid-x)/(mid1-min)) ) );
				}
              	for (int x=mid; x<=max; x++)
				{
			   		truths[x] = 1 / ( 1 + ( ((x-mid)/(mid2-mid)) * ((x-mid)/(mid2-mid)) ) );
				}
				for (int x=max+1; x<DOMAIN_POINTS; x++)	{ truths[x] = ZERO_TRUTH; }
            }
            break;
    	case GAUSSIAN_CURVE:
        	{
				for (int x=0; x<min; x++)				{ truths[x] = ZERO_TRUTH; }
              	for (int x=min; x<mid; x++)
				{
			   		truths[x] = minTruth+(maxTruth-minTruth)*(exp( -(mid-min) * ((mid-x)/max)*((mid-x)/max) ));
				}
              	for (int x=mid; x<=max; x++)
				{
			   		truths[x] = minTruth+(maxTruth-minTruth)*(exp( -(max-mid) * ((x-mid)/max)*((x-mid)/max) ));
				}
				for (int x=max+1; x<DOMAIN_POINTS; x++)	{ truths[x] = ZERO_TRUTH; }
            }
            break;
    	default:
				for (int x=0; x<DOMAIN_POINTS; x++) 	{ truths[x] = ZERO_TRUTH; }
    }
}




void TFuzzySet::clear_hedges()
{
  for(int i=0; i<MAX_HEDGES; i++)
  {
  	hedges[i] == INVALID_HEDGE;
  }
}


void TFuzzySet::add_hedge (int type)
{
  for(int i=0; i<MAX_HEDGES; i++)
  {
  	if (hedges[i] == INVALID_HEDGE)
    {
		hedges[i] = type;

		switch(type)
        {
// negation
          case NOT:				// !
          {
			*this = !*this;
          }
          break;
// fuzzify a scalar, or widen
          case ABOUT:			// fuzzify(scalar), or widen by 1.3
          //case AROUND:
          //case ROUGHLY:
          {
		  		//if(shape == SINGLETON) then become PI Curve, else widen
               	widen(1.3);
          }
          break;
// restrict
          case ABOVE:			// shift_right(50%)
          //case MORE_THAN:
          {
			shift_right(0.5);
          }
          break;
// restrict
          case BELOW:			// shift_left(50%)
          //case LESS_THAN:
          {
			shift_left(0.5);
          }
          break;
// contrast intensification (rotate_anti_clock_wise)
          case POSITIVELY:      // if truths[x]<=0.5 -->     n*((  x)^n) , n=2.0
          						// if truths[x]> 0.5 --> 1 - n*((1-x)^n) , n=2.0
          //case ALMOST:
          //case DEFINITELY:
          {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
    			if (truths[x] <= 0.5)
		        {
       			   	truths[x] =        ( 2.0 * pow(truths[x], 2.0) );
                }
   			    else if (truths[x] > 0.5)
	        	{
    		       	truths[x] = MAX_TRUTH -  ( 2.0 * pow((MAX_TRUTH-truths[x]), 2.0) );
        		}
		    }
          }
          break;
// contrast diffusion  (rotate_clock_wise)
          case GENERALLY:       // if truths[x]<=0.5 -->     ((1/n)*(  x))^(1/n) , n=2.0
          						// if truths[x]> 0.5 --> 1 - ((1/n)*(1-x))^(1/n) , n=2.0
          //case USUALLY:
          {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
    			if (truths[x] <= 0.5)
		        {
       			   	truths[x] =        pow( ((1/2.0)*truths[x]), (1/2.0) );
                }
   			    else if (truths[x] > 0.5)
	        	{
    		       	truths[x] = MAX_TRUTH -  pow((1/2.0)*((MAX_TRUTH-truths[x])), (1/2.0) );
        		}
		    }
          }
          break;
// approximate broadly
          case IN_THE_VICINITY_OF:	// widen to double
          {
		  		widen(2.0);
          }
          break;
// approximate narrowly
          case CLOSE_TO:        // narrow by half
          //case NEIGHBORING:
          {
		  		narrow(0.5);
          }
          break;
// dilute
          case SLIGHTLY:		// x^(1/n)	, n=3.0
          {
			for (int x=0; x<DOMAIN_POINTS; x++)
	        {
    	    	truths[x] = pow(truths[x], 1/3.0);
        	}
          }
          break;
          case SOMEWHAT:		// x^(1/n)	, n=2.0
          //case RATHER:
          //case QUITE:
          {
			for (int x=0; x<DOMAIN_POINTS; x++)
	        {
    	    	truths[x] = pow(truths[x], 1/2.0);
        	}
          }
          break;
// intensify
          case VERY:			// x^n		, n=2.0
          {
			for (int x=0; x<DOMAIN_POINTS; x++)
	        {
    	    	truths[x] = pow(truths[x], 2.0);
        	}
          }
          break;
          case EXTREMELY:		// x^n		, n=3.0
          {
			for (int x=0; x<DOMAIN_POINTS; x++)
	        {
    	    	truths[x] = pow(truths[x], 3.0);
        	}
          }
          break;

        }

	    minTruth  = MAX_TRUTH;
    	maxTruth  = MIN_TRUTH;
		for (int x=0; x<DOMAIN_POINTS; x++)
		{
		    minTruth  = (minTruth < truths[x]) ? minTruth : truths[x];
		    maxTruth  = (maxTruth > truths[x]) ? maxTruth : truths[x];
        }

  	    break;	// for loop
    }// end if
  }
}






void TFuzzySet::remove_hedge(int type)
{
  for(int i=0; i<MAX_HEDGES; i++)
  {
   	if (hedges[i] == type)
    {
		// shift up
		for(int j=i; j<MAX_HEDGES-1; j++)
    	{
			hedges[j] = hedges[j+1];
	    }
		// invalidate last hedge in case it had a valid hedge
        hedges[MAX_HEDGES-1] = INVALID_HEDGE;


		switch(type)
        {
// negation
          case NOT:				// !
          {
			*this = !*this;
          }
          break;
// fuzzify a scalar, or widen
          case ABOUT:			// fuzzify(scalar), or widen by 1.3
          //case AROUND:
          //case ROUGHLY:
          {
		  		if(shape == SINGLETON)
                {
                	//PI_CURVE_it(mid);
                }
                else
                {
	               	narrow(1/1.3);
                }
          }
          break;
// restrict
          case ABOVE:			// shift_right(50%)
          //case MORE_THAN:
          {
			shift_left(0.5);
          }
          break;
// restrict
          case BELOW:			// shift_left(50%)
          //case LESS_THAN:
          {
			shift_right(0.5);
          }
          break;
// contrast intensification (rotate_anti_clock_wise)
          case POSITIVELY:      // if truths[x]<=0.5 -->     n*((  x)^n) , n=2.0
          						// if truths[x]> 0.5 --> 1 - n*((1-x)^n) , n=2.0
          //case ALMOST:
          //case DEFINITELY:
          {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
    			if (truths[x] <= 0.5)
		        {
       			   	truths[x] =        pow( (truths[x]*(1/2.0)), (1/2.0) );
                }
   			    else if (truths[x] > 0.5)
	        	{
    		       	truths[x] = MAX_TRUTH -  pow(((MAX_TRUTH-truths[x]))*(1/2.0), (1/2.0) );
        		}
		    }
          }
          break;
// contrast diffusion  (rotate_clock_wise)
          case GENERALLY:       // if truths[x]<=0.5 -->     ((1/n)*(  x))^(1/n) , n=2.0
          						// if truths[x]> 0.5 --> 1 - ((1/n)*(1-x))^(1/n) , n=2.0
          //case USUALLY:
          {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
    			if (truths[x] <= 0.5)
		        {
       			   	truths[x] =        ( 2.0 * pow(truths[x], 2.0) );
                }
   			    else if (truths[x] > 0.5)
	        	{
    		       	truths[x] = MAX_TRUTH -  ( 2.0 * pow((MAX_TRUTH-truths[x]), 2.0) );
        		}
		    }
          }
          break;
// approximate broadly
          case IN_THE_VICINITY_OF:	// widen to double
          {
		  		narrow(1/2.0);
          }
          break;
// approximate narrowly
          case CLOSE_TO:        // narrow by half
          //case NEIGHBORING:
          {
		  		widen(1/0.5);
          }
          break;
// dilute
          case SLIGHTLY:		// x^(1/n)	, n=3.0
          {
			for (int x=0; x<DOMAIN_POINTS; x++)
	        {
    	    	truths[x] = pow(truths[x], 1/(1/3.0));
        	}
          }
          break;
          case SOMEWHAT:		// x^(1/n)	, n=2.0
          //case RATHER:
          //case QUITE:
          {
			for (int x=0; x<DOMAIN_POINTS; x++)
	        {
    	    	truths[x] = pow(truths[x], 1/(1/2.0));
        	}
          }
          break;
// intensify
          case VERY:			// x^n		, n=2.0
          {
			for (int x=0; x<DOMAIN_POINTS; x++)
	        {
    	    	truths[x] = pow(truths[x], 1/2.0);
        	}
          }
          break;
          case EXTREMELY:		// x^n		, n=3.0
          {
			for (int x=0; x<DOMAIN_POINTS; x++)
	        {
    	    	truths[x] = pow(truths[x], 1/3.0);
        	}
          }
          break;

        }

	    minTruth  = MAX_TRUTH;
    	maxTruth  = MIN_TRUTH;
		for (int x=0; x<DOMAIN_POINTS; x++)
		{
		    minTruth  = (minTruth < truths[x]) ? minTruth : truths[x];
		    maxTruth  = (maxTruth > truths[x]) ? maxTruth : truths[x];
        }

        break; // for loop
    }// end if
  }
}

////////////////////////////////////////////////////////////////////////////////














////////////////////////////////////////////////////////////////////////////////
// manipulator functions
////////////////////////////////////////////////////////////////////////////////
// multiply all by MAX_TRUTH/maxTruth() so that maxTruth()=MAX_TRUTH
void  TFuzzySet::normalize()
{
	for (int x=0; x<DOMAIN_POINTS; x++)
   	{
		truths[x] = truths[x] * (MAX_TRUTH/maxTruth);
    }

    minTruth  = MAX_TRUTH;
    maxTruth  = MIN_TRUTH;
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
	    minTruth  = (minTruth < truths[x]) ? minTruth : truths[x];
	    maxTruth  = (maxTruth > truths[x]) ? maxTruth : truths[x];
    }
}


// is normalised?
bool  TFuzzySet::normalized()
{
	return (maxTruth == MAX_TRUTH);
}


// for all truth(x) > _maxTruth make truth(x) = _maxTruth
void	TFuzzySet::truncate(double _maxTruth)
{
	for (int x=0; x<DOMAIN_POINTS; x++)
   	{
		if (truths[x] > _maxTruth)
        {
        	truths[x] = _maxTruth;
        }
    }

    minTruth  = MAX_TRUTH;
    maxTruth  = MIN_TRUTH;
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
	    minTruth  = (minTruth < truths[x]) ? minTruth : truths[x];
	    maxTruth  = (maxTruth > truths[x]) ? maxTruth : truths[x];
    }
}


// STRONG	restricts domain to x:truth(x) >  alphaCut
// WEAK		restricts domain to x:truth(x) >= alphaCut
void TFuzzySet::apply_alphaCut(int type)
{
	for (int x=0; x<DOMAIN_POINTS; x++)
    {
		if (type == STRONG)
        {
	    	if (truths[x] <= alphaCut)
    	    {
        		truths[x] = ZERO_TRUTH;
	        }
        }
        else if (type == WEAK)
        {
	    	if (truths[x] < alphaCut)
    	    {
        		truths[x] = ZERO_TRUTH;
	        }
        }
    }

	double left  = INVALID_DOMAIN;
	double right = INVALID_DOMAIN;
	for (int x=0; x<DOMAIN_POINTS; x++)
    {
        if (truths[x] != ZERO_TRUTH)
        {
           	if (left == INVALID_DOMAIN)
            {
	            left = x;
	            right = x;	// in case there is only one point above alpha
            }
            else
	        {
                right = x;
            }
        }
    }
    min = left;
    max = right;

    minTruth  = MAX_TRUTH;
   	maxTruth  = MIN_TRUTH;
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
	    minTruth  = (minTruth < truths[x]) ? minTruth : truths[x];
	    maxTruth  = (maxTruth > truths[x]) ? maxTruth : truths[x];
    }
}


// convert to a scalar
double	TFuzzySet::defuzzify(int how)
{
	double scalar = INVALID_DOMAIN;
	switch(how)
   	{
    	case CENTROID:				//find domain point x at the centre of gravity, {(Xi*Yi)/{Yi
    	//case COMPOSITE_MOMENTS:
        {
			double productsum = 0.0;
			double D = 0.0;
			for (int x=0; x<DOMAIN_POINTS; x++)
		   	{
            	double y = (MIN_TRUTH > truths[x]) ? 0.0 : truths[x];
				productsum += x*y;
                D += y;
		    }

            if (D != 0.0)
              scalar = productsum / D;
            else
              scalar = mid;
        }
        break;
    	case COMPOSITE_MAXIMUM:		//find domain point x with maximum truth, if plateau, then use either AVERAGE_MAXIMUM, or CENTRE_OF_MAXIMUMS
    	//case MAXIMUM_HEIGHT:
        {
		    minTruth  = MAX_TRUTH;
    		maxTruth  = MIN_TRUTH;
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
			    minTruth  = (minTruth < truths[x]) ? minTruth : truths[x];
			    maxTruth  = (maxTruth > truths[x]) ? maxTruth : truths[x];
	        }

			double MaxLeft  = INVALID_DOMAIN;
			double MaxRight = INVALID_DOMAIN;
			for (int x=0; x<DOMAIN_POINTS; x++)
		   	{
//				if (abs(truths[x] - maxTruth) < TOLERANCE)
				if (truths[x] == maxTruth)
                {
                	if (MaxLeft == INVALID_DOMAIN)
                    {
	                  MaxLeft = x;     		// first x of maxTruth
			          MaxRight = x;			// in case there is only one max point
                    }
                    else
	                {
                      MaxRight = x;			// overright with last x of maxTruth
                    }
                }
		    }

//           	if  (abs(truths[(int)( (MaxLeft + MaxRight)/(2) )] - maxTruth) < TOLERANCE) // multi plateaus
         	if  (truths[(int)( (MaxLeft + MaxRight)/(2) )] == maxTruth) // multi plateaus
                scalar = (MaxLeft + MaxRight) / 2;
            else
                scalar = MaxLeft;
        }
        break;
    	case NEAR_EDGE:				//find x at the left edge of usually a single-edged plateau
        {
		    minTruth  = MAX_TRUTH;
    		maxTruth  = MIN_TRUTH;
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
			    minTruth  = (minTruth < truths[x]) ? minTruth : truths[x];
			    maxTruth  = (maxTruth > truths[x]) ? maxTruth : truths[x];
	        }

			for (int x=0; x<DOMAIN_POINTS; x++)
		   	{

//				if (abs(truths[x] - maxTruth) < TOLERANCE)
				if (truths[x] == maxTruth)
                {
		            scalar = x;		// first x of maxTruth
                    break;
                }
		    }
        }
        break;
    	case FAR_EDGE:				//find x at the right edge of usually a single-edged plateau
        {
		    minTruth  = MAX_TRUTH;
    		maxTruth  = MIN_TRUTH;
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
			    minTruth  = (minTruth < truths[x]) ? minTruth : truths[x];
			    maxTruth  = (maxTruth > truths[x]) ? maxTruth : truths[x];
	        }

			for (int x=0; x<DOMAIN_POINTS; x++)
		   	{
//				if (abs(truths[x] - maxTruth) < TOLERANCE)
				if (truths[x] == maxTruth)
                {
		            scalar = x;		// last x of maxTruth
                }
		    }
        }
        break;
    	case CENTROID_SUPPORT_SET:	//similar to CENTROID, but for only the part of the domain with truth(X) > alphaCut (STRONG)
        {
        	// NOT IMPLEMENTED YET
        }
        break;
    	case AVERAGE_MAXIMUM:		//similar to MAXIMUM_HEIGHT, but for only the highest plateau, or singlton
        {
        	// NOT IMPLEMENTED YET
        }
        break;
    	case CENTRE_OF_MAXIMUMS:	//similar to MAXIMUM_HEIGHT, but for only the two highest plateaus, or singltons
        {
        	// NOT IMPLEMENTED YET
        }
        break;
    	case CENTRE_OF_MASS:		//find x at the centre of the region supported by the maximum number of rules or evidence
    	//case BEST_EVIDENCE:
        {
        	// NOT IMPLEMENTED YET
        }
        break;
    	case LEAST_ENTROPY:			//find x for the least degree of fuzzyness (information entropy)
        {
        	// NOT IMPLEMENTED YET
        }
        break;
    }

	scalar = (scalar * (variable->end - variable->begin))/DOMAIN_WIDTH + variable->begin;
	return scalar;
}

////////////////////////////////////////////////////////////////////////////////











////////////////////////////////////////////////////////////////////////////////
// move and resize functions
////////////////////////////////////////////////////////////////////////////////
void	TFuzzySet::shift_left	(double percent)
{
	double temp[DOMAIN_POINTS];
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
        if ( (x>=min) && (x<=max) )
        {
	        temp[x] = truths[x];
        }
        else
        {
        	temp[x] = ZERO_TRUTH;
        }
    }

	double shift = (max - min) * percent;
    min  -= shift;
    mid1 -= shift;
    mid  -= shift;
    mid2 -= shift;
    max  -= shift;

	for (int x=0; x<DOMAIN_POINTS; x++)
	{
        if ( (x>=min) && (x<=max) )
        {
        	truths[x] = temp[(int)(x+shift)];
        }
        else
        {
        	truths[x] = ZERO_TRUTH;
        }
    }
}


void	TFuzzySet::shift_right	(double percent)
{
	double temp[DOMAIN_POINTS];
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
        if ( (x>=min) && (x<=max) )
        {
	        temp[x] = truths[x];
        }
        else
        {
        	temp[x] = ZERO_TRUTH;
        }
    }

	double shift = (max - min) * percent;
    min  += shift;
    mid1 += shift;
    mid  += shift;
    mid2 += shift;
    max  += shift;

	for (int x=0; x<DOMAIN_POINTS; x++)
	{
        if ( (x>=min) && (x<=max) )
        {
        	truths[x] = temp[(int)(x-shift)];
        }
        else
        {
        	truths[x] = ZERO_TRUTH;
        }
    }
}


void	TFuzzySet::shift_up	  	(double percent)
{
	double shift = (maxTruth - minTruth) * percent;
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
    	if (truths[x] != ZERO_TRUTH)
        {
        	truths[x] += shift;
        }
    }
    minTruth += shift;
    maxTruth += shift;
}


void	TFuzzySet::shift_down	(double percent)
{
	double shift = (maxTruth - minTruth) * percent;
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
    	if (truths[x] != ZERO_TRUTH)
        {
        	truths[x] -= shift;
        }
    }
    minTruth -= shift;
    maxTruth -= shift;
}


void	TFuzzySet::narrow 		(double percent)
{
	double temp[DOMAIN_POINTS];
    int x, j;
	for (x=min,j=0; x<=max; x=x+(1/percent),j++)
	{
        temp[j] = truths[x];
    }

    double _min  = mid - percent * (mid-min );
    double _mid1 = mid - percent * (mid-mid1);
    double _mid  = mid - percent * (mid-mid );
    double _mid2 = mid - percent * (mid-mid2);
    double _max  = mid - percent * (mid-max );
	min  = _min;
	mid1 = _mid1;
	mid  = _mid;
	mid2 = _mid2;
    max  = _max;

	for (int x=0; x<DOMAIN_POINTS; x++)
	{
        if ( (x>=min) && (x<=max) )
        {
        	truths[x] = temp[(int)(x-min)];
        }
        else
        {
        	truths[x] = ZERO_TRUTH;
        }
    }
}


void	TFuzzySet::widen  		(double factor)
{
	double temp[DOMAIN_POINTS];
	for (int x=min; x<max; x++)
	{
		for (int j=0; j<factor; j++)
		{
       		temp[(int)((x-min)*factor) + j] = truths[x];
        }
    }

    double _min  = mid - factor * (mid-min );
    double _mid1 = mid - factor * (mid-mid1);
    double _mid  = mid - factor * (mid-mid );
    double _mid2 = mid - factor * (mid-mid2);
    double _max  = mid - factor * (mid-max );
	min  = _min;
	mid1 = _mid1;
	mid  = _mid;
	mid2 = _mid2;
    max  = _max;

	for (int x=0; x<DOMAIN_POINTS; x++)
	{
        if ( (x>=min) && (x<=max) )
        {
        	truths[x] = temp[(int)(x-min)];
        }
        else
        {
        	truths[x] = ZERO_TRUTH;
        }
    }
}



void	TFuzzySet::scale_up  	(double factor)
{
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
    	if (truths[x] != ZERO_TRUTH)
        {
	    	truths[x] *= factor;
        }
    }
    minTruth *= factor;
    maxTruth *= factor;
}


void	TFuzzySet::scale_down	(double percent)
{
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
    	if (truths[x] != ZERO_TRUTH)
        {
	    	truths[x] *= percent;
        }
    }
    minTruth *= percent;
    maxTruth *= percent;
}

////////////////////////////////////////////////////////////////////////////////








////////////////////////////////////////////////////////////////////////////////
// logical operators (Zadeh and non_Zadeh)
////////////////////////////////////////////////////////////////////////////////
TFuzzySet&	TFuzzySet::and(TFuzzySet& set2, int type)
{
	TFuzzySet* result = new TFuzzySet(*this);
	switch(type)
   	{
		case ZADEH:			//					min(s1,s2)
        {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
		   		result->truths[x] = (truths[x] < set2.truths[x]) ? truths[x] : set2.truths[x];
			}
		    result->minTruth  = (minTruth < set2.minTruth) ? minTruth : set2.minTruth;
		    result->maxTruth  = (maxTruth < set2.maxTruth) ? maxTruth : set2.maxTruth;
        }
        break;
		case MEAN:			//      	    	(s1+s2)/2
        {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
		   		result->truths[x] = (truths[x] + set2.truths[x]) / 2.0;
			}
		    result->minTruth  = (minTruth + set2.minTruth) / 2.0;
		    result->maxTruth  = (maxTruth + set2.maxTruth) / 2.0;
        }
        break;
		case MEANSQUARE:	//         			((s1+s2)/2)^2
        {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
		   		result->truths[x] = pow((truths[x] + set2.truths[x]) / 2.0,  2);
			}
		    result->minTruth  = pow((minTruth + set2.minTruth) / 2.0, 2);
		    result->maxTruth  = pow((maxTruth + set2.maxTruth) / 2.0, 2);
        }
        break;
		case MEANROOT:		//					((s1+s2)/2)^.5
        {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
		   		result->truths[x] = pow((truths[x] + set2.truths[x]) / 2.0,  0.5);
			}
		    result->minTruth  = pow((minTruth + set2.minTruth) / 2.0, 0.5);
		    result->maxTruth  = pow((maxTruth + set2.maxTruth) / 2.0, 0.5);
        }
        break;
		case PRODUCT:		//						s1*s2
        {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
		   		result->truths[x] = truths[x] * set2.truths[x];
			}
		    result->minTruth  = minTruth * set2.minTruth;
		    result->maxTruth  = maxTruth * set2.maxTruth;
        }
        break;
		case BOUNDEDSUM:	//					max(0,(s1+s2-1))
        {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
		   		result->truths[x] = (MIN_TRUTH > (truths[x] + set2.truths[x] - 1)) ? 0.0 : (truths[x] + set2.truths[x] - 1);
			}
		    result->minTruth  = (MIN_TRUTH > (minTruth + set2.minTruth - 1)) ? 0.0 : (minTruth + set2.minTruth - 1);
		    result->maxTruth  = (MIN_TRUTH > (maxTruth + set2.maxTruth - 1)) ? 0.0 : (maxTruth + set2.maxTruth - 1);
        }
        break;
    }
	return *result;
}




TFuzzySet&	TFuzzySet::or(TFuzzySet& set2, int type)
{
	TFuzzySet* result = new TFuzzySet(*this);
	switch(type)
   	{
		case ZADEH:			//					max(s1,s2)
        {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
		   		result->truths[x] = (truths[x] > set2.truths[x]) ? truths[x] : set2.truths[x];
			}
		    result->minTruth  = (minTruth > set2.minTruth) ? minTruth : set2.minTruth;
		    result->maxTruth  = (maxTruth > set2.maxTruth) ? maxTruth : set2.maxTruth;
        }
        break;
		case MEAN:			//      	    	(2*min(s1,s2)+4*max(s1,s2))/6
        {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
		   		result->truths[x] = (
                    					2 * ((truths[x] < set2.truths[x]) ? truths[x] : set2.truths[x])
                    					+
										4 * ((truths[x] > set2.truths[x]) ? truths[x] : set2.truths[x])
                                    ) / 6.0;
			}
		    result->minTruth  = (
                    			2 * ((minTruth < set2.minTruth) ? minTruth : set2.minTruth)
                    			+
								4 * ((minTruth > set2.minTruth) ? minTruth : set2.minTruth)
                                ) / 6.0;
		    result->maxTruth  = (
                    			2 * ((maxTruth < set2.maxTruth) ? maxTruth : set2.maxTruth)
                    			+
								4 * ((maxTruth > set2.maxTruth) ? maxTruth : set2.maxTruth)
                                ) / 6.0;
        }
        break;
		case MEANSQUARE:	//         			((2*min(s1,s2)+4*max(s1,s2))/6)^2
        {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
		   		result->truths[x] = pow(((
                    					2 * ((truths[x] < set2.truths[x]) ? truths[x] : set2.truths[x])
                    					+
										4 * ((truths[x] > set2.truths[x]) ? truths[x] : set2.truths[x])
		                            ) / 6.0), 2);
			}
		    result->minTruth  = pow(((
                    			2 * ((minTruth < set2.minTruth) ? minTruth : set2.minTruth)
                    			+
								4 * ((minTruth > set2.minTruth) ? minTruth : set2.minTruth)
                                ) / 6.0), 2);
		    result->maxTruth  = pow(((
                    			2 * ((maxTruth < set2.maxTruth) ? maxTruth : set2.maxTruth)
                    			+
								4 * ((maxTruth > set2.maxTruth) ? maxTruth : set2.maxTruth)
                                ) / 6.0), 2);
        }
        break;
		case MEANROOT:		//					((2*min(s1,s2)+4*max(s1,s2))/6)^.5
        {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
		   		result->truths[x] = pow(((
                    					2 * ((truths[x] < set2.truths[x]) ? truths[x] : set2.truths[x])
                    					+
										4 * ((truths[x] > set2.truths[x]) ? truths[x] : set2.truths[x])
                                    ) / 6.0), 0.5);
			}
		    result->minTruth  = pow(((
                    			2 * ((minTruth < set2.minTruth) ? minTruth : set2.minTruth)
                    			+
								4 * ((minTruth > set2.minTruth) ? minTruth : set2.minTruth)
                                ) / 6.0), 0.5);
		    result->maxTruth  = pow(((
                    			2 * ((maxTruth < set2.maxTruth) ? maxTruth : set2.maxTruth)
                    			+
								4 * ((maxTruth > set2.maxTruth) ? maxTruth : set2.maxTruth)
                                ) / 6.0), 0.5);
        }
        break;
		case PRODUCT:		//					(s1+s2)-(s1*s2)
        {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
		   		result->truths[x] = (truths[x] + set2.truths[x]) - (truths[x] * set2.truths[x]);
			}
		    result->minTruth  = (minTruth + set2.minTruth) - (minTruth * set2.minTruth);
		    result->maxTruth  = (maxTruth + set2.maxTruth) - (maxTruth * set2.maxTruth);
        }
        break;
		case BOUNDEDSUM:	//					min(1,(s1+s2))
        {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
		   		result->truths[x] = (MAX_TRUTH < (truths[x] + set2.truths[x])) ? 1.0 : (truths[x] + set2.truths[x]);
			}
		    result->minTruth  = (MAX_TRUTH < (minTruth + set2.minTruth)) ? 1.0 : (minTruth + set2.minTruth);
		    result->maxTruth  = (MAX_TRUTH < (maxTruth + set2.maxTruth)) ? 1.0 : (maxTruth + set2.maxTruth);
        }
        break;
    }
	return *result;
}





TFuzzySet&	TFuzzySet::not(int type)
{
	TFuzzySet* result = new TFuzzySet(*this);
	switch(type)
   	{
		case ZADEH:			//					1-s1
		case MEAN:			//      	    	1-s1
		case MEANSQUARE:	//         			1-s1
		case MEANROOT:		//					1-s1
		case PRODUCT:		//					1-s1
		case BOUNDEDSUM:	//                  1-s1
        {
			for (int x=0; x<DOMAIN_POINTS; x++)
			{
   				result->truths[x] = MAX_TRUTH - truths[x];
			}
		    result->minTruth  = MAX_TRUTH - maxTruth;
		    result->maxTruth  = MAX_TRUTH - minTruth;
		         if (direction == UP)   result->direction = DOWN;
		    else if (direction == DOWN) result->direction = UP;
        }
        break;
    }
	return *result;
}


////////////////////////////////////////////////////////////////////////////////









////////////////////////////////////////////////////////////////////////////////
// operators
////////////////////////////////////////////////////////////////////////////////

bool       TFuzzySet::operator == 		(TFuzzySet& set2)   		 	  		// comparison
{
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
   		if (this->truths[x] != set2.truths[x]) return false;
	}
	return true;
}

bool       TFuzzySet::operator != 		(TFuzzySet& set2)   		 	  		// comparison
{
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
   		if (this->truths[x] != set2.truths[x]) return true;
	}
	return false;
}

bool	   TFuzzySet::operator <  		(TFuzzySet& set2)   		   			// comparison
{
    double integral1 = 0.0;
    double integral2 = 0.0;
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
   		integral1 += this->truths[x];
   		integral2 += set2.truths[x];
	}
	return (integral1 < integral2);
}

bool	   TFuzzySet::operator >  		(TFuzzySet& set2)   		   			// comparison
{
    double integral1 = 0.0;
    double integral2 = 0.0;
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
   		integral1 += this->truths[x];
   		integral2 += set2.truths[x];
	}
	return (integral1 > integral2);
}

bool	   TFuzzySet::operator <=  		(TFuzzySet& set2)   		   			// comparison
{
	return !(*this > set2);
}

bool	   TFuzzySet::operator >=  		(TFuzzySet& set2)   		   			// comparison
{
	return !(*this < set2);
}

TFuzzySet& TFuzzySet::operator =  		(TFuzzySet& set2)      					// assignment
{
	set_variable(set2.get_variable());
    if (set2.get_name() != NULL)
   	{
   		name	= new char[strlen(set2.get_name())+1];
        strcpy(name, set2.get_name());
   	}
	set_shape(set2.get_shape());
	set_direction(set2.get_direction());
	set_min (set2.get_min());
	set_mid1(set2.get_mid1());
	set_mid (set2.get_mid());
	set_mid2(set2.get_mid2());
	set_max (set2.get_max());
	set_minTruth(set2.get_minTruth());
	set_maxTruth(set2.get_maxTruth());
	for(int i=0; i<MAX_HEDGES; i++)    	hedges[i] = set2.hedges[i];
	for(int i=0; i<DOMAIN_POINTS; i++) 	truths[i] = set2.truths[i];
	set_alphaCut(set2.get_alphaCut());
	return *this;
}



TFuzzySet& 		 	  operator +  		(TFuzzySet& set1, TFuzzySet& set2)		// WARNING: maxTruth can be > MAX_TRUTH
{
	TFuzzySet* result = new TFuzzySet(set1);
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
   		result->truths[x] = set1.truths[x] + set2.truths[x];
	}
    result->minTruth  = set1.minTruth + set2.minTruth;
    result->maxTruth  = set1.maxTruth + set2.maxTruth;
	return *result;
}

TFuzzySet& 		 	  operator +  		(TFuzzySet& set1, double shift)			// shift-up
{
	TFuzzySet* result = new TFuzzySet(set1);
    result->shift_up(shift);
	return *result;
}

TFuzzySet& 		 	  operator +  		(double shift, TFuzzySet& set1)			// shift-up
{
	TFuzzySet* result = new TFuzzySet(set1);
    result->shift_up(shift);
	return *result;
}


TFuzzySet& 		 	  operator -  		(TFuzzySet& set1, TFuzzySet& set2)		// WARNING: maxTruth can be > MAX_TRUTH
{
	TFuzzySet* result = new TFuzzySet(set1);
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
   		result->truths[x] = set1.truths[x] - set2.truths[x];
	}
    result->minTruth  = set1.minTruth - set2.minTruth;
    result->maxTruth  = set1.maxTruth - set2.maxTruth;
	return *result;
}

TFuzzySet& 		 	  operator -  		(TFuzzySet& set1, double shift)			// shift-down
{
	TFuzzySet* result = new TFuzzySet(set1);
    result->shift_down(shift);
	return *result;
}

TFuzzySet& 		 	  operator -  		(double shift, TFuzzySet& set1)			// shift-down
{
	TFuzzySet* result = new TFuzzySet(set1);
    result->shift_down(shift);
	return *result;
}


TFuzzySet& 		 	  operator *  		(TFuzzySet& set1, TFuzzySet& set2)		// WARNING: maxTruth can be > MAX_TRUTH
{
	TFuzzySet* result = new TFuzzySet(set1);
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
   		result->truths[x] = set1.truths[x] * set2.truths[x];
	}
    result->minTruth  = set1.minTruth * set2.minTruth;
    result->maxTruth  = set1.maxTruth * set2.maxTruth;
	return *result;
}

TFuzzySet& 			  operator *  		(TFuzzySet& set1, double factor)		// scale up
{
	TFuzzySet* result = new TFuzzySet(set1);
    result->scale_up(factor);
	return *result;
}

TFuzzySet& 			  operator *  		(double factor, TFuzzySet& set1)		// scale up
{
	TFuzzySet* result = new TFuzzySet(set1);
    result->scale_up(factor);
	return *result;
}


TFuzzySet& 		 	  operator /  		(TFuzzySet& set1, TFuzzySet& set2)		//
{
	TFuzzySet* result = new TFuzzySet(set1);
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
   		result->truths[x] = set1.truths[x] / set2.truths[x];
	}
    result->minTruth  = set1.minTruth / set2.minTruth;
    result->maxTruth  = set1.maxTruth / set2.maxTruth;
	return *result;
}

TFuzzySet& 			  operator /  		(TFuzzySet& set1, double percent)		// scale down
{
	TFuzzySet* result = new TFuzzySet(set1);
    result->scale_down(percent);
	return *result;
}

TFuzzySet& 			  operator /  		(double percent, TFuzzySet& set1)		// scale down
{
	TFuzzySet* result = new TFuzzySet(set1);
    result->scale_down(percent);
	return *result;
}


TFuzzySet& 		 	  operator ^  		(TFuzzySet& set1, double power)			// hedge (dilute or intensify)
{
	TFuzzySet* result = new TFuzzySet(set1);
	for (int x=0; x<DOMAIN_POINTS; x++)
	{
   		result->truths[x] = pow(set1.truths[x], power);
	}
    result->minTruth  = pow(set1.minTruth, power);
    result->maxTruth  = pow(set1.maxTruth, power);
	return *result;
}


TFuzzySet& 		 	  operator && 		(TFuzzySet& set1, TFuzzySet& set2)  	// and
{
	return set1.and(set2);
}

TFuzzySet& 		 	  operator || 		(TFuzzySet& set1, TFuzzySet& set2)		// or
{
	return set1.or(set2);
}

TFuzzySet& 		 	  operator !  		(TFuzzySet& set1)						// not
{
	return set1.not();
}


double	   TFuzzySet::operator [] 		(double x)								// truth(double x)
{
	return truths[(int)x];
}


		   TFuzzySet::operator double 	()  									// defuzzify(CENTROID)
{
	return defuzzify();
}


ostream&   			  operator <<		(ostream& s, TFuzzySet& set1)			// save
{
	s << set1.variable->name  << " ";
	s << set1.variable->unit  << " ";
	s << set1.variable->begin << " ";
	s << set1.variable->end   << " ";
	s << set1.name      << " ";
    s << set1.shape     << " ";
	s << set1.direction << " ";
    s << set1.min       << " ";
	s << set1.mid1      << " ";
    s << set1.mid       << " ";
	s << set1.mid2      << " ";
    s << set1.max       << " ";
	s << set1.minTruth  << " ";
    s << set1.maxTruth  << " ";
	s << set1.alphaCut  << " ";
   	s << endl;

	for (int x=0; x<DOMAIN_POINTS; x++)
	{
   		s << set1.truths[x] << " ";
	}
    s << endl;

	for (int x=0; x<MAX_HEDGES; x++)
	{
   		s << set1.hedges[x] << " ";
	}
    s << endl;

	return s;
}



istream&   			  operator >>		(istream& s, TFuzzySet& set1)			// load
{
    char variable_name[256];	s >> variable_name; set1.variable->set_name(variable_name);
    char variable_unit[256];	s >> variable_unit; set1.variable->set_unit(variable_unit);
    s >> set1.variable->begin;
    s >> set1.variable->end;
    char name[256];				s >> name; 			set1.set_name(name);
	s >> set1.shape;
	s >> set1.direction;
    s >> set1.min;
	s >> set1.mid1;
    s >> set1.mid;
	s >> set1.mid2;
    s >> set1.max;
	s >> set1.minTruth;
	s >> set1.maxTruth;
    s >> set1.alphaCut;

	for (int x=0; x<DOMAIN_POINTS; x++)
	{
   		s >> set1.truths[x];
	}

	for (int x=0; x<MAX_HEDGES; x++)
	{
   		s >> set1.hedges[x];
	}

	return s;
}

////////////////////////////////////////////////////////////////////////////////












////////////////////////////////////////////////////////////////////////////////
// graphics
////////////////////////////////////////////////////////////////////////////////

void TFuzzySet::Draw(TPaintBox* PaintBox)
{
	double XScale = PaintBox->Width/DOMAIN_WIDTH;
	double YScale = PaintBox->Height/MAX_TRUTH;
	for (int x=0; x<DOMAIN_POINTS-1; x++)
   	{
		PaintBox->Canvas->MoveTo((int)(x    *XScale), (int)((MAX_TRUTH - truths[x  ])*YScale));
		PaintBox->Canvas->LineTo((int)((x+1)*XScale), (int)((MAX_TRUTH - truths[x+1])*YScale));
    }
   	PaintBox->Canvas->Refresh();
}

////////////////////////////////////////////////////////////////////////////////







