//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <values.h>
#include <math.h>
#pragma hdrstop

#include "TLink.h"
#include "TNetwork.h"
//---------------------------------------------------------------------------

#include "TNode.h"

	int TLink::scale 	 	= 20;
	int TLink::serial_num	= 0;

    TLink::TLink(TLink* _link)
	{
      if (_link != NULL)
      {
	    set_id(_link->get_id());
    	Node_a     = NULL;
    	Node_b     = NULL;
	    set_type(_link->get_type());
		set_initial_delay(_link->get_initial_delay());

		set_capacity(_link->get_capacity());

		selected = _link->IsSelected();
		set_utilization(_link->get_utilization());
      }
	}


    TLink::TLink(TNetwork* _Network, TNode* _node_a, TNode* _node_b)
	{
		serial_num++;

		Network		= _Network;
	    set_id(serial_num);
    	Node_a     = _node_a;
    	if (Node_a != NULL)
        {
        	Node_a->Links->Add(this);
        }
    	Node_b     = _node_b;
    	if (Node_b != NULL)
        {
        	Node_b->Links->Add(this);
        }

        Node_a->update_connectivity();
        Node_b->update_connectivity();

	    if(random(2))
        {
	    	set_type(LAN);
			set_initial_delay(1+random(9));
        }
        else
        {
	    	set_type(WAN);
            set_initial_delay(10+random(40));
		}

        set_capacity(100);

		selected = false;

		set_utilization(0);
	}





    TLink::~TLink()
	{
		DeSelect();

    	if (Node_a != NULL)
        {
        	Node_a->Links->Remove(this);
			Node_a->update_connectivity();
        }

    	if (Node_a != NULL)
        {
	    	Node_b->Links->Remove(this);
			Node_b->update_connectivity();
        }
	}







	void TLink::set_utilization(float _utilization)
    {
       	utilization	= _utilization;
        //if (utilization <   0) utilization = 0;
        //if (utilization > 100) utilization = 100;
		if (utilization < 50)	//  0 .. 50
        {
         	delay = initial_delay;
	        color	= (TColor) RGB(
           							255 - (utilization*2)*2.55,
                           	        255 - (utilization*2)*2.55,
                               	    255
           						  );
        }
        else	//  50 .. 100
        {
         	delay = initial_delay + initial_delay * (utilization/100.0)  ;
	        color	= (TColor) RGB(
           							0,
                           	        0,
                               	    ((100-utilization)*2)*2.55
           						  );
        }

		if (utilization > 100)	// Overflow
        {
         	delay = 999;        // Help
	        color	= (TColor) RGB(
           							255,
                           	        0,
                               	    0
           						  );
        }

		if (!selected)
        {
       		selection_color = color;
	    }
    }




	void TLink::set_capacity(int _capacity)
    {
       	capacity	= _capacity;
        width 		= capacity/scale;
    }




	void TLink::Draw(TCanvas* _canvas)
	{
		_canvas->Pen->Color   = selection_color;
		_canvas->Brush->Color	= color;
		TPoint* selection_boundary = calculateSelectionBoundary();
        _canvas->Polygon(selection_boundary, 3);


        int x0 = selection_boundary[0].x;
        int x1 = selection_boundary[1].x;
        //int x2 = selection_boundary[2].x;
        //int x3 = selection_boundary[3].x;
        int y0 = selection_boundary[0].y;
        int y1 = selection_boundary[1].y;
        //int y2 = selection_boundary[2].y;
        //int y3 = selection_boundary[3].y;
        int x = x0 + (x1 - x0)/2;
        int y = y0 + (y1 - y0)/2;
//
        x = x - 5;
//
		_canvas->Brush->Color = cl3DLight;
		if (selected)
        {
			_canvas->Font->Color = selection_color;
        }
   		if (Network->info_type != 0)
        {
			_canvas->Font->Color = clGreen;
            _canvas->TextOut(x, y, IntToStr((int)delay));
        }
        else
        {
			_canvas->Font->Color = clBlue;
            _canvas->TextOut(x, y, IntToStr((int)id));
        }

        _canvas->Refresh();
    }

	void TLink::DrawDetails(TCanvas* _canvas)
	{
    }


	void TLink::DrawCaption(TCanvas* _canvas)
	{
		_canvas->Pen->Color = clInfoBk;
		_canvas->Brush->Color = clInfoBk;
		_canvas->Rectangle(0, 0, 300, 15);

		_canvas->Font->Name  = "MS Sans Serif";
		_canvas->Font->Size  = 8;
		_canvas->Font->Color = clBlue;
		_canvas->Brush->Color = clInfoBk;
   		_canvas->TextOut(10, 0, IntToStr((int)id));
   		_canvas->TextOut(42, 0, IntToStr((int)capacity));
   		_canvas->TextOut(98, 0, IntToStr((int)utilization));
   		_canvas->TextOut(145, 0, IntToStr((int)type)==0 ? "LAN" : "WAN");
   		_canvas->TextOut(192, 0, IntToStr((int)delay));
        _canvas->Refresh();
    }


	bool TLink::IsSelected()
	{
	   	return selected;
    }


    void TLink::Select(TColor _selection_color)
	{
        Network->set_selected_link_index(Network->Links->IndexOf(this));
    	selection_color = _selection_color;
   		selected = true;
    }


    void TLink::DeSelect()
	{
	    if (selected)
        {
            Network->set_selected_link_index(-1);
	    	selection_color = color;
   			selected = false;
        }
    }





///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

TPoint* TLink::calculateSelectionBoundary()
{
   if( (Node_a != NULL) && (Node_a != NULL) )
   {
      int W  = width;
      int x1 = Node_a->get_position().x;
      int y1 = Node_a->get_position().y;
      int x2 = Node_b->get_position().x;
      int y2 = Node_b->get_position().y;

      float dx = (x2 - x1);
      float dy = (y2 - y1);
      float D = (float)sqrt(dx*dx + dy*dy);

      float COS;
      float SIN;
      if (D != 0)
      {
    	COS = dx/D;
    	SIN = dy/D;
      }
      else
      {
    	COS = 0;
    	SIN = 0;
      }

      float A_SIZE = Node_a->get_radius();
      float B_SIZE = Node_b->get_radius();
      float AX, AY, BX, BY;  // border coordinates to be calculated
      AX = x1 + A_SIZE*COS;
      AY = y1 + A_SIZE*SIN;
      BX = x2 - B_SIZE*COS;
      BY = y2 - B_SIZE*SIN;


      // leading .._ / lagging _.. region coordinates for node A
      float AX_ = AX + W*SIN;
      float AY_ = AY - W*COS;
      float _AX = AX - W*SIN;
      float _AY = AY + W*COS;
      // signs are inverted for  ( A--->   &    ---->B ) both on one side of link
      // and                     ( A<---   &    <----B ) both on the other side
      // leading .._ / lagging _.. region coordinates for node B
      float BX_ = BX - W*SIN;
      float BY_ = BY + W*COS;
      float _BX = BX + W*SIN;
      float _BY = BY - W*COS;


      float DX = (BX - AX);
      float DY = (BY - AY);
      float Distance = (float)sqrt(DX*DX + DY*DY);



      //     1           5           7            3
      //     |-----A-----|----MID----|-----B------|
      //     2           6           8            4


      // Find 1/3 distances from node A
      float AX13 = AX + (.333f*Distance)*COS;
      float AY13 = AY + (.333f*Distance)*SIN;
      // leading .._ / lagging _.. 1/3 region coordinates for node A
      float AX13_ = AX13 + W*SIN;
      float AY13_ = AY13 - W*COS;
      float _AX13 = AX13 - W*SIN;
      float _AY13 = AY13 + W*COS;

      // Find 2/3 distances from node A
      float AX23 = AX + (.666f*Distance)*COS;
      float AY23 = AY + (.666f*Distance)*SIN;
      // leading .._ / lagging _.. 2/3 region coordinates for node A
      float AX23_ = AX23 + W*SIN;
      float AY23_ = AY23 - W*COS;
      float _AX23 = AX23 - W*SIN;
      float _AY23 = AY23 + W*COS;

      /*
      // Find 3/3 distances from node A
      float AX33 = AX + (.999f*Distance)*COS;
      float AY33 = AY + (.999f*Distance)*SIN;
      // leading .._ / lagging _.. 3/3 region coordinates for node A
      float AX33_ = AX33 + W*SIN;
      float AY33_ = AY33 - W*COS;
      float _AX33 = AX33 - W*SIN;
      float _AY33 = AY33 + W*COS;
      */

      TPoint p1, p2, p3, p4, p5, p6, p7, p8;
	  p1.x = (int)AX_;
	  p1.y = (int)AY_;
	  p2.x = (int)_AX;
	  p2.y = (int)_AY;
	  p3.x = (int)_BX;
	  p3.y = (int)_BY;
	  p4.x = (int)BX_;
	  p4.y = (int)BY_;
	  p5.x = (int)AX13_;
	  p5.y = (int)AY13_;
	  p6.x = (int)_AX13;
	  p6.y = (int)_AY13;
	  p7.x = (int)AX23_;
	  p7.y = (int)AY23_;
	  p8.x = (int)_AX23;
	  p8.y = (int)_AY23;

/*
      //     1           5
      //     |-----A-----|
      //     2           6
	  TPoint* points1562 = new TPoint[4];
	  points1562[0] = p1;
	  points1562[1] = p5;
	  points1562[2] = p6;
	  points1562[3] = p2;

      //                 5           7
      //                 |----MID----|
      //                 6           8
	  TPoint* points5786 = new TPoint[4];
	  points5786[0] = p5;
	  points5786[1] = p7;
	  points5786[2] = p8;
	  points5786[3] = p6;

      //                             7            3
      //                             |-----B------|
      //                             8            4
	  TPoint* points7438 = new TPoint[4];
	  points7438[0] = p7;
	  points7438[1] = p4;
	  points7438[2] = p3;
	  points7438[3] = p8;
*/
      //     1                                    3
      //     |------------------------------------|
      //     2                                    4
	  TPoint* points1342 = new TPoint[4];
	  points1342[0] = p1;
	  points1342[1] = p3;
	  points1342[2] = p4;
	  points1342[3] = p2;

      return points1342;
   }
   else
   {
      return NULL;
   }
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////













///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

	bool TLink::Contains(TPoint _p)
	{
		TPoint* points = calculateSelectionBoundary();
	    int npoints   = 4;
	    int xpoints[] = {points[0].x, points[1].x, points[2].x, points[3].x};
    	int ypoints[] = {points[0].y, points[1].y, points[2].y, points[3].y};

        if (insideRect(calculateBounds(xpoints, ypoints, npoints), Point(_p.x, _p.y)))
        {
            int hits = 0;
            int ySave = 0;

            // Find a vertex that's not on the halfline
            int i = 0;
            while (i < npoints && ypoints[i] == _p.y)
            {
                i++;
		    }

    	    // Walk the edges of the polygon
        	for (int n = 0; n < npoints; n++)
	        {
                int j = (i + 1) % npoints;

                int dx = xpoints[j] - xpoints[i];
                int dy = ypoints[j] - ypoints[i];

                // Ignore horizontal edges completely
                if (dy != 0)
                {
                    // Check to see if the edge intersects
                    // the horizontal halfline through (x, y)
                    int rx = _p.x - xpoints[i];
                    int ry = _p.y - ypoints[i];

                    // Deal with edges starting or ending on the halfline
                    if (ypoints[j] == _p.y && xpoints[j] >= _p.x)
                    {
                        ySave = ypoints[i];
				    }
                    if (ypoints[i] == _p.y && xpoints[i] >= _p.x)
                    {
                        if ((ySave > _p.y) != (ypoints[j] > _p.y))
                        {
						    hits--;
						}
				    }

                    // Tally intersections with halfline
                    float s = (float)ry / (float)dy;
                    if (s >= 0.0 && s <= 1.0 && (s * dx) >= rx)
                    {
                        hits++;
		    		}
                }
                i = j;
            }

            // Inside if number of intersections odd
            return (hits % 2) != 0;
        }
        return false;
    }





    TRect* TLink::calculateBounds(int xpoints[], int ypoints[], int npoints)
    {
		int boundsMinX = MAXINT;
		int boundsMinY = MAXINT;
		int boundsMaxX = 0;
		int boundsMaxY = 0;

		for (int i = 0; i < npoints; i++) {
		    int x = xpoints[i];
		    boundsMinX = (boundsMinX < x) ? boundsMinX : x;
		    boundsMaxX = (boundsMaxX > x) ? boundsMaxX : x;
		    int y = ypoints[i];
		    boundsMinY = (boundsMinY < y) ? boundsMinY : y;
		    boundsMaxY = (boundsMaxY > y) ? boundsMaxY : y;
		}

		TRect* r = new TRect();
		r->Left	 = boundsMinX;
		r->Top	 = boundsMinY;
		r->Right = boundsMaxX;
		r->Bottom = boundsMaxY;
        return r;
    }



    bool TLink::insideRect(TRect* r, TPoint p)
    {
		return (p.x >= r->Left) && ((p.x - r->Left) < (r->Right - r->Left)) && (p.y >= r->Top) && ((p.y - r->Top) < (r->Bottom - r->Top));
    }



///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

