//---------------------------------------------------------------------------
#include "TMigrationFC.h"
//---------------------------------------------------------------------------

#define MIGRATION_RECOMMECNDATION_FACTOR   	0.85        // 85 main criterion
#define MIGRATION_LINKS_DELAY_FACTOR 	   	0.00		// 00 only important in replication to divert traffic away
#define MIGRATION_NODE_CONNECTIVITY_FACTOR	0.10        // 15 important
#define MIGRATION_COST_FACTOR			   	0.00        // 00 fragment size covered in mobility


// static members, set out side constuctor
double TMigrationFC::CandidacyThreshold	= 33.0;		// 33.0 minmum level to justfy consideration    // 05 statistically significant
double TMigrationFC::MigrationThreshold	= 50.0;		// 50.0 minmum level to justfy migration


////////////////////////////////////////////////////////////////////////////////
TMigrationFC::TMigrationFC(TFragment* _Fragment)
{
	Fragment = _Fragment;

	C1 = NULL;
	C2 = NULL;
	C3 = NULL;

	C1 = new TFuzzyController();    // attraction
	C2 = new TFuzzyController();    // repulsion
	C3 = new TFuzzyController();    // force

    // needed for Shape Function setup
    C1Input1Centre = 50;  C1Input2Centre = 50;  C1OutputCentre = 50;   // demand = linear increase
    C2Input1Centre = 75;  C2Input2Centre = 75;  C2OutputCentre = 50;   // load   = slow to increase
    C3OutputCentre = 75;	// scale-up output before mobility drags it down
    A4 = 0.33;				// output = input * (mobility ^ A4) = fast rise

    // static members, set out side constuctor
    //CandidacyThreshold	= 33.0;		// 33.0 minmum level to justfy consideration    // 05 statistically significant
    //MigrationThreshold	= 50.0;		// 50.0 minmum level to justfy migratation

    // needed for Rules setup
/*  too bad, it made high migration all the time, 30, 18, 14, 16, 18, 16, ...
    C1r1out = 0;  C1r1alpha = 0.0;  C1r1weight = 1.0;
    C1r2out = 0;  C1r2alpha = 0.0;  C1r2weight = 1.0;
    C1r3out = 0;  C1r3alpha = 0.0;  C1r3weight = 1.0;
    C1r4out = 2;  C1r4alpha = 0.0;  C1r4weight = 1.0;
    C1r5out = 1;  C1r5alpha = 0.0;  C1r5weight = 1.0;
    C1r6out = 0;  C1r6alpha = 0.0;  C1r6weight = 1.0;
    C1r7out = 2;  C1r7alpha = 0.0;  C1r7weight = 1.0;
    C1r8out = 2;  C1r8alpha = 0.0;  C1r8weight = 1.0;
    C1r9out = 1;  C1r9alpha = 0.0;  C1r9weight = 1.0;
*/
//  very good, migrations actions   25, 12, 8, 5, 3, 0, 1, 0, 1, ...
    C1r1out = 0;  C1r1alpha = 0.0;  C1r1weight = 1.0;
    C1r2out = 0;  C1r2alpha = 0.0;  C1r2weight = 1.0;
    C1r3out = 0;  C1r3alpha = 0.0;  C1r3weight = 1.0;
    C1r4out = 1;  C1r4alpha = 0.0;  C1r4weight = 1.0;
    C1r5out = 0;  C1r5alpha = 0.0;  C1r5weight = 1.0;
    C1r6out = 0;  C1r6alpha = 0.0;  C1r6weight = 1.0;
    C1r7out = 2;  C1r7alpha = 0.0;  C1r7weight = 1.0;
    C1r8out = 1;  C1r8alpha = 0.0;  C1r8weight = 1.0;
    C1r9out = 0;  C1r9alpha = 0.0;  C1r9weight = 1.0;

    C2r1out = 0;  C2r1alpha = 0.0;  C2r1weight = 1.0;
    C2r2out = 0;  C2r2alpha = 0.0;  C2r2weight = 1.0;
    C2r3out = 0;  C2r3alpha = 0.0;  C2r3weight = 1.0;
    C2r4out = 1;  C2r4alpha = 0.0;  C2r4weight = 1.0;
    C2r5out = 0;  C2r5alpha = 0.0;  C2r5weight = 1.0;
    C2r6out = 0;  C2r6alpha = 0.0;  C2r6weight = 1.0;
    C2r7out = 2;  C2r7alpha = 0.0;  C2r7weight = 1.0;
    C2r8out = 1;  C2r8alpha = 0.0;  C2r8weight = 1.0;
    C2r9out = 0;  C2r9alpha = 0.0;  C2r9weight = 1.0;

    C3r1out = 0;  C3r1alpha = 0.0;  C3r1weight = 1.0;
    C3r2out = 1;  C3r2alpha = 0.0;  C3r2weight = 1.0;
    C3r3out = 2;  C3r3alpha = 0.0;  C3r3weight = 1.0;
    C3r4out = 1;  C3r4alpha = 0.0;  C3r4weight = 1.0;
    C3r5out = 2;  C3r5alpha = 0.0;  C3r5weight = 1.0;
    C3r6out = 2;  C3r6alpha = 0.0;  C3r6weight = 1.0; // more biased toward first param
    C3r7out = 2;  C3r7alpha = 0.0;  C3r7weight = 1.0;
    C3r8out = 3;  C3r8alpha = 0.0;  C3r8weight = 1.0;
    C3r9out = 4;  C3r9alpha = 0.0;  C3r9weight = 1.0;


	num_of_demanding_nodes = 0;
	num_of_candidate_nodes = 0;

	// inputs
    local_demand = 0.0;
    for (int i=0; i<MAX_DEMANDING_NODES; i++)
    {
        remote_demands[i] = 0.0;
    }
    local_load = 0.0;
    for (int i=0; i<MAX_CANDIDATE_NODES; i++)
    {
        remote_loads[i] = 0.0;
    }
    mobility = 0.0;

	// outputs
    for (int i=0; i<MAX_DEMANDING_NODES; i++)
    {
        demanding_nodes_ids[i] = INVALID_ID;
    }
    for (int i=0; i<MAX_CANDIDATE_NODES; i++)
    {
        candidate_nodes_ids[i] = INVALID_ID;
    }
    for (int i=0; i<MAX_DEMANDING_NODES; i++)
    {
        remote_attractionsets[i] = NULL;
    }
    for (int i=0; i<MAX_DEMANDING_NODES; i++)
    {
        remote_attractions[i] = 0.0;
    }
    for (int i=0; i<MAX_CANDIDATE_NODES; i++)
    {
        local_repulsionsets[i] = NULL;
    }
    for (int i=0; i<MAX_CANDIDATE_NODES; i++)
    {
        local_repulsions[i] = 0.0;
    }
    for (int i=0; i<MAX_CANDIDATE_NODES; i++)
    {
        forcesets[i] = NULL;
    }
    for (int i=0; i<MAX_CANDIDATE_NODES; i++)
    {
        forces[i] = 0.0;
    }
    for (int i=0; i<MAX_CANDIDATE_NODES; i++)
    {
        migrations[i] = 0.0;
    }
    migration_node_id = INVALID_ID;

///////////////////////
	SetupControllers();
///////////////////////

}
////////////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////////////
TMigrationFC::~TMigrationFC()
{
	Fragment = NULL;
    if (C1 != NULL)	delete	C1;    C1 = NULL;
    if (C2 != NULL)	delete	C2;    C2 = NULL;
    if (C3 != NULL)	delete	C3;    C3 = NULL;
}
////////////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////////////
void TMigrationFC::SetupControllers()
{
  // rules predicate (if) side
  int Low		= 0;
  int Medium	= 1;
  int High		= 2;


  // setup "remote node's attraction force on fragmnet" evaluator

  if (C1 != NULL)
  {	// add input and output variables
  	C1->clear_input_variables();         //  name,         unit, begin,  end
	C1->add_input_variable(new TVariable("RemoteDemand",     "",  00.0, 100.0));
	C1->add_input_variable(new TVariable("LocalDemand",      "",  00.0, 100.0));
	C1->output_variable =  new TVariable("RemoteAttraction", "",  00.0, 100.0);

	// add fuzzysets to the input and output variables

    centre = C1Input1Centre; // %
    begin = C1->input_variables[0]->begin;
    end   = C1->input_variables[0]->end;
    ratio = ( centre / ((end-begin) - centre) );
    // areas
    // end-begin = a        + b;
    //           = ratio*b  + b;
    //           = b * (ratio + 1);
    //    b      = (end-begin)/(ratio + 1);

    // width      percent
        b     = (end-begin)/(ratio + 1);
        a     = ratio*b;
    p0 = begin;
    p1 = p0 + a;
    p2 = p1 + b;
  	C1->input_variables[0]->clear_fuzzysets();
    C1->input_variables[0]->add_fuzzyset(new TFuzzySet(C1->input_variables[0], "Low", 	    S_CURVE  , DOWN , p0, -0.0       , (p0+p1)/2.0, -0.0        , p1));
    C1->input_variables[0]->add_fuzzyset(new TFuzzySet(C1->input_variables[0], "Medium", 	PI_CURVE , UP   , p0, (p0+p1)/2.0, p1         , (p1+p2)/2.0 , p2));
    C1->input_variables[0]->add_fuzzyset(new TFuzzySet(C1->input_variables[0], "High",   	S_CURVE  , UP   , p1, -0.0       , (p1+p2)/2.0, -0.0        , p2));


    centre = C1Input2Centre; // %
    begin = C1->input_variables[1]->begin;
    end   = C1->input_variables[1]->end;
    ratio = ( centre / ((end-begin) - centre) );
    // areas
    // end-begin = a        + b;
    //           = ratio*b  + b;
    //           = b * (ratio + 1);
    //    b      = (end-begin)/(ratio + 1);

    // width      percent
        b     = (end-begin)/(ratio + 1);
        a     = ratio*b;
    p0 = begin;
    p1 = p0 + a;
    p2 = p1 + b;
  	C1->input_variables[1]->clear_fuzzysets();
    C1->input_variables[1]->add_fuzzyset(new TFuzzySet(C1->input_variables[1], "Low", 	    S_CURVE  , DOWN , p0, -0.0        , (p0+p1)/2.0, -0.0        , p1));
    C1->input_variables[1]->add_fuzzyset(new TFuzzySet(C1->input_variables[1], "Medium", 	PI_CURVE , UP   , p0, (p0+p1)/2.0 , p1          , (p1+p2)/2.0, p2));
    C1->input_variables[1]->add_fuzzyset(new TFuzzySet(C1->input_variables[1], "High",   	S_CURVE  , UP   , p1, -0.0        , (p1+p2)/2.0, -0.0        , p2));


    centre = C1OutputCentre; // %
    begin = C1->output_variable->begin;
    end   = C1->output_variable->end;
    ratio = ( centre / ((end-begin) - centre) );
    // areas
    // end-begin = a        + b;
    //           = ratio*b  + b;
    //           = b * (ratio + 1);
    //    b      = (end-begin)/(ratio + 1);

    // width      percent
        b     = (end-begin)/(ratio + 1);
        a     = ratio*b;
    p0 = begin;
    p1 = p0 + a;
    p2 = p1 + b;
  	C1->output_variable->clear_fuzzysets();
    C1->output_variable->add_fuzzyset(new TFuzzySet(C1->output_variable, "Low", 	S_CURVE  , DOWN , p0, -0.0        , (p0+p1)/2.0, -0.0        , p1));
    C1->output_variable->add_fuzzyset(new TFuzzySet(C1->output_variable, "Medium", 	PI_CURVE , UP   , p0, (p0+p1)/2.0 , p1          , (p1+p2)/2.0, p2));
    C1->output_variable->add_fuzzyset(new TFuzzySet(C1->output_variable, "High",   	S_CURVE  , UP   , p1, -0.0        , (p1+p2)/2.0, -0.0        , p2));




	// create empty rules    (conditional, alpha,  weight)

    C1->clear_rules();
    C1->add_rule(new TFuzzyRule(true, C1r1alpha, C1r1weight));
    C1->add_rule(new TFuzzyRule(true, C1r2alpha, C1r2weight));
    C1->add_rule(new TFuzzyRule(true, C1r3alpha, C1r3weight));
    C1->add_rule(new TFuzzyRule(true, C1r4alpha, C1r4weight));
    C1->add_rule(new TFuzzyRule(true, C1r5alpha, C1r5weight));
    C1->add_rule(new TFuzzyRule(true, C1r6alpha, C1r6weight));
    C1->add_rule(new TFuzzyRule(true, C1r7alpha, C1r7weight));
    C1->add_rule(new TFuzzyRule(true, C1r8alpha, C1r8weight));
    C1->add_rule(new TFuzzyRule(true, C1r9alpha, C1r9weight));

	// setup rules              remote_demand        local_demand      remote_attraction
    //----------------------------------------------------------------------------------
    //
    if (C1->rules[0] != NULL)
    {
    C1->rules[0]->predicate->clear_premises();
    C1->rules[0]->predicate->clear_operators();
    C1->rules[0]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[Low]));
    C1->rules[0]->predicate->add_operator(AND);
    C1->rules[0]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[Low]));
    C1->rules[0]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r1out]));
    }
    if (C1->rules[1] != NULL)
    {
    C1->rules[1]->predicate->clear_premises();
    C1->rules[1]->predicate->clear_operators();
    C1->rules[1]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[Low]));
    C1->rules[1]->predicate->add_operator(AND);
    C1->rules[1]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[Medium]));
    C1->rules[1]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r2out]));
    }
    if (C1->rules[2] != NULL)
    {
    C1->rules[2]->predicate->clear_premises();
    C1->rules[2]->predicate->clear_operators();
    C1->rules[2]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[Low]));
    C1->rules[2]->predicate->add_operator(AND);
    C1->rules[2]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[High]));
    C1->rules[2]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r3out]));
    }
    if (C1->rules[3] != NULL)
    {
    C1->rules[3]->predicate->clear_premises();
    C1->rules[3]->predicate->clear_operators();
    C1->rules[3]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[Medium]));
    C1->rules[3]->predicate->add_operator(AND);
    C1->rules[3]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[Low]));
    C1->rules[3]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r4out]));
    }
    if (C1->rules[4] != NULL)
    {
    C1->rules[4]->predicate->clear_premises();
    C1->rules[4]->predicate->clear_operators();
    C1->rules[4]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[Medium]));
    C1->rules[4]->predicate->add_operator(AND);
    C1->rules[4]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[Medium]));
    C1->rules[4]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r5out]));
    }
    if (C1->rules[5] != NULL)
    {
    C1->rules[5]->predicate->clear_premises();
    C1->rules[5]->predicate->clear_operators();
    C1->rules[5]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[Medium]));
    C1->rules[5]->predicate->add_operator(AND);
    C1->rules[5]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[High]));
    C1->rules[5]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r6out]));
    }
    if (C1->rules[6] != NULL)
    {
    C1->rules[6]->predicate->clear_premises();
    C1->rules[6]->predicate->clear_operators();
    C1->rules[6]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[High]));
    C1->rules[6]->predicate->add_operator(AND);
    C1->rules[6]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[Low]));
    C1->rules[6]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r7out]));
    }
    if (C1->rules[7] != NULL)
    {
    C1->rules[7]->predicate->clear_premises();
    C1->rules[7]->predicate->clear_operators();
    C1->rules[7]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[High]));
    C1->rules[7]->predicate->add_operator(AND);
    C1->rules[7]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[Medium]));
    C1->rules[7]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r8out]));
    }
    if (C1->rules[8] != NULL)
    {
    C1->rules[8]->predicate->clear_premises();
    C1->rules[8]->predicate->clear_operators();
    C1->rules[8]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[High]));
    C1->rules[8]->predicate->add_operator(AND);
    C1->rules[8]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[High]));
    C1->rules[8]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r9out]));
    }
  }



  // setup "local node's repulsion force on fragmnet" evaluator

  if (C2 != NULL)
  {
	// add input and output variables
  	C2->clear_input_variables();         //  name,       unit, begin,  end
	C2->add_input_variable(new TVariable("LocalLoad",      "",  00.0, 100.0));
	C2->add_input_variable(new TVariable("RemoteLoad",     "",  00.0, 100.0));
	C2->output_variable =  new TVariable("LocalRepulsion", "",  00.0, 100.0);


	// add fuzzysets to the input and output variables

    centre = C2Input1Centre; // %
    begin = C2->input_variables[0]->begin;
    end   = C2->input_variables[0]->end;
    ratio = ( centre / ((end-begin) - centre) );
    // areas
    // end-begin = a        + b;
    //           = ratio*b  + b;
    //           = b * (ratio + 1);
    //    b      = (end-begin)/(ratio + 1);

    // width      percent
        b     = (end-begin)/(ratio + 1);
        a     = ratio*b;
    p0 = begin;
    p1 = p0 + a;
    p2 = p1 + b;
  	C2->input_variables[0]->clear_fuzzysets();
    C2->input_variables[0]->add_fuzzyset(new TFuzzySet(C2->input_variables[0], "Low", 	    S_CURVE  , DOWN , p0, -0.0       , (p0+p1)/2.0, -0.0        , p1));
    C2->input_variables[0]->add_fuzzyset(new TFuzzySet(C2->input_variables[0], "Medium", 	PI_CURVE , UP   , p0, (p0+p1)/2.0, p1         , (p1+p2)/2.0 , p2));
    C2->input_variables[0]->add_fuzzyset(new TFuzzySet(C2->input_variables[0], "High",   	S_CURVE  , UP   , p1, -0.0       , (p1+p2)/2.0, -0.0        , p2));


    centre = C2Input2Centre; // %
    begin = C2->input_variables[1]->begin;
    end   = C2->input_variables[1]->end;
    ratio = ( centre / ((end-begin) - centre) );
    // areas
    // end-begin = a        + b;
    //           = ratio*b  + b;
    //           = b * (ratio + 1);
    //    b      = (end-begin)/(ratio + 1);

    // width      percent
        b     = (end-begin)/(ratio + 1);
        a     = ratio*b;
    p0 = begin;
    p1 = p0 + a;
    p2 = p1 + b;
  	C2->input_variables[1]->clear_fuzzysets();
    C2->input_variables[1]->add_fuzzyset(new TFuzzySet(C2->input_variables[1], "Low", 	    S_CURVE  , DOWN , p0, -0.0        , (p0+p1)/2.0, -0.0        , p1));
    C2->input_variables[1]->add_fuzzyset(new TFuzzySet(C2->input_variables[1], "Medium", 	PI_CURVE , UP   , p0, (p0+p1)/2.0 , p1          , (p1+p2)/2.0, p2));
    C2->input_variables[1]->add_fuzzyset(new TFuzzySet(C2->input_variables[1], "High",   	S_CURVE  , UP   , p1, -0.0        , (p1+p2)/2.0, -0.0        , p2));


    centre = C2OutputCentre; // %
    begin = C2->output_variable->begin;
    end   = C2->output_variable->end;
    ratio = ( centre / ((end-begin) - centre) );
    // areas
    // end-begin = a        + b;
    //           = ratio*b  + b;
    //           = b * (ratio + 1);
    //    b      = (end-begin)/(ratio + 1);

    // width      percent
        b     = (end-begin)/(ratio + 1);
        a     = ratio*b;
    p0 = begin;
    p1 = p0 + a;
    p2 = p1 + b;
  	C2->output_variable->clear_fuzzysets();
    C2->output_variable->add_fuzzyset(new TFuzzySet(C2->output_variable, "Low", 	S_CURVE  , DOWN , p0, -0.0        , (p0+p1)/2.0, -0.0        , p1));
    C2->output_variable->add_fuzzyset(new TFuzzySet(C2->output_variable, "Medium", 	PI_CURVE , UP   , p0, (p0+p1)/2.0 , p1          , (p1+p2)/2.0, p2));
    C2->output_variable->add_fuzzyset(new TFuzzySet(C2->output_variable, "High",   	S_CURVE  , UP   , p1, -0.0        , (p1+p2)/2.0, -0.0        , p2));



	// create empty rules    (conditional, alpha,  weight)
    C2->clear_rules();
    C2->add_rule(new TFuzzyRule(true, C2r1alpha, C2r1weight));
    C2->add_rule(new TFuzzyRule(true, C2r2alpha, C2r2weight));
    C2->add_rule(new TFuzzyRule(true, C2r3alpha, C2r3weight));
    C2->add_rule(new TFuzzyRule(true, C2r4alpha, C2r4weight));
    C2->add_rule(new TFuzzyRule(true, C2r5alpha, C2r5weight));
    C2->add_rule(new TFuzzyRule(true, C2r6alpha, C2r6weight));
    C2->add_rule(new TFuzzyRule(true, C2r7alpha, C2r7weight));
    C2->add_rule(new TFuzzyRule(true, C2r8alpha, C2r8weight));
    C2->add_rule(new TFuzzyRule(true, C2r9alpha, C2r9weight));


	// setup rules                local_load         remote_load        local_repulsion
    //----------------------------------------------------------------------------------
    //
    if (C2->rules[0] != NULL)
    {
    C2->rules[0]->predicate->clear_premises();
    C2->rules[0]->predicate->clear_operators();
    C2->rules[0]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[Low]));
    C2->rules[0]->predicate->add_operator(AND);
    C2->rules[0]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[Low]));
    C2->rules[0]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r1out]));
    }
    if (C2->rules[1] != NULL)
    {
    C2->rules[1]->predicate->clear_premises();
    C2->rules[1]->predicate->clear_operators();
    C2->rules[1]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[Low]));
    C2->rules[1]->predicate->add_operator(AND);
    C2->rules[1]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[Medium]));
    C2->rules[1]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r2out]));
    }
    if (C2->rules[2] != NULL)
    {
    C2->rules[2]->predicate->clear_premises();
    C2->rules[2]->predicate->clear_operators();
    C2->rules[2]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[Low]));
    C2->rules[2]->predicate->add_operator(AND);
    C2->rules[2]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[High]));
    C2->rules[2]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r3out]));
    }
    if (C2->rules[3] != NULL)
    {
    C2->rules[3]->predicate->clear_premises();
    C2->rules[3]->predicate->clear_operators();
    C2->rules[3]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[Medium]));
    C2->rules[3]->predicate->add_operator(AND);
    C2->rules[3]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[Low]));
    C2->rules[3]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r4out]));
    }
    if (C2->rules[4] != NULL)
    {
    C2->rules[4]->predicate->clear_premises();
    C2->rules[4]->predicate->clear_operators();
    C2->rules[4]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[Medium]));
    C2->rules[4]->predicate->add_operator(AND);
    C2->rules[4]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[Medium]));
    C2->rules[4]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r5out]));
    }
    if (C2->rules[5] != NULL)
    {
    C2->rules[5]->predicate->clear_premises();
    C2->rules[5]->predicate->clear_operators();
    C2->rules[5]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[Medium]));
    C2->rules[5]->predicate->add_operator(AND);
    C2->rules[5]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[High]));
    C2->rules[5]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r6out]));
    }
    if (C2->rules[6] != NULL)
    {
    C2->rules[6]->predicate->clear_premises();
    C2->rules[6]->predicate->clear_operators();
    C2->rules[6]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[High]));
    C2->rules[6]->predicate->add_operator(AND);
    C2->rules[6]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[Low]));
    C2->rules[6]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r7out]));
    }
    if (C2->rules[7] != NULL)
    {
    C2->rules[7]->predicate->clear_premises();
    C2->rules[7]->predicate->clear_operators();
    C2->rules[7]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[High]));
    C2->rules[7]->predicate->add_operator(AND);
    C2->rules[7]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[Medium]));
    C2->rules[7]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r8out]));
    }
    if (C2->rules[8] != NULL)
    {
    C2->rules[8]->predicate->clear_premises();
    C2->rules[8]->predicate->clear_operators();
    C2->rules[8]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[High]));
    C2->rules[8]->predicate->add_operator(AND);
    C2->rules[8]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[2]));
    C2->rules[8]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r9out]));
    }
  }





  // setup "migration force towards remote node" evaluator

  if ( (C1 != NULL) && (C2 != NULL) && (C3 != NULL) )
  {
	// add input and output variables
  	C3->clear_input_variables();         //  name,         unit, begin,  end
	C3->add_input_variable(new TVariable(*(C1->output_variable)));
	C3->add_input_variable(new TVariable(*(C2->output_variable)));
	C3->output_variable =  new TVariable("MigrationForce",        "", 00.0,   100.0);


	// add fuzzysets to the input and output variables

    centre = C3OutputCentre; // %
    begin = C3->output_variable->begin;
    end   = C3->output_variable->end;
    ratio = ( centre / ((end-begin) - centre) );
    // areas
    // end-begin = a                   + b             + c        + d;
    //           = ratio*ratio*ratio*d + ratio*ratio*d + ratio*d  + d;
    //           = d * (ratio*ratio*ratio + ratio*ratio + ratio  + 1);
    //    d      = (end-begin)/(ratio*ratio*ratio + ratio*ratio + ratio  + 1);

    // width      percent
        d     = (end-begin)/(ratio*ratio*ratio + ratio*ratio + ratio  + 1);
        c     = ratio*d;
        b     = ratio*ratio*d;
        a     = ratio*ratio*ratio*d;

    p0 = begin;
    p1 = p0 + a;
    p2 = p1 + b;
    p3 = p2 + c;
    p4 = p3 + d;
	// add fuzzysets to the output variable
  	C3->output_variable->clear_fuzzysets();
    C3->output_variable->add_fuzzyset(new TFuzzySet(C3->output_variable, "VeryLow", S_CURVE  , DOWN , p0, -0.0       , (p0+p1)/2.0, -0.0       , p1));
    C3->output_variable->add_fuzzyset(new TFuzzySet(C3->output_variable, "Low", 	PI_CURVE , UP   , p0, (p0+p1)/2.0, p1         , (p1+p2)/2.0, p2));
    C3->output_variable->add_fuzzyset(new TFuzzySet(C3->output_variable, "Medium", 	PI_CURVE , UP   , p1, (p1+p2)/2.0, p2         , (p2+p3)/2.0, p3));
    C3->output_variable->add_fuzzyset(new TFuzzySet(C3->output_variable, "High", 	PI_CURVE , UP   , p2, (p2+p3)/2.0, p3         , (p3+p4)/2.0, p4));
    C3->output_variable->add_fuzzyset(new TFuzzySet(C3->output_variable, "VeryHigh",S_CURVE  , UP   , p3, -0.0       , (p3+p4)/2.0, -0.0       , p4));



	// create empty rules    (conditional, alpha,  weight)
    C3->clear_rules();
    C3->add_rule(new TFuzzyRule(true, C3r1alpha, C3r1weight));
    C3->add_rule(new TFuzzyRule(true, C3r2alpha, C3r2weight));
    C3->add_rule(new TFuzzyRule(true, C3r3alpha, C3r3weight));
    C3->add_rule(new TFuzzyRule(true, C3r4alpha, C3r4weight));
    C3->add_rule(new TFuzzyRule(true, C3r5alpha, C3r5weight));
    C3->add_rule(new TFuzzyRule(true, C3r6alpha, C3r6weight));
    C3->add_rule(new TFuzzyRule(true, C3r7alpha, C3r7weight));
    C3->add_rule(new TFuzzyRule(true, C3r8alpha, C3r8weight));
    C3->add_rule(new TFuzzyRule(true, C3r9alpha, C3r9weight));


	// setup rules             remote_attraction     local_repulsion       migration_force
    //----------------------------------------------------------------------------------
    //
    if (C3->rules[0] != NULL)
    {
    C3->rules[0]->predicate->clear_premises();
    C3->rules[0]->predicate->clear_operators();
    C3->rules[0]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[Low]));
    C3->rules[0]->predicate->add_operator(AND);
    C3->rules[0]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[Low]));
    C3->rules[0]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r1out]));
    }
    if (C3->rules[1] != NULL)
    {
    C3->rules[1]->predicate->clear_premises();
    C3->rules[1]->predicate->clear_operators();
    C3->rules[1]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[Low]));
    C3->rules[1]->predicate->add_operator(AND);
    C3->rules[1]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[Medium]));
    C3->rules[1]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r2out]));
    }
    if (C3->rules[2] != NULL)
    {
    C3->rules[2]->predicate->clear_premises();
    C3->rules[2]->predicate->clear_operators();
    C3->rules[2]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[Low]));
    C3->rules[2]->predicate->add_operator(AND);
    C3->rules[2]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[High]));
    C3->rules[2]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r3out]));
    }
    if (C3->rules[3] != NULL)
    {
    C3->rules[3]->predicate->clear_premises();
    C3->rules[3]->predicate->clear_operators();
    C3->rules[3]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[Medium]));
    C3->rules[3]->predicate->add_operator(AND);
    C3->rules[3]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[Low]));
    C3->rules[3]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r4out]));
    }
    if (C3->rules[4] != NULL)
    {
    C3->rules[4]->predicate->clear_premises();
    C3->rules[4]->predicate->clear_operators();
    C3->rules[4]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[Medium]));
    C3->rules[4]->predicate->add_operator(AND);
    C3->rules[4]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[Medium]));
    C3->rules[4]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r5out]));
    }
    if (C3->rules[5] != NULL)
    {
    C3->rules[5]->predicate->clear_premises();
    C3->rules[5]->predicate->clear_operators();
    C3->rules[5]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[Medium]));
    C3->rules[5]->predicate->add_operator(AND);
    C3->rules[5]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[High]));
    C3->rules[5]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r6out]));
    }
    if (C3->rules[6] != NULL)
    {
    C3->rules[6]->predicate->clear_premises();
    C3->rules[6]->predicate->clear_operators();
    C3->rules[6]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[High]));
    C3->rules[6]->predicate->add_operator(AND);
    C3->rules[6]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[Low]));
    C3->rules[6]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r7out]));
    }
    if (C3->rules[7] != NULL)
    {
    C3->rules[7]->predicate->clear_premises();
    C3->rules[7]->predicate->clear_operators();
    C3->rules[7]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[High]));
    C3->rules[7]->predicate->add_operator(AND);
    C3->rules[7]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[Medium]));
    C3->rules[7]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r8out]));
    }
    if (C3->rules[8] != NULL)
    {
    C3->rules[8]->predicate->clear_premises();
    C3->rules[8]->predicate->clear_operators();
    C3->rules[8]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[High]));
    C3->rules[8]->predicate->add_operator(AND);
    C3->rules[8]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[High]));
    C3->rules[8]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r9out]));
    }
  }



  // setup "migration recommendation" evaluator

  if ( (C3 != NULL) )
  {
  // do nothing here
  // A4 must be setup outside and
  // the algorithm is applied in RunA4()
  }

}
////////////////////////////////////////////////////////////////////////////////











// fill demanding_nodes_ids[MAX_DEMANDING_NODES]
void TMigrationFC::IdentifyDemandingNodes() // using Fragment->DemandLevelTable[i]
{
  if (Fragment != NULL)
  {
    if (Fragment->DemandLevelTable != NULL)
    {
      num_of_demanding_nodes = Fragment->DemandLevelTable->Count;

      for (int i=0; i<num_of_demanding_nodes; i++)
      {
        demanding_nodes_ids[i] = ((TDemandLevel*)Fragment->DemandLevelTable->Items[i])->NodeId;
      }
    }
  }
}





// average info from fragment's statistics in last T (+ forcast)
void TMigrationFC::FillLocalDemand()	// using Fragment->DemandLevelTable[i]
{
  if (Fragment != NULL)
  {
    for (int i=0; i<num_of_demanding_nodes; i++)
    {
      if (Fragment->Node != NULL)
      {
        if (((TDemandLevel*)Fragment->DemandLevelTable->Items[i])->NodeId == Fragment->Node->get_id())  // local
        {
        	local_demand = ((TDemandLevel*)Fragment->DemandLevelTable->Items[i])->Demand;
            break;
        }
      }
    }
  }
}


// for each demanding node in demanding_nodes_ids[MAX_DEMANDING_NODES]
// fill remote_demands[i]
void TMigrationFC::FillRemoteDemands() 		// using Fragment->DemandLevelTable[i]
{
  if (Fragment != NULL)
  {
    for (int i=0; i<num_of_demanding_nodes; i++)
    {
      if (Fragment->Node != NULL)
      {
        if (((TDemandLevel*)Fragment->DemandLevelTable->Items[i])->NodeId != Fragment->Node->get_id())  // remote
        {
        	remote_demands[i] = ((TDemandLevel*)Fragment->DemandLevelTable->Items[i])->Demand;
        }
      }
    }
  }
}


// for each demanding node, run C1 to get its attraction force and
// fill remote_attractions[MAX_DEMANDING_NODES] with the results
double TMigrationFC::RunC1(int index, double input1, double input2,
						   int AggregationMethod,
                           int CorrelationMethod,
                           int AlphaCutType,
                           int CompositionMethod,
                           int DefuzzificationMethod
					      )
{
  if ( C1 != NULL )
  {
    // apply new inputs
	C1->rules[0]->predicate->premises[0]->set_scalar(input1);
	C1->rules[0]->predicate->premises[1]->set_scalar(input2);
	C1->rules[1]->predicate->premises[0]->set_scalar(input1);
	C1->rules[1]->predicate->premises[1]->set_scalar(input2);
	C1->rules[2]->predicate->premises[0]->set_scalar(input1);
	C1->rules[2]->predicate->premises[1]->set_scalar(input2);
	C1->rules[3]->predicate->premises[0]->set_scalar(input1);
	C1->rules[3]->predicate->premises[1]->set_scalar(input2);
	C1->rules[4]->predicate->premises[0]->set_scalar(input1);
	C1->rules[4]->predicate->premises[1]->set_scalar(input2);
	C1->rules[5]->predicate->premises[0]->set_scalar(input1);
	C1->rules[5]->predicate->premises[1]->set_scalar(input2);
	C1->rules[6]->predicate->premises[0]->set_scalar(input1);
	C1->rules[6]->predicate->premises[1]->set_scalar(input2);
	C1->rules[7]->predicate->premises[0]->set_scalar(input1);
	C1->rules[7]->predicate->premises[1]->set_scalar(input2);
	C1->rules[8]->predicate->premises[0]->set_scalar(input1);
	C1->rules[8]->predicate->premises[1]->set_scalar(input2);

    // run the controller for the new inputs
	remote_attractionsets[index]        = C1->GenerateOutputSet(AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod);
	return remote_attractions[index]    = C1->GenerateOutput   (AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod, DefuzzificationMethod);
  }
  return INVALID_DOMAIN;
}









// fill candidate_nodes_ids[MAX_CANDIDATE_NODES] with ids of
// first nodes with attraction force > CandidacyThreshold, the rest fill with INVALID_ID
void TMigrationFC::SelectCandidates()
{
	// temporary storage areas
    double 		candidate_nodes_remote_demands[MAX_CANDIDATE_NODES];
    double 		candidate_nodes_remote_attractions[MAX_CANDIDATE_NODES];
    TFuzzySet* 	candidate_nodes_remote_attractionsets[MAX_CANDIDATE_NODES];

    int count = 0;
    for (int i=0; i<num_of_demanding_nodes; i++)
    {
//
//PRE-CONDITION:
//
       	if (remote_attractions[i] >= CandidacyThreshold)
        {
	        candidate_nodes_ids[count] 					= demanding_nodes_ids[i];
            candidate_nodes_remote_demands[count]		= remote_demands[i];
            candidate_nodes_remote_attractions[count]	= remote_attractions[i];
            candidate_nodes_remote_attractionsets[count]= new TFuzzySet(*remote_attractionsets[i]);
            count++;
        }
    }
    num_of_candidate_nodes = count;

	// clear-out demandings
    for (int i=0; i<MAX_DEMANDING_NODES; i++)
    {
        remote_demands[i] 		 = 0.0;
        remote_attractions[i] 	 = 0.0;
        delete remote_attractionsets[i]; remote_attractionsets[i] = NULL;
    }

	// copy-back from candidates
    for (int i=0; i<num_of_candidate_nodes; i++)
    {
        remote_demands[i] 		 = candidate_nodes_remote_demands[i];
        remote_attractions[i] 	 = candidate_nodes_remote_attractions[i];
        remote_attractionsets[i] = candidate_nodes_remote_attractionsets[i];
                                   candidate_nodes_remote_attractionsets[i] = NULL;
    }
}









// fill local_load
void TMigrationFC::FillLocalLoad()   // using Fragment->Node->NodeMonitor
{
  if (Fragment != NULL)
  {
   	local_load = Fragment->Node->get_utilization();
  }
}


// for each chosen candidate node
// fill remote_loads[index]
void TMigrationFC::FillRemoteLoads()  //using Fragment->Node->Network->GetNode(id)->NodeMonitor
{
  if (Fragment != NULL)
  {
    for (int i=0; i<num_of_candidate_nodes; i++)
    {
        TNode* remote_node = Fragment->Node->Network->GetNode(candidate_nodes_ids[i]);
        if (remote_node != NULL)
        {
        	remote_loads[i] = remote_node->get_utilization();
        }
    }
  }
}


// for each chosen candidate node
// run C2 to get the local_repulsion force relative to its load
// fill local_repulsions[MAX_CANDIDATE_NODES] with the results
double TMigrationFC::RunC2(int index, double input1, double input2,
						   int AggregationMethod,
                           int CorrelationMethod,
                           int AlphaCutType,
                           int CompositionMethod,
                           int DefuzzificationMethod
					      )
{
  if ( C2 != NULL )
  {
    // apply new inputs
	C2->rules[0]->predicate->premises[0]->set_scalar(input1);
	C2->rules[0]->predicate->premises[1]->set_scalar(input2);
	C2->rules[1]->predicate->premises[0]->set_scalar(input1);
	C2->rules[1]->predicate->premises[1]->set_scalar(input2);
	C2->rules[2]->predicate->premises[0]->set_scalar(input1);
	C2->rules[2]->predicate->premises[1]->set_scalar(input2);
	C2->rules[3]->predicate->premises[0]->set_scalar(input1);
	C2->rules[3]->predicate->premises[1]->set_scalar(input2);
	C2->rules[4]->predicate->premises[0]->set_scalar(input1);
	C2->rules[4]->predicate->premises[1]->set_scalar(input2);
	C2->rules[5]->predicate->premises[0]->set_scalar(input1);
	C2->rules[5]->predicate->premises[1]->set_scalar(input2);
	C2->rules[6]->predicate->premises[0]->set_scalar(input1);
	C2->rules[6]->predicate->premises[1]->set_scalar(input2);
	C2->rules[7]->predicate->premises[0]->set_scalar(input1);
	C2->rules[7]->predicate->premises[1]->set_scalar(input2);
	C2->rules[8]->predicate->premises[0]->set_scalar(input1);
	C2->rules[8]->predicate->premises[1]->set_scalar(input2);

    // run the controller for the new inputs
	local_repulsionsets[index]      = C2->GenerateOutputSet(AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod);
	return local_repulsions[index]  = C2->GenerateOutput   (AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod, DefuzzificationMethod);
  }
  return INVALID_DOMAIN;
}



// for each chosen candidate node, run C3 to get the migration force
// fill forces[MAX_CANDIDATE_NODES] with the results
double TMigrationFC::RunC3(int index, double input1, double input2,
						   int AggregationMethod,
                           int CorrelationMethod,
                           int AlphaCutType,
                           int CompositionMethod,
                           int DefuzzificationMethod
					      )
{
  if ( (C1 != NULL) && (C2 != NULL) && (C3 != NULL) )
  {
    // apply new inputs
	C3->rules[0]->predicate->premises[0]->set_scalar(input1);
	C3->rules[0]->predicate->premises[1]->set_scalar(input2);
	C3->rules[1]->predicate->premises[0]->set_scalar(input1);
	C3->rules[1]->predicate->premises[1]->set_scalar(input2);
	C3->rules[2]->predicate->premises[0]->set_scalar(input1);
	C3->rules[2]->predicate->premises[1]->set_scalar(input2);
	C3->rules[3]->predicate->premises[0]->set_scalar(input1);
	C3->rules[3]->predicate->premises[1]->set_scalar(input2);
	C3->rules[4]->predicate->premises[0]->set_scalar(input1);
	C3->rules[4]->predicate->premises[1]->set_scalar(input2);
	C3->rules[5]->predicate->premises[0]->set_scalar(input1);
	C3->rules[5]->predicate->premises[1]->set_scalar(input2);
	C3->rules[6]->predicate->premises[0]->set_scalar(input1);
	C3->rules[6]->predicate->premises[1]->set_scalar(input2);
	C3->rules[7]->predicate->premises[0]->set_scalar(input1);
	C3->rules[7]->predicate->premises[1]->set_scalar(input2);
	C3->rules[8]->predicate->premises[0]->set_scalar(input1);
	C3->rules[8]->predicate->premises[1]->set_scalar(input2);

    // run the controller for the new inputs
	forcesets[index]        = C3->GenerateOutputSet(AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod);
	return forces[index]    = C3->GenerateOutput   (AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod, DefuzzificationMethod);
  }
  return INVALID_DOMAIN;
}




// fill mobility
void TMigrationFC::FillMobility()  	// using Fragment->Size & Fragment->NumOfRelationships
{
  if (Fragment != NULL)
  {
   	mobility = Fragment->calculate_mobility();
  }
}


// for each chosen candidate node, run A4 to get the migration
// fill migrations[MAX_CANDIDATE_NODES] with the results
double TMigrationFC::RunA4(int index, double input1, double input2)
{
  if ( (C1 != NULL) && (C2 != NULL) && (C3 != NULL) )
  {
    // run the controller for the new inputs
    //                           force                mobility
	return (migrations[index] = (input1/100.0) * pow((input2/100.0), A4)*100.0);
  }
  return INVALID_DOMAIN;
}




// fill migration_node_id with the id of the node with
// (migrations[MAX_CANDIDATE_NODES] AND connectivity) >= MigrationThreshold
// or fill migration_node_id with INVALID_ID for no migration
int TMigrationFC::SelectBestCandidate()
{
    double goodness;
    double first_best  = 0.0;   int index_of_first  = -1;
    double second_best = 0.0;   int index_of_second = -1;

    for (int i=0; i<num_of_candidate_nodes; i++)
    {
	    TNode* migration_node = Fragment->Node->Network->GetNode(candidate_nodes_ids[i]);
        if (migration_node != NULL)
        {

//
//POST-CONDITION:
//
        // skip if candidate is local node
        if (migration_node->get_id() == Fragment->Node->get_id())
        {
          continue;
        }

//
//POST-CONDITION:
//
        // skip candidate if target is FULL or BUSY or SMALL or SLOW

            int needed_storage_capacity = Fragment->get_size() / MAX_FULL_SIZE_FRAGMENTS_PER_NODE;

		    int weighted_demand             = ( Fragment->get_demand() * Fragment->get_size() ) / 100;
            int needed_processor_capacity   = weighted_demand / MAX_FULLY_DEMANDED_FULL_SIZE_FRAGMNETS_PER_NODE;


            // skip candidate if target is FULL or BUSY
            if ( (migration_node->get_storage_utilization() > 95) || (migration_node->get_processor_utilization() > 95) )
            {
              continue;
            }


            // skip candidate if target is SMALL or SLOW
            if (
                ( migration_node->get_storage_capacity()   < needed_storage_capacity )
              ||
                ( migration_node->get_processor_capacity() < needed_processor_capacity )
               )
            {
              continue;
            }



            TNode* remote_node       = Fragment->Node->Network->GetNode(candidate_nodes_ids[i]);
            int    remote_node_index = Fragment->Node->Network->Nodes->IndexOf(remote_node);
            int    local_node_index  = Fragment->Node->Network->Nodes->IndexOf(Fragment->Node);

            //goodness    = migration recommendation
            // 	    	  + delay of all intermediate links on fastest path
            //   	      + migration node connectivity
            //    	      - migration node utilization
            //        	  + cost of migrating size-data
            // 	            		  ;
        	goodness      = MIGRATION_RECOMMECNDATION_FACTOR   * migrations[i]
           	    		  + MIGRATION_LINKS_DELAY_FACTOR       * Fragment->Node->Network->delays[local_node_index][remote_node_index]
       	    	    	  + MIGRATION_NODE_CONNECTIVITY_FACTOR * migration_node->get_connectivity()
       	        		  + MIGRATION_COST_FACTOR              * Fragment->get_size()/100.0 * Fragment->Node->Network->delays[Fragment->Node->Network->Nodes->IndexOf(Fragment->Node)][candidate_nodes_ids[i]]
   	            		  ;


//
//POST-FILITERING:
//
	        // get the best candidate
			if (goodness > first_best)
	    	{
        	    first_best = goodness;
                index_of_first = i;
	        }
            else if (goodness > second_best)
            {
        	    second_best = goodness;
                index_of_second = i;
            }
            else
            {
                continue;
            }




//
//POST-CONDITION:
//
     // select the best candidate
            if (migrations[index_of_first ] >= MigrationThreshold)
            {
                migration_node_id = candidate_nodes_ids[index_of_first];
            }



  //////////////////////////////////////////////////////////////////////////////
  // goto first-best, replication will take care of other demanding nodes
  //////////////////////////////////////////////////////////////////////////////


//
//POST-CONDITION:
//
     // choose a common node between the first and second best on
     // shortest paths between the local node and the two best
     //
     //
     //        nearest_&_capable node on common path
     //
     //

            // if both bests high then choose a node common to the first and second best nearest to the first
            if (
                  (migrations[index_of_first ] >= MigrationThreshold)
                &&
                  (migrations[index_of_second] >= MigrationThreshold)
                )
            {
                TNode* local_node  = Fragment->Node;
                TNode* first_node  = Fragment->Node->Network->GetNode(candidate_nodes_ids[index_of_first ]);
                TNode* second_node = Fragment->Node->Network->GetNode(candidate_nodes_ids[index_of_second]);

                int local_node_index_in_network  = Fragment->Node->Network->Nodes->IndexOf(local_node);
                int first_node_index_in_network  = Fragment->Node->Network->Nodes->IndexOf(first_node);
                int second_node_index_in_network = Fragment->Node->Network->Nodes->IndexOf(second_node);

                ////////////////////////////////////////////////////////////////
                int common_nodes_ids[MAX_NODES]; for (int n=0; n<MAX_NODES; n++) common_nodes_ids[n] = -1;
                int common_nodes_count = 0;
                ////////////////////////////////////////////////////////////////

                                          /////////////// CHANGING DEMANDS /////////////////////////////
                                          // first choice --> response = 00 18 87    resource = 11 12 49
bool common_on_path_to_target   = false;  // on_path      --> response = 00 19 77    resource = 12 12 49
bool common_between_two         = false;  // between      --> response = 00 18 90    resource = 11 12 49
                                          /////////////// FIXED DEMANDS ////////////////////////////////
                                          // first choice --> response = 00 19 48    resource = 09 09 48
                                          // on_path      --> response = 00 19 48    resource = 09 09 48
                                          // between      --> response = 00 18 48    resource = 09 09 47
                                          //////////////////////////////////////////////////////////////
if (common_on_path_to_target)
{
                // find all common nodes including self from source to targets
                for (int x=0; x<MAX_NODES+1; x++)
                {
                  int intermediate1_index;
                  if (x == 0)
                  {
                    intermediate1_index = first_node_index_in_network;
                  }
                  else
                  {
                    intermediate1_index = Fragment->Node->Network->intermediates[first_node_index_in_network ][local_node_index_in_network][x-1];
                    if (intermediate1_index == -1)  break;
                  }

                  for (int y=0; y<MAX_NODES; y++)
                  {
                    int intermediate2_index;
                    if (y == 0)
                    {
                      intermediate2_index = second_node_index_in_network;
                    }
                    else
                    {
                      intermediate2_index = Fragment->Node->Network->intermediates[second_node_index_in_network ][local_node_index_in_network][y-1];
                      if (intermediate2_index == -1)  break;
                    }

                    if (intermediate1_index == intermediate2_index) // all common nodes
                    {
                        int common_node_id = ((TNode*)(Fragment->Node->Network->Nodes->Items[intermediate1_index]))->get_id();

                        ////////////////////////////////////////////////////////////////
                        common_nodes_ids[common_nodes_count++] = common_node_id;
                        ////////////////////////////////////////////////////////////////
                    }
                  }
                }

                // find nearest common node which is less than 80% busy
                for (int c=0; c<common_nodes_count; c++)
                {
                   TNode* common_node = Fragment->Node->Network->GetNode(common_nodes_ids[c]);
                   int utilization = common_node->get_utilization();
                   if (utilization < 80)
                   {
                       migration_node_id = common_node->get_id();
                   }
                }

}

                // OR


else if (common_between_two)
{

                // find all common nodes between the best two targets including targets
                int first_node_id = ((TNode*)(Fragment->Node->Network->Nodes->Items[first_node_index_in_network]))->get_id();
                common_nodes_ids[common_nodes_count++] = first_node_id;

                for (int x=0; x<MAX_NODES; x++)
                {
                  int intermediate_node_index = Fragment->Node->Network->intermediates[first_node_index_in_network][second_node_index_in_network][x];
                  if (intermediate_node_index == -1)  break;

                  int intermediate_node_id = ((TNode*)(Fragment->Node->Network->Nodes->Items[intermediate_node_index]))->get_id();
                  common_nodes_ids[common_nodes_count++] = intermediate_node_id;
                }

                int second_node_id = ((TNode*)(Fragment->Node->Network->Nodes->Items[second_node_index_in_network]))->get_id();
                common_nodes_ids[common_nodes_count++] = second_node_id;


                /*
                // find largest common node
                int max_capacity = 0;
                for (int c=0; c<common_nodes_count; c++)
                {
                   TNode* common_node = Fragment->Node->Network->GetNode(common_nodes_ids[c]);
                   int capacity = common_node->get_capacity();
                   if (capacity > max_capacity)
                   {
                       max_capacity = capacity;
                       migration_node_id = common_node->get_id();
                   }
                }
                */

                // find middle common node which is less than 80% busy
                for (int i=0; i<common_nodes_count; i++)
                {
                   int c;
                   if ( !(i%2) )    // i = even
                   {
                       c = common_nodes_count/2 + i/2;
                   }
                   else             // i = odd
                   {
                       c = common_nodes_count/2 - i/2 - 1;
                   }
                   TNode* common_node = Fragment->Node->Network->GetNode(common_nodes_ids[c]);
                   int utilization = common_node->get_utilization();
                   if (utilization < 80)
                   {
                     migration_node_id = common_node->get_id();
                     break;
                   }
                }

}

            }


        }
    }

    // igonre migration to self (local node)
    if ( migration_node_id == Fragment->Node->get_id() )
    {
        return INVALID_ID;
    }

	return migration_node_id;
}









// returns the migration_node_id, or INVALID_ID if none
int TMigrationFC::Run()
{
    IdentifyDemandingNodes();


    FillLocalDemand();
    FillRemoteDemands();
    for (int i=0; i<num_of_demanding_nodes; i++)
    {
	    RunC1(i, remote_demands[i], local_demand,
        //    AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod, DefuzzificationMethod
//                   PRODUCT,           TRUNCATE,        STRONG,        BOUNDEDSUM,        NEAR_EDGE        );
                   ZADEH    ,           TRUNCATE,        STRONG,        BOUNDEDSUM,        NEAR_EDGE        );
    }


	SelectCandidates();


    FillLocalLoad();
    FillRemoteLoads();
    for (int i=0; i<num_of_candidate_nodes; i++)
    {
	    RunC2(i, local_load, remote_loads[i],
        //    AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod, DefuzzificationMethod
//                   PRODUCT,           TRUNCATE,        STRONG,        BOUNDEDSUM,        NEAR_EDGE        );
                   ZADEH    ,           TRUNCATE,        STRONG,        BOUNDEDSUM,        NEAR_EDGE        );
    }



    for (int i=0; i<num_of_candidate_nodes; i++)
    {
	    RunC3(i, remote_attractions[i], local_repulsions[i],
        //    AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod, DefuzzificationMethod
                   PRODUCT,           TRUNCATE,        STRONG,        BOUNDEDSUM,        NEAR_EDGE        );
    }


	FillMobility();
    for (int i=0; i<num_of_candidate_nodes; i++)
    {
	    RunA4(i, forces[i], mobility);
    }


	return SelectBestCandidate();
}






