//---------------------------------------------------------------------------
#ifndef TMigrationFCH
#define TMigrationFCH
//---------------------------------------------------------------------------

#include "TFuzzySet.h"
#include "TFuzzyRule.h"
#include "TFuzzyController.h"

#include "TFragment.h"
#include "TNode.h"
#include "TLink.h"
#include "TNetwork.h"
//---------------------------------------------------------------------------


#define INVALID_ID				 -1

#define MAX_DEMANDING_NODES		256
#define MAX_CANDIDATE_NODES		 16


// pre-conditions
//
// RemoteDemand > CandidacyThreshold = MIN_DEMAND
//


//
// multi-satge controller
//
//
//  RemoteDemand---C1 ++
//                 C1---RemoteAttraction---C3 +
//  LocalDemand----C1 -                    C3
//                                         C3
//                                         C3---MigrationForce---A4
//  LocalLoad------C2 +                    C3                    A4
//                 C2---LocalRepulsion-----C3 +                  A4
//  RemoteLoad-----C2 -                                          A4
//                                                               A4---MigrationRecommendation---
//                                                               A4
//  Mobility (constant)------------------------------------------A4
//


// post-conditions
//
// Recommendation > MigrationThreshold
// RemoteLoad < MAX_LOAD
//





class TMigrationFC
{
  public:
	TFuzzyController* 	C1; // Demand    controller
	TFuzzyController* 	C2; // Load      controller
	TFuzzyController* 	C3; // Force     controller
    double              A4; // Mobility  algorithm  [output = input * mobility^A4]

    int C1Input1Centre;
					int C1OutputCentre;
    int C1Input2Centre;
  									int C3OutputCentre;
    int C2Input1Centre;
    				int C2OutputCentre;
    int C2Input2Centre;

    static double CandidacyThreshold;		// minmum level to justfy consideration
    static double MigrationThreshold;		// minmum level to justfy migratation

    int C1r1out; double C1r1alpha; double C1r1weight;
    int C1r2out; double C1r2alpha; double C1r2weight;
    int C1r3out; double C1r3alpha; double C1r3weight;
    int C1r4out; double C1r4alpha; double C1r4weight;
    int C1r5out; double C1r5alpha; double C1r5weight;
    int C1r6out; double C1r6alpha; double C1r6weight;
    int C1r7out; double C1r7alpha; double C1r7weight;
    int C1r8out; double C1r8alpha; double C1r8weight;
    int C1r9out; double C1r9alpha; double C1r9weight;

    int C2r1out; double C2r1alpha; double C2r1weight;
    int C2r2out; double C2r2alpha; double C2r2weight;
    int C2r3out; double C2r3alpha; double C2r3weight;
    int C2r4out; double C2r4alpha; double C2r4weight;
    int C2r5out; double C2r5alpha; double C2r5weight;
    int C2r6out; double C2r6alpha; double C2r6weight;
    int C2r7out; double C2r7alpha; double C2r7weight;
    int C2r8out; double C2r8alpha; double C2r8weight;
    int C2r9out; double C2r9alpha; double C2r9weight;

    int C3r1out; double C3r1alpha; double C3r1weight;
    int C3r2out; double C3r2alpha; double C3r2weight;
    int C3r3out; double C3r3alpha; double C3r3weight;
    int C3r4out; double C3r4alpha; double C3r4weight;
    int C3r5out; double C3r5alpha; double C3r5weight;
    int C3r6out; double C3r6alpha; double C3r6weight;
    int C3r7out; double C3r7alpha; double C3r7weight;
    int C3r8out; double C3r8alpha; double C3r8weight;
    int C3r9out; double C3r9alpha; double C3r9weight;


  private:
    int		num_of_demanding_nodes;
    int		num_of_candidate_nodes;

	// inputs
    // average info from fragment's statistics in last T (+ forcast)
    double remote_demands[MAX_DEMANDING_NODES];
    double local_demand;

    double local_load;
    double remote_loads[MAX_CANDIDATE_NODES];

    double mobility;

  public:
	// outputs
    int    		demanding_nodes_ids[MAX_DEMANDING_NODES];
    int    		candidate_nodes_ids[MAX_CANDIDATE_NODES];

    TFuzzySet* 	remote_attractionsets[MAX_DEMANDING_NODES];
    double 		remote_attractions[MAX_DEMANDING_NODES];

    TFuzzySet* 	local_repulsionsets[MAX_CANDIDATE_NODES];
    double 		local_repulsions[MAX_CANDIDATE_NODES];

    TFuzzySet* 	forcesets[MAX_CANDIDATE_NODES];
    double 		forces[MAX_CANDIDATE_NODES];

    double 		migrations[MAX_CANDIDATE_NODES];

    int    		migration_node_id;

  public:
	TFragment*		Fragment;

	TMigrationFC(TFragment* _Fragment);
	~TMigrationFC();


  	double begin;                  // begin of domain
  	double end;                    // end of domain
  	double p0, p1, p2, p3, p4;     // points in domain's begin..end
  	double a, b, c, d;             // areas in 5-value variable

                      				// < 50% means fast becomes high
  	double centre;    				// = 50% means linearly becomes high
                      				// > 50% means slow becomes high

	double ratio;     				// centre/total

    void SetupControllers();





    // fill demanding_nodes_ids[MAX_DEMANDING_NODES]
    void IdentifyDemandingNodes();  // using Fragment->DemandLevelTable[i]


    // fill local_demand
    void FillLocalDemand();  // using Fragment->DemandLevelTable[i]

    // fill remote_demands[i]
    void FillRemoteDemands();  // using Fragment->DemandLevelTable[i]

    // for each demanding node, run C1 to get its attraction force and
	// fill remote_attractions[MAX_DEMANDING_NODES] with the results
    double RunC1(int index, double input1, double input2,
    			 int AggregationMethod,
    			 int CorrelationMethod,
    			 int AlphaCutType,
    			 int CompositionMethod,
    			 int DefuzzificationMethod
    			);


	// fill candidate_nodes_ids[MAX_CANDIDATE_NODES] with ids of nodes with attraction force > MEDIUM
	void SelectCandidates();



    // fill local_load
    void FillLocalLoad();       //using Fragment->Node->NodeMonitor

    // fill remote_loads[i]
    void FillRemoteLoads();		//using Fragment->Node->Network->GetNode(id)->NodeMonitor


    // for each chosen candidate node
    // run C2 to get the local_repulsion force relative to its load
	// fill local_repulsions[MAX_CANDIDATE_NODES] with the results
    double RunC2(int index, double input1, double input2,
    			 int AggregationMethod,
    			 int CorrelationMethod,
    			 int AlphaCutType,
    			 int CompositionMethod,
    			 int DefuzzificationMethod
    			);

    // for each chosen candidate node, run C3 to get the migration force
	// fill forces[MAX_CANDIDATE_NODES] with the results
    double RunC3(int index, double input1, double input2,
    			 int AggregationMethod,
    			 int CorrelationMethod,
    			 int AlphaCutType,
    			 int CompositionMethod,
    			 int DefuzzificationMethod
    			);


	void FillMobility();  //using Fragment->Size + Fragment->NumOfRelationships

    // for each chosen candidate node, run A4 to get the migration recommendation
	// fill migrations[MAX_CANDIDATE_NODES] with the results
    double RunA4(int index, double input1, double input2);

	// fill migration_node_id with the id of the node with
    // (migrations[MAX_CANDIDATE_NODES] AND connectivity) >= MigrationThreshold
    // or fill migration_node_id with INVALID_ID for no migration
	int SelectBestCandidate();

    // returns the migration_node_id to migrate to.
	int Run();
};

//---------------------------------------------------------------------------
#endif

