//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop

//#include <values.h> // MAXINT
#include <math.h>   // pow
#define MAXINT 10000
#include "TNetwork.h"
#include "TFragment.h"
//---------------------------------------------------------------------------



    void TNetwork::calculate_shortest_paths()
    {
        // initialize
       	for (int i=0; i<MAX_NODES; i++)
        {
       	  for (int j=0; j<MAX_NODES; j++)
          {
            neighbours[i][j] = -1;
            if (i == j)
            {
              delays[i][j]   = 0;
            }
            else
            {
              delays[i][j]   = MAXINT;
            }
          }
        }
       	for (int i=0; i<MAX_NODES; i++)
        {
       	  for (int j=0; j<MAX_NODES; j++)
          {
       	    for (int k=0; k<MAX_NODES; k++)
            {
           	  intermediates[i][j][k]  = -1;
            }
          }
        }


        // get neighbours of all nodes -> neighbours[][]
		for (int i=0; i<Nodes->Count; i++)
		{
            TNode* node = ((TNode*)Nodes->Items[i]);
            int index = Nodes->IndexOf(node);
            node->get_neighbours(neighbours[index]);
        }

        // fill delay[][] with initial configuration
		for (int i=0; i<Links->Count; i++)
		{
          TLink* link = ((TLink*)Links->Items[i]);
          TNode* node_a = link->Node_a;
          TNode* node_b = link->Node_b;

          if ( (node_a != NULL) && (node_b != NULL) )
          {
            int index_a = Nodes->IndexOf(node_a);
            int index_b = Nodes->IndexOf(node_b);

            if ( (index_a != -1) && (index_b != -1) )
            {
               delays[index_a][index_b] = link->get_delay();
               delays[index_b][index_a] = link->get_delay();
            }
          }
        }




        // Floyed
        // get shortest_paths from every node to all other nodes -> intermediates[][][0]
        // update delays[][] with sum of delays on shortest_path between k,j nodes
        // add direct intermediate node between k,j
       	for (int i=0; i<Nodes->Count; i++)
        {
       	  for (int j=0; j<Nodes->Count; j++)
          {
            for (int k=0; k<Nodes->Count; k++)
            {
              if ( (delays[k][i] + delays[i][j])   <   (delays[k][j]) )
              {
                 delays[k][j] = (delays[k][i] + delays[i][j]);
                 // add intermediate node to path
                 intermediates[k][j][0] = i;
              }
            }
          }
        }





        // find intemediates between i---mid------j starting form mid in both directions
        // fill -> intermediates[][][]
        TList* Path  = new TList();
        int from;
        int to;
       	for (int i=0; i<Nodes->Count; i++)
        {
       	  for (int j=0; j<Nodes->Count; j++)
          {
            Path->Clear();

            // find forward intermediates
            from = i;
            to   = j;
            while (intermediates[from][to][0] != -1)
            {
                // insert after last found intermediate
                Path->Insert(Path->Count, new int(intermediates[from][to][0]));
                from = intermediates[from][to][0];
            }


            // find backward intermediates
            from = i;
            to   = j;
            while (intermediates[from][to][0] != -1)
            {
                // insert before last found intermediate
                Path->Insert(0, new int(intermediates[from][to][0]));
                to = intermediates[from][to][0];
            }


            // remove duplicates
        	for (int p=0; p<Path->Count-1; p++)
            {
        	  for (int q=p+1; q<Path->Count; q++)
              {
                if( (*((int*)Path->Items[p])) == (*((int*)Path->Items[q])) )
                {
                  Path->Delete(q); q--;
                }
              }
            }


            // copy Path to intermediates
        	for (int p=0; p<Path->Count; p++)
            {
              intermediates[i][j][p] = *((int*)Path->Items[p]);
            }
            intermediates[i][j][Path->Count] = -1;

          }
        }
        delete Path;







        // clear  TList* shortest_paths[][]; // intermediate_links
		for (int i=0; i<Nodes->Count; i++)
		{
	      for (int j=0; j<Nodes->Count; j++)
		  {
            shortest_paths[i][j]->Clear();
          }
        }


        int mids[MAX_NODES][MAX_NODES][MAX_NODES];	for (int i=0; i<MAX_NODES; i++) for (int j=0; j<MAX_NODES; j++) for (int k=0; k<MAX_NODES; k++) mids[i][j][k] = -1;
		for (int i=0; i<Nodes->Count; i++)
		{
	      for (int j=0; j<Nodes->Count; j++)
		  {
            if (i == j) continue; // next j

            mids[i][j][0] = i;
	        for (int k=0; k<Nodes->Count; k++)
		    {
              if (intermediates[i][j][k] != -1)
              {
                mids[i][j][k+1] = intermediates[i][j][k];
              }
              else
              {
                mids[i][j][k+1] = j;
                break; // k
              }
            }
          }
        }



        // fill  TList* shortest_paths[][]; // intermediate_links
		for (int i=0; i<Nodes->Count; i++)
		{
	      for (int j=0; j<Nodes->Count; j++)
		  {
            if (i == j) continue; // next j

            from = i;
            to   = i;

            bool stop = false;
        	for (int k=0; k<MAX_NODES; k++) // unknown length, but terminated by -1
            {
              if (mids[i][j][k] != -1)
              {
                from = mids[i][j][k];

                if (mids[i][j][k+1] != -1)
                {
                  to = mids[i][j][k+1];
                }
                else
                {
                    stop = true;
                }
              }
              else
              {
                  stop = true;
              }
              if (stop) break;


              // find the link
              TNode* from_node = (TNode*)Nodes->Items[from];
              TNode* to_node   = (TNode*)Nodes->Items[to];
              if (from_node && to_node)
              {
      	        for (int l=0; l<from_node->Links->Count; l++)
        	    {
       			  TLink* link = (TLink*)from_node->Links->Items[l];
                  if (link)
                  {
                    TNode* node_a = (TNode*)link->Node_a;
    	            TNode* node_b = (TNode*)link->Node_b;
                    if (node_a == to_node || node_b == to_node)
                    {
                      shortest_paths[i][j]->Add(link);
                      break; // l
                    }
                  }
                }// for l
              }

            }// for k

          } // for j

        } // for i

    }







    void TNetwork::update_links_utilization()
    {
        calculate_shortest_paths(); // fill intermediates[i][j]
        // based on last utilization -> last shortest paths



        ////////////////////////////////////////////////////////////////////////////
        // add fragment->demands[from] to links' traffic on shortest_paths[from][to]
        ////////////////////////////////////////////////////////////////////////////

        // reset links utilization
        for(int l = 0; l < Links->Count; l++)
   	    {
            TLink* link = ((TLink*)Links->Items[l]);
            link->set_utilization(0);
        }

        // increment links utilization
        for(int to=0; to<Nodes->Count; to++)
   	    {
            TNode* to_node = ((TNode*)Nodes->Items[to]);
	    	for(int f=0; f<to_node->Fragments->Count; f++)
	        {
    	        TFragment* to_fragment = ((TFragment*)to_node->Fragments->Items[f]);
    	    	for(int from=0; from<Nodes->Count; from++)
	            {
                    if (from == to) continue; //skip

        	    	for(int l=0; l<shortest_paths[from][to]->Count; l++)
	                {
                        TLink* link = ((TLink*)(shortest_paths[from][to]->Items[l]));

                        // size-weighted averaged-demand of fragment
                        float used_capacity;
                        used_capacity  =  ( (to_fragment->demands[from] * to_fragment->get_size()) / 100.0 );


                        float extra_utilization = 0;
                        if (link->get_capacity() > 0)
                        {
                            extra_utilization = (used_capacity  * 100.0) / link->get_capacity();
                        }
                        link->set_utilization(link->get_utilization() + extra_utilization);
                    }
                }
    	    }
        }

        //                                                     MAX_FULL_SIZE_FRAGMENTS_PER_NODE
        // allow links to handle demands from MAX_FULLY_DEMANDED_FULLY_POPULATED_NODES_SERVICED_PER_FULL_CAPACITY_LINK nodes
        for(int l = 0; l < Links->Count; l++)
   	    {
            TLink* link = ((TLink*)Links->Items[l]);
            link->set_utilization(link->get_utilization() / (MAX_FULL_SIZE_FRAGMENTS_PER_NODE * MAX_FULLY_DEMANDED_FULLY_POPULATED_NODES_SERVICED_PER_FULL_CAPACITY_LINK) );
        }
    }










    TNetwork::TNetwork(TNetwork* _network)
	{
	  if(_network != NULL)
	  {
	    id      = _network->get_id();

    	Nodes   =   new TList();
		for (int i=0; i<_network->Nodes->Count; i++)
		{
			TNode* copyNode = new TNode((TNode*)_network->Nodes->Items[i]);
			copyNode->Network = this;
			Nodes->Add(copyNode);
		}
        selected_node_index 	= _network->get_selected_node_index();

    	Links   =   new TList();
		for (int i=0; i<_network->Links->Count; i++)
		{
			TLink* copyLink = new TLink((TLink*)_network->Links->Items[i]);
			copyLink->Network = this;
			Links->Add(copyLink);
		}
        selected_link_index 	= _network->get_selected_link_index();


       	for (int i=0; i<MAX_NODES; i++)
        {
       	  for (int j=0; j<MAX_NODES; j++)
          {
           	  for (int k=0; k<MAX_NODES; k++)
              {
                  intermediates[i][j][k] = -1;
              }
          }
        }

       	for (int i=0; i<MAX_NODES; i++)
        {
       	  for (int j=0; j<MAX_NODES; j++)
          {
              shortest_paths[i][j] = new TList();
          }
        }
        calculate_shortest_paths();


		for (int i=0; i<_network->Nodes->Count; i++)
		{
			TNode* current_node = (TNode*)(Nodes->Items[i]);
			current_node->update_connectivity();
		}


		CPUCapacity	= _network->CPUCapacity;
		MEMCapacity	= _network->MEMCapacity;
		IOCapacity	= _network->IOCapacity;
		STOCapacity	= _network->STOCapacity;
		COMCapacity	= _network->COMCapacity;
        TOTALCapacity = CPUCapacity + MEMCapacity + IOCapacity + STOCapacity + COMCapacity;
    	ResourceUsageHistory   =   new TList();
		for (int i=0; i<_network->ResourceUsageHistory->Count; i++)
		{
			ResourceUsageHistory->Add(new int(*((int*)_network->ResourceUsageHistory->Items[i])));
		}
        minResource = _network->minResource;
        maxResource = _network->maxResource;

    	ResponseTimeHistory   =   new TList();
		for (int i=0; i<_network->ResponseTimeHistory->Count; i++)
		{
			ResponseTimeHistory->Add(new int(*((int*)_network->ResponseTimeHistory->Items[i])));
		}
        minResponse = _network->minResponse;
        maxResponse = _network->maxResponse;


        for (int i=0; i<Links->Count; i++)
        {
		  // forEach link
          TLink* i_link = (TLink*)Links->Items[i];
	      if (i_link != NULL)
	      {
		    // 1. find twin_link in _network
		    TLink* twin_link = NULL;
            for (int j=0; j<_network->Links->Count; j++)
	        {
	          TLink* j_link = (TLink*)_network->Links->Items[j];
	          if (j_link != NULL)
	          {
	          	if ( i_link->get_id() == j_link->get_id() )
		        {
		          twin_link = j_link;
	    	      break;
	            }
              }
            }


		    // 2. find ids of twin_node_a and twin_node_b in _network
            int node_a_id = -1;
            int node_b_id = -1;
		    if (twin_link != NULL)
		    {
              TNode* twin_node_a = (TNode*)twin_link->Node_a;
              TNode* twin_node_b = (TNode*)twin_link->Node_b;
              if (twin_node_a != NULL) node_a_id = twin_node_a->get_id();
              if (twin_node_b != NULL) node_b_id = twin_node_b->get_id();
		    }


		    // 3. Add the link to nodes with id = node_a_id and node_b_id, and update nodes
            for (int k=0; k<Nodes->Count; k++)
	        {
	          TNode* k_node = (TNode*)Nodes->Items[k];
	          if (k_node != NULL)
	          {
	            if( k_node->get_id() == node_a_id )
	            {
	              k_node->Links->Add(i_link);
                  i_link->Node_a = k_node;
                }
	            else if( k_node->get_id() == node_b_id )
	            {
	              k_node->Links->Add(i_link);
                  i_link->Node_b = k_node;
                }
              }
	        }
          }
		}

      }

	}




    TNetwork::TNetwork()
	{
    	id 		= 0;
    	Nodes   =   new TList();
    	Links   =   new TList();

       	for (int i=0; i<MAX_NODES; i++)
        {
       	  for (int j=0; j<MAX_NODES; j++)
          {
           	  for (int k=0; k<MAX_NODES; k++)
              {
                  intermediates[i][j][k] = -1;
              }
          }
        }

       	for (int i=0; i<MAX_NODES; i++)
        {
       	  for (int j=0; j<MAX_NODES; j++)
          {
              shortest_paths[i][j] = new TList();
          }
        }

        selected_node_index 	= -1;
        selected_link_index 	= -1;
		DeSelectChildren();

		CPUCapacity	= 0;
		MEMCapacity	= 0;
		IOCapacity	= 0;
		STOCapacity	= 0;
		COMCapacity	= 0;
        TOTALCapacity = CPUCapacity + MEMCapacity + IOCapacity + STOCapacity + COMCapacity;
		ResourceUsageHistory	= new TList();
		ResponseTimeHistory		= new TList();
        minResource = 100;
        maxResource = 0;
        minResponse = 100;
        maxResponse = 0;

		TNode::serial_num	= 0;
		TLink::serial_num	= 0;


        //randomize();
		AddNode( 35,  45);
		AddNode(110,  45);
		AddNode(185,  45);
		AddNode( 35, 125);
		AddNode(110, 125);
		AddNode(185, 125);

///*
        ((TNode*)Nodes->Items[0])->set_num_of_processors(1);
        ((TNode*)Nodes->Items[1])->set_num_of_processors(4);
        ((TNode*)Nodes->Items[2])->set_num_of_processors(16);
        ((TNode*)Nodes->Items[3])->set_num_of_processors(64);
        ((TNode*)Nodes->Items[4])->set_num_of_processors(256);
        ((TNode*)Nodes->Items[5])->set_num_of_processors(1024);

        ((TNode*)Nodes->Items[0])->set_storage_capacity(90);
        ((TNode*)Nodes->Items[1])->set_storage_capacity(70);
        ((TNode*)Nodes->Items[2])->set_storage_capacity(80);
        ((TNode*)Nodes->Items[3])->set_storage_capacity(65);
        ((TNode*)Nodes->Items[4])->set_storage_capacity(80);
        ((TNode*)Nodes->Items[5])->set_storage_capacity(95);
        ((TNode*)Nodes->Items[0])->set_processor_capacity(90);
        ((TNode*)Nodes->Items[1])->set_processor_capacity(70);
        ((TNode*)Nodes->Items[2])->set_processor_capacity(80);
        ((TNode*)Nodes->Items[3])->set_processor_capacity(65);
        ((TNode*)Nodes->Items[4])->set_processor_capacity(80);
        ((TNode*)Nodes->Items[5])->set_processor_capacity(95);
//*/

		AddLink((TNode*)Nodes->Items[0], (TNode*)Nodes->Items[1]);
		AddLink((TNode*)Nodes->Items[1], (TNode*)Nodes->Items[2]);
		AddLink((TNode*)Nodes->Items[2], (TNode*)Nodes->Items[5]);
		AddLink((TNode*)Nodes->Items[5], (TNode*)Nodes->Items[4]);
		AddLink((TNode*)Nodes->Items[4], (TNode*)Nodes->Items[3]);
		AddLink((TNode*)Nodes->Items[2], (TNode*)Nodes->Items[4]);

        ((TLink*)Links->Items[0])->set_type(LAN);
        ((TLink*)Links->Items[1])->set_type(WAN);
        ((TLink*)Links->Items[2])->set_type(LAN);
        ((TLink*)Links->Items[3])->set_type(LAN);
        ((TLink*)Links->Items[4])->set_type(WAN);
        ((TLink*)Links->Items[5])->set_type(LAN);

        ((TLink*)Links->Items[0])->set_initial_delay(10);
        ((TLink*)Links->Items[1])->set_initial_delay(50);
        ((TLink*)Links->Items[2])->set_initial_delay(10);
        ((TLink*)Links->Items[3])->set_initial_delay(10);
        ((TLink*)Links->Items[4])->set_initial_delay(50);
        ((TLink*)Links->Items[5])->set_initial_delay(15);

///*
        ((TLink*)Links->Items[0])->set_capacity( 30);
        ((TLink*)Links->Items[1])->set_capacity( 90);
        ((TLink*)Links->Items[2])->set_capacity( 30);
        ((TLink*)Links->Items[3])->set_capacity( 30);
        ((TLink*)Links->Items[4])->set_capacity( 90);
        ((TLink*)Links->Items[5])->set_capacity( 30);
//*/

		for (int i=0; i<Nodes->Count; i++)
		{
            TNode* node = ((TNode*)Nodes->Items[i]);
    		node->update_connectivity();
	    	node->update_capacity();
            node->update_utilization();   // also will re-calculate
                                		  //processor_utilization
                                		  //storage_utilization
        }

        info_type   = 0;
	}






    TNetwork::~TNetwork()
	{
    	Nodes->Clear();
    	delete Nodes;

    	Links->Clear();
    	delete Links;

       	for (int i=0; i<MAX_NODES; i++)
        {
       	  for (int j=0; j<MAX_NODES; j++)
          {
             delete shortest_paths[i][j];
          }
        }

	    ResourceUsageHistory->Clear();
	    delete ResourceUsageHistory;

	    ResponseTimeHistory->Clear();
	    delete ResponseTimeHistory;
	}






	TNode* TNetwork::GetNode(int id)
	{
   		for (int i=0; i<Nodes->Count; i++)
   		{
			TNode* current_node = (TNode*)(Nodes->Items[i]);
            if (current_node != NULL)
            {
				if ( current_node->get_id() == id )
    	        {
        	    	return current_node;
            	}
            }
	    }
   	    return NULL;
    }



	TNode* TNetwork::AddNode(int x, int y)
	{
       	TNode* node = new TNode(this, Point(x, y));
	    Nodes->Add(node);
		DeSelectChildren();
        return node;
    }



	TLink* TNetwork::AddLink(TNode* _node_a, TNode* _node_b)
	{
	    TLink* link = new TLink(this, _node_a, _node_b);
	    Links->Add(link);
		DeSelectChildren();

    	for (int i=0; i<Nodes->Count; i++)
    	{
            ((TNode*)Nodes->Items[i])->update_connectivity();
   		}

        return link;
    }




	TNode* TNetwork::AddFullyConnectedNode(int x, int y)
	{
       	TNode* node = AddNode(x, y);
		if (node != NULL)
        {
	   		for (int i=0; i<Nodes->Count-1; i++)
    		{
				AddLink((TNode*)Nodes->Items[i], node);
		    }

    		for (int i=0; i<Nodes->Count; i++)
    		{
                ((TNode*)Nodes->Items[i])->update_connectivity();
    		}

	        return node;
        }
        return NULL;
    }



	void TNetwork::DeleteNodeAndItsLinks(int _index)
	{
        TNode* node = (TNode*)Nodes->Items[_index];
        if (node != NULL)
        {
           	for (int i=0; i<node->Links->Count; i++)
            {
                TLink* link = (TLink*)node->Links->Items[i];

                // delete link from network
                Links->Remove(link);

                // delete link (call the destructor)
		        delete link;

                // after deletetion, shifted up, so current i has to be retried.
                i--;
            }

            // delete node from network
            Nodes->Remove(node);
			DeSelectChildren();
    	    delete node;

    		for (int i=0; i<Nodes->Count; i++)
	    	{
                ((TNode*)Nodes->Items[i])->update_connectivity();
    		}
        }
    }




	void TNetwork::DeleteLink(int _index)
	{
        TLink* link = (TLink*)Links->Items[_index];

        if (link != NULL)
        {
	        // delete link from network
	        Links->Remove(link);
			if (_index == selected_link_index)
    	    {
        		selected_link_index = -1;
	        }

	        delete link;

    		for (int i=0; i<Nodes->Count; i++)
	    	{
                ((TNode*)Nodes->Items[i])->update_connectivity();
    		}
        }
    }




	bool TNetwork::IsEmpty()
	{
	    return ( (Nodes->Count == 0) && (Links->Count == 0) );
	}



	void TNetwork::Clear()
	{
    	Nodes->Clear();
    	Links->Clear();
		ResourceUsageHistory->Clear();
		ResponseTimeHistory->Clear();

		DeSelectChildren();
	}



	void TNetwork::Draw(TCanvas* _canvas)
	{
	  _canvas->Pen->Color = clInfoBk;
	  _canvas->Brush->Color = clInfoBk;
	  _canvas->Rectangle(0, 0, 800, 15);

      if (!IsEmpty())
      {
		update_links_utilization();
		for (int i=0; i<Links->Count; i++)
		{
			((TLink*)Links->Items[i])->Draw(_canvas);
		}
		//update_nodes_utilization();
		for (int i=0; i<Nodes->Count; i++)
		{
			((TNode*)Nodes->Items[i])->Draw(_canvas);
		}

        calculate_shortest_paths();



        // draw matrix values
       	for (int i=0; i<Nodes->Count; i++)
        {
       	  for (int j=0; j<Nodes->Count; j++)
          {
             if      (info_type == 0)
             {
                // clear drawing
                // show response/resource in childform
             }
             else if (info_type == 1)
             {
                // draw matrix indexes
               	for (int i=0; i<Nodes->Count; i++)
                {
                  _canvas->Font->Name  = "MS Sans Serif";
                  _canvas->Font->Size  = 8;
                  _canvas->Font->Color = clGray;
                  _canvas->Brush->Color = cl3DLight;
                  _canvas->TextOut(20+35*i, 200, i+1);
                  _canvas->TextOut(2, 220+25*i, i+1);
                }
              	if (delays[i][j] != MAXINT)
                {
            	  _canvas->Font->Name  = "MS Sans Serif";
    		      _canvas->Font->Size  = 8;
            	  _canvas->Font->Color = clBlack;
            	  _canvas->Brush->Color = cl3DLight;
                  _canvas->TextOut(20, 180, "Shortest Path's Delay");

            	  _canvas->Font->Color = clGreen;
            	  _canvas->Brush->Color = cl3DLight;
                  _canvas->TextOut(20+35*j, 220+25*i, IntToStr((int)delays[i][j]));
                }
             }
             else if (info_type == 2)
             {
                // draw matrix indexes
               	for (int i=0; i<Nodes->Count; i++)
                {
                  _canvas->Font->Name  = "MS Sans Serif";
                  _canvas->Font->Size  = 8;
                  _canvas->Font->Color = clGray;
                  _canvas->Brush->Color = cl3DLight;
                  _canvas->TextOut(20+35*i, 200, i+1);
                  _canvas->TextOut(2, 220+25*i, i+1);
                }
                for (int k=0; k<MAX_NODES; k++)
                {
                  if (intermediates[i][j][k] == -1)
                  {
                    break;
                  }
               	  _canvas->Font->Name  = "MS Sans Serif";
		          _canvas->Font->Size  = 8;
            	  _canvas->Font->Color = clBlack;
            	  _canvas->Brush->Color = cl3DLight;
                  _canvas->TextOut(20, 180, "Intermediate Nodes on Shortest Path");

            	  _canvas->Font->Color = clFuchsia;
            	  _canvas->Brush->Color = cl3DLight;
                  _canvas->TextOut(20+35*j+8*k, 220+25*i, IntToStr((int)((TNode*)Nodes->Items[intermediates[i][j][k]])->get_id()) );
                }
             }
             else if (info_type == 3)
             {
                // draw matrix indexes
               	for (int i=0; i<Nodes->Count; i++)
                {
                  _canvas->Font->Name  = "MS Sans Serif";
                  _canvas->Font->Size  = 8;
                  _canvas->Font->Color = clGray;
                  _canvas->Brush->Color = cl3DLight;
                  _canvas->TextOut(20+35*i, 200, i+1);
                  _canvas->TextOut(2, 220+25*i, i+1);
                }
           		for (int l=0; l<shortest_paths[i][j]->Count; l++)
                {
               	  _canvas->Font->Name  = "MS Sans Serif";
		          _canvas->Font->Size  = 8;
            	  _canvas->Font->Color = clBlack;
            	  _canvas->Brush->Color = cl3DLight;
                  _canvas->TextOut(20, 180, "Intermediate Links on Shortest Path");

            	  _canvas->Font->Color = clBlue;
            	  _canvas->Brush->Color = cl3DLight;
                  _canvas->TextOut(20+35*j+8*l, 220+25*i, IntToStr((int)((TLink*)shortest_paths[i][j]->Items[l])->get_id()) );
                }
             }
             else if (info_type == 4)
             {
                // clear drawing
                // empty paintbox to display the network only
             }// end info_type
          }
        }

	  }
    }





	void TNetwork::DeSelectChildren()
	{
        for(int i=0; i<Nodes->Count; i++)
        {
	        ((TNode*)Nodes->Items[i])->DeSelect();
    	}
        selected_node_index 	= -1;

        for(int i=0; i<Links->Count; i++)
        {
            ((TLink*)Links->Items[i])->DeSelect();
        }
        selected_link_index 	= -1;
	}



	bool TNetwork::AnySelection()
	{
       	return ( (selected_node_index != -1) || (selected_link_index != -1) );
	}


