//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <math.h>
#pragma hdrstop


#include "TNode.h"
#include "TNetwork.h"
//---------------------------------------------------------------------------


	int TNode::scale 	 	= 4;
	int TNode::serial_num	= 0;

	const FragmentsPaintBoxWidth  = 557;
	const FragmentsPaintBoxHeight = 443;
	// all nodes have the same drawing area dimensions
	int TNode::area_width  = 0.333 * (FragmentsPaintBoxWidth);
	int TNode::area_height = 0.333 * (FragmentsPaintBoxHeight)  - 5;

	void TNode::calculate_area_lefttop()
	{
		int x[4];
		x[0] = 0.000 * (FragmentsPaintBoxWidth);
		x[1] = 0.333 * (FragmentsPaintBoxWidth);
		x[2] = 0.666 * (FragmentsPaintBoxWidth);
		x[3] = 0.999 * (FragmentsPaintBoxWidth);
		int y[4];
		y[0] = 0.000 * (FragmentsPaintBoxHeight-1) + 15;
		y[1] = 0.333 * (FragmentsPaintBoxHeight-1) + 10;
		y[2] = 0.666 * (FragmentsPaintBoxHeight-1) + 05;
		y[3] = 0.999 * (FragmentsPaintBoxHeight-1) + 00;
    	area_left	= x[Network->Nodes->IndexOf(this) % 3];
	    area_top	= y[Network->Nodes->IndexOf(this) / 3];
	}



    TNode::TNode(TNode* _node)
	{
      if (_node != NULL)
      {
		fragment_serial_num = _node->fragment_serial_num;

		Network		= _node->Network;
		id   	    = _node->get_id();
        num_of_processors = _node->get_num_of_processors();
		position	= _node->position;
        area_left	= _node->area_left;
        area_top	= _node->area_top;

		processor_utilization = 0;
		storage_utilization   = 0;

		Links       = new TList();
        // don't create twice in Network and here,
        // Network will fill Node->Links with pointers to Network->Links
       	//for (int i=0; i<_node->Links->Count; i++)
        //{
	    //    TLink* copyLink = new TLink((TLink*)_node->Links->Items[i]);
		//	Links->Add(copyLink);
        //}

		set_processor_capacity(_node->processor_capacity);
		set_storage_capacity(_node->storage_capacity);

		Fragments	= new TList();
       	for (int i=0; i<_node->Fragments->Count; i++)
        {
	        TFragment* copyFragment = new TFragment((TFragment*)_node->Fragments->Items[i]);
            copyFragment->Node = this;
			Fragments->Add(copyFragment);
        }

		selected = _node->IsSelected();

		update_connectivity();
		update_capacity();
        update_utilization();   // also will re-calculate
                        		//processor_utilization
                        		//storage_utilization
      }
	}




	TNode::TNode(TNetwork* _Network, TPoint _position)
	{
		fragment_serial_num = 0;

		Network		= _Network;
		id   	    = ++serial_num; // 1,2,3,.. no zero to avoid the need for formatting fragments ids as 01,02,..
		position 	= _position;

		set_num_of_processors(1);     // also will update_capacity

		processor_utilization = 0;
		storage_utilization   = 0;

		Links       = new TList();
        // don't create twice in Network and here,
        // Network will fill Node->Links with pointers to Network->Links
       	//for (int i=0; i<_node->Links->Count; i++)
        //{
	    //    TLink* copyLink = new TLink((TLink*)_node->Links->Items[i]);
		//	Links->Add(copyLink);
        //}

		set_storage_capacity(100);    // also will update_capacity
		set_processor_capacity(100);  // also will update_capacity

		Fragments	= new TList();
	    int NewFragments = 5;//3 + random(8);
       	for (int i=0; i<NewFragments; i++)
        {
			TFragment* Fragment = new TFragment(this);
		    AddFragment(Fragment);
        }
		DeSelectChildren();
		DeSelect();

		selected = false;

		update_connectivity();
		update_capacity();
        update_utilization();   // also will re-calculate
                        		//processor_utilization
                        		//storage_utilization
	}



    TNode::~TNode()
	{
    	// doesn't own Links
    	delete Links;
    	Fragments->Clear();
		delete Fragments;
    }



     void TNode::get_neighbours(int neighbours[])
     {
       	for (int i=0; i<Links->Count; i++)
        {
          TNode* node_a = ((TLink*)Links->Items[i])->Node_a;
          TNode* node_b = ((TLink*)Links->Items[i])->Node_b;

          if ( (node_a != NULL) && (node_b != NULL) )
          {
            int node_a_id = node_a->get_id();
            int node_b_id = node_b->get_id();

            if ( (node_a_id == id) && (node_b_id != id) )
            {
               neighbours[i] = node_b_id;
            }
            else if ( (node_a_id != id) && (node_b_id == id) )
            {
               neighbours[i] = node_a_id;
            }
            else
            {
                // error
                color = clBlue;
            }
          }
        }
     }




	void TNode::update_capacity()
    {
        int _capacity  = storage_capacity   * STORAGE_CAPACITY_FACTOR
                       + processor_capacity * PROCESSOR_CAPACITY_FACTOR;

      	if (_capacity > 10)
        {
        	capacity = _capacity;
        }
        else
        {
        	capacity = 10;
        }

       	radius 	 = capacity / scale;

        update_utilization();
    }




	float TNode::update_utilization()
    {
      if (Fragments != NULL)
      {
		int total_size = 0;      // Fragments Sizes
       	for (int i=0; i<Fragments->Count; i++)
        {
		    total_size += ( ((TFragment*)Fragments->Items[i])->get_size() );
        }
        int used_storage_capacity = total_size / MAX_FULL_SIZE_FRAGMENTS_PER_NODE;
        if (storage_capacity > 0)
        {
            storage_utilization = (used_storage_capacity * 100) / storage_capacity;
        }

		int total_weighted_demand = 0;     // Processing Demands // data shipping = f(size, ...)
       	for (int i=0; i<Fragments->Count; i++)
        {
		    total_weighted_demand += ( ((TFragment*)Fragments->Items[i])->get_demand() * ((TFragment*)Fragments->Items[i])->get_size() ) / 100;
        }
        int used_processor_capacity = total_weighted_demand / MAX_FULLY_DEMANDED_FULL_SIZE_FRAGMNETS_PER_NODE;
        if (processor_capacity > 0)
        {
		    processor_utilization = (used_processor_capacity * 100) / processor_capacity;
        }


        utilization = storage_utilization   * STORAGE_UTILIZATION_FACTOR
                    + processor_utilization * PROCESSOR_UTILIZATION_FACTOR;


        //if (utilization <   0) utilization = 0;
        //if (utilization > 100) utilization = 100;
		if (utilization < 50)	//  0 .. 50
	    {
    	    color	= (TColor) RGB(
									255,
    	                   	        255 - (utilization*2)*2.55,
        	                   	    255
       							  );
	    }
    	else	//  50 .. 100
	    {
    	     color	= (TColor) RGB(
       								((100-utilization)*2)*2.55,
 		    	        	        0,
           						    ((100-utilization)*2)*2.55
       							  );
	    }

		if (utilization > 100)	//  Overflow
        {
	        color	= (TColor) RGB(
           							255,
                           	        0,
                               	    0
           						  );
        }


   		if (!selected)
        {
        	selection_color = color;
        }

        return utilization;
      }

      return 0;
    }







   	int TNode::update_connectivity()
    {
        // WAN linked_node gives better location, but slow connectivity, therefore
        // connectivity = number_of_links regardless of their type / max_number_of_links_per_node
        if ((Network->Nodes->Count - 1) != 0)
        {
            connectivity = (Links->Count * 100)/ (Network->Nodes->Count - 1);
        }
        else
        {
            connectivity = 0;
        }
/*
    	double sum = 0;
       	for (int i=0; i<Links->Count; i++)
        {
            TLink* link = ((TLink*)Links->Items[i]);
            sum += link->get_delay() * link->get_capacity();
        }
    	connectivity = (int)(sum/100.0);  // > N*100%
*/
    	return connectivity;
    }




   	void TNode::AddFragment(TFragment* _Fragment)
    {
	  if(_Fragment != NULL)
      {
        int Columns = 5;
		int xGap    = 35;
	    int yGap    = 35;
		int xShift  = 23;
    	int yShift  = 20;

	    int Column, Row, x, y;
    	Column	= (Fragments->Count)%Columns;
		Row		= (Fragments->Count)/Columns;
		x	= xShift + xGap*Column;
    	y	= yShift + yGap*Row;

   	    ////////////////////////////////////////////////////////////////////////////////////////////////////////////
		_Fragment->Node = this;	// must do this first to calculate_area_lefttop();
   	    ////////////////////////////////////////////////////////////////////////////////////////////////////////////
        // as if it were coming from mouse input, must modify first
		calculate_area_lefttop();
	   	TPoint	modified_position;
    	modified_position.x = x + area_left;
	    modified_position.y = y + area_top;
		_Fragment->set_position(modified_position);		// will set to original position
   	    ////////////////////////////////////////////////////////////////////////////////////////////////////////////

        Fragments->Add(_Fragment);
		DeSelectChildren();

		update_utilization();
      }
    }




	TFragment* TNode::DeleteFragment(int _index, int _Owner)
    {
		// shift subsequent fragments back to fill any gap in drwaing after deletion
       	for (int i=Fragments->Count-1; i>_index; i--)
        {
    	    ////////////////////////////////////////////////////////////////////////////////////////////////////////////
                             	// will set to original position					//will modify original position
		    ((TFragment*)Fragments->Items[i])->set_position(((TFragment*)Fragments->Items[i-1])->get_position());
    	    ////////////////////////////////////////////////////////////////////////////////////////////////////////////
        }

        TFragment* fragment = (TFragment*)Fragments->Items[_index];
        Fragments->Delete(_index);

		if (_index == selected_fragment_index)
        {
        	selected_fragment_index = -1;
        }

		update_utilization();

        if (_Owner)
        {
          delete fragment;
          return NULL;
        }
        return fragment;
    }




	TFragment* TNode::RemoveFragment(TFragment* _Fragment, int _Owner)
    {
	  int index = Fragments->IndexOf(_Fragment);
      if ( (index >= 0) && (index < Fragments->Count) )
      {
        return DeleteFragment(index, _Owner);
      }
      return NULL;
    }





	void TNode::Draw(TCanvas* _canvas)
	{
		_canvas->Pen->Color     = selection_color;
		_canvas->Brush->Color	= color;
		_canvas->Ellipse(position.x-radius, position.y-radius, position.x+radius, position.y+radius);

		if (selected)
        {
			_canvas->Font->Color = selection_color;
        }
        else
        {
    	   	if (utilization > 19)
	        {
				_canvas->Font->Color = clWhite;
    	    }
			else
    	    {
				_canvas->Font->Color = clFuchsia;
    	    }
        }

		_canvas->Font->Name = "NewCareer";
		_canvas->Font->Size  = 14;
   		int left = position.x - _canvas->Font->Size * 0.3;
   		int top  = position.y - _canvas->Font->Size * 0.7;
   		_canvas->TextOut(left, top, IntToStr((int)id));

		_canvas->Font->Name = "NewCareer";
		_canvas->Font->Size  = 6;
   		 left = position.x - 2 * floor(log10(num_of_processors)+1) - 0;
   		 top  = position.y + 9;
   		_canvas->TextOut(left, top, IntToStr((int)num_of_processors));


/*
		_canvas->Pen->Color     = clAqua;
		//_canvas->Brush->Color	= clInfoBk;
        int x,y;
		for (int i=0; i<num_of_processors; i++)
        {
            x = position.x + (i%5)*4 - 10;
            y = position.y + (i/5)*4 - 10;
            _canvas->Ellipse(x, y, x+4, y+4);
        }
		//_canvas->Brush->Color	= color;
*/

        _canvas->Refresh();
    }




	void TNode::DrawCaption(TCanvas* _canvas)
	{
		_canvas->Pen->Color = clInfoBk;
		_canvas->Brush->Color = clInfoBk;
		_canvas->Rectangle(0, 0, 800, 15);

		_canvas->Font->Name  = "MS Sans Serif";
		_canvas->Font->Size  = 8;
		_canvas->Font->Color = clFuchsia;
		_canvas->Brush->Color = clInfoBk;
   		_canvas->TextOut(12, 0, IntToStr((int)id));
   		_canvas->TextOut(43, 0, IntToStr((int)capacity));
   		_canvas->TextOut(100, 0, IntToStr((int)utilization));
   		//_canvas->TextOut(142, 0, IntToStr((int)Links->Count));
        _canvas->TextOut(165, 0, IntToStr((int)Fragments->Count));
   		//_canvas->TextOut(165, 0, IntToStr((int)connectivity));

        _canvas->Refresh();
	}


	void TNode::DrawDetails(TCanvas* _canvas)
	{
        /*
        if (IsSelected())
        {
			_canvas->Pen->Color = selection_color;
        }
        else
        {
			_canvas->Pen->Color = (TColor) RGB(255, 192, 255);
        }
        */
    	calculate_area_lefttop();

	    // light pink background
		_canvas->Pen->Color = (TColor) RGB(255, 192, 255);
		_canvas->Brush->Color = (TColor) RGB(255, 240, 255);
		_canvas->Rectangle(area_left+1, area_top+1, area_left+area_width-1, area_top+area_height-1);

		// node id at corner
		//_canvas->Font->Name  = "MS Sans Serif";
		_canvas->Font->Name  = "New Times Romans";
		_canvas->Font->Size  = 20;
		_canvas->Font->Color = clFuchsia;
		_canvas->Brush->Color = (TColor) RGB(255, 240, 255);
   		_canvas->TextOut(area_left+167, area_top+108, IntToStr((int)id));

       	for (int i=0; i<Fragments->Count; i++)
        {
			((TFragment*)(Fragments->Items[i]))->Draw(_canvas);
        }
    }




	bool TNode::IsSelected()
	{
	   	return selected;
    }


    void TNode::Select(TColor _selection_color)
	{
        Network->set_selected_node_index(Network->Nodes->IndexOf(this));
		selection_color = _selection_color;
   		selected = true;
    }


    void TNode::DeSelect()
	{
        if (selected)
        {
   			selected = false;
	    	selection_color = color;
        	Network->set_selected_node_index(-1);
			DeSelectChildren();
        }
    }


    void TNode::DeSelectChildren()
	{
       	for (int i=0; i<Fragments->Count; i++)
        {
			((TFragment*)(Fragments->Items[i]))->DeSelect();
        }
    }




	bool TNode::AnySelection()
	{
       	return (selected_fragment_index != -1);
	}


