//---------------------------------------------------------------------------
#include "TReplicationFC.h"
//---------------------------------------------------------------------------

#define REPLICATION_RECOMMECNDATION_FACTOR   	0.85        // 85 main criterion
//#define REPLICATION_LINKS_UTILIZATION_FACTOR    0.15		// 15 important in replication to dirvert traffic away
#define REPLICATION_LINKS_DELAY_FACTOR          0.15		// 15 only replicate outside LANs


// static members, set out side constuctor
double TReplicationFC::CandidacyThreshold	= 33.0;		// 33.0 minmum level to justfy consideration    // 05 statistically significant
// for both local machine and replication candidate machine, if local is low demand then migration will take care of it.
double TReplicationFC::ReplicationThreshold	= 50.0;		// 50.0 minmum level to justfy replication


////////////////////////////////////////////////////////////////////////////////
TReplicationFC::TReplicationFC(TFragment* _Fragment)
{
	Fragment = _Fragment;

	C1 = NULL;
	C2 = NULL;
	C3 = NULL;

	C1 = new TFuzzyController();    // attraction
	C2 = new TFuzzyController();    // dumpening
	C3 = new TFuzzyController();    // force

    // needed for Shape Function setup
    C1Input1Centre = 50;  C1Input2Centre = 50;  C1OutputCentre = 50;   // demand = linear increase
    C2Input1Centre = 25;  C2Input2Centre = 40;  C2OutputCentre = 50;   // update_level = fast to increase, availability = linearfast to icrease above medium
    C3OutputCentre = 55;	// scale-up output before link_delay drags it down
    A4 = 0.15;				// output = input * (link_delay ^ A4) = very fast rise

    // static members, set out side constuctor
    //CandidacyThreshold	= 33.0;		// 33.0 minmum level to justfy consideration    // 05 statistically significant
    //ReplicationThreshold	= 50.0;		// 50.0 minmum level to justfy migratation

    // needed for Rules setup
/*
    C1r1out = 0;  C1r1alpha = 0.0;  C1r1weight = 1.0;
    C1r2out = 0;  C1r2alpha = 0.0;  C1r2weight = 1.0;
    C1r3out = 0;  C1r3alpha = 0.0;  C1r3weight = 1.0;   // migrate
    C1r4out = 0;  C1r4alpha = 0.0;  C1r4weight = 1.0;
    C1r5out = 1;  C1r5alpha = 0.0;  C1r5weight = 1.0;
    C1r6out = 2;  C1r6alpha = 0.0;  C1r6weight = 1.0;
    C1r7out = 0;  C1r7alpha = 0.0;  C1r7weight = 1.0;   // migrate
    C1r8out = 2;  C1r8alpha = 0.0;  C1r8weight = 1.0;
    C1r9out = 2;  C1r9alpha = 0.0;  C1r9weight = 1.0;
*/
/*
// can be run without migration, therefore:
    C1r1out = 0;  C1r1alpha = 0.0;  C1r1weight = 1.0;
    C1r2out = 0;  C1r2alpha = 0.0;  C1r2weight = 1.0;
    C1r3out = 1;  C1r3alpha = 0.0;  C1r3weight = 1.0;   // without migrate
    C1r4out = 0;  C1r4alpha = 0.0;  C1r4weight = 1.0;
    C1r5out = 1;  C1r5alpha = 0.0;  C1r5weight = 1.0;
    C1r6out = 2;  C1r6alpha = 0.0;  C1r6weight = 1.0;
    C1r7out = 2;  C1r7alpha = 0.0;  C1r7weight = 1.0;   // without migrate
    C1r8out = 2;  C1r8alpha = 0.0;  C1r8weight = 1.0;
    C1r9out = 2;  C1r9alpha = 0.0;  C1r9weight = 1.0;
*/
///*
// run without migration, controlled by replication level:
    C1r1out = 0;  C1r1alpha = 0.05;  C1r1weight = 1.0;
    C1r2out = 0;  C1r2alpha = 0.05;  C1r2weight = 1.0;
    C1r3out = 1;  C1r3alpha = 0.05;  C1r3weight = 1.0;   // without migrate
    C1r4out = 1;  C1r4alpha = 0.05;  C1r4weight = 1.0;
    C1r5out = 1;  C1r5alpha = 0.05;  C1r5weight = 1.0;
    C1r6out = 1;  C1r6alpha = 0.05;  C1r6weight = 1.0;
    C1r7out = 1;  C1r7alpha = 0.05;  C1r7weight = 1.0;   // without migrate
    C1r8out = 2;  C1r8alpha = 0.05;  C1r8weight = 1.0;
    C1r9out = 2;  C1r9alpha = 0.05;  C1r9weight = 1.0;
//*/

/*
    // stabilise around Medium availability, if update_level allows it.
    C2r1out = 0;  C2r1alpha = 0.0;  C2r1weight = 1.0;
    C2r2out = 1;  C2r2alpha = 0.0;  C2r2weight = 1.0;
    C2r3out = 2;  C2r3alpha = 0.0;  C2r3weight = 1.0;
    C2r4out = 1;  C2r4alpha = 0.0;  C2r4weight = 1.0;
    C2r5out = 2;  C2r5alpha = 0.0;  C2r5weight = 1.0;
    C2r6out = 2;  C2r6alpha = 0.0;  C2r6weight = 1.0;
    C2r7out = 2;  C2r7alpha = 0.0;  C2r7weight = 1.0;
    C2r8out = 2;  C2r8alpha = 0.0;  C2r8weight = 1.0;
    C2r9out = 2;  C2r9alpha = 0.0;  C2r9weight = 1.0;
*/
// run without migration, controlled by replication level:
    C2r1out = 0;  C2r1alpha = 0.0;  C2r1weight = 1.0;
    C2r2out = 1;  C2r2alpha = 0.0;  C2r2weight = 1.0;
    C2r3out = 1;  C2r3alpha = 0.0;  C2r3weight = 1.0;
    C2r4out = 1;  C2r4alpha = 0.0;  C2r4weight = 1.0;
    C2r5out = 1;  C2r5alpha = 0.0;  C2r5weight = 1.0;
    C2r6out = 2;  C2r6alpha = 0.0;  C2r6weight = 1.0;
    C2r7out = 1;  C2r7alpha = 0.0;  C2r7weight = 1.0;
    C2r8out = 2;  C2r8alpha = 0.0;  C2r8weight = 1.0;
    C2r9out = 2;  C2r9alpha = 0.0;  C2r9weight = 1.0;


    C3r1out = 1;  C3r1alpha = 0.0;  C3r1weight = 1.0;
    C3r2out = 1;  C3r2alpha = 0.0;  C3r2weight = 1.0;
    C3r3out = 0;  C3r3alpha = 0.0;  C3r3weight = 1.0;
    C3r4out = 2;  C3r4alpha = 0.0;  C3r4weight = 1.0;
    C3r5out = 1;  C3r5alpha = 0.0;  C3r5weight = 1.0;
    C3r6out = 0;  C3r6alpha = 0.0;  C3r6weight = 1.0;
    C3r7out = 3;  C3r7alpha = 0.0;  C3r7weight = 1.0;
    C3r8out = 2;  C3r8alpha = 0.0;  C3r8weight = 1.0;
    C3r9out = 1;  C3r9alpha = 0.0;  C3r9weight = 1.0;


	num_of_demanding_nodes = 0;
	num_of_candidate_nodes = 0;



	// inputs
    for (int i=0; i<MAX_DEMANDING_NODES; i++)
    {
        remote_demands[i] = 0.0;
    }
    local_demand = 0.0;

    update_level = 0.0;
    availability = 0.0;

    for (int i=0; i<MAX_CANDIDATE_NODES; i++)
    {
        link_delays[i] = 0.0;
    }


	// outputs
    for (int i=0; i<MAX_DEMANDING_NODES; i++)
    {
        demanding_nodes_ids[i] = INVALID_ID;
    }
    for (int i=0; i<MAX_CANDIDATE_NODES; i++)
    {
        candidate_nodes_ids[i] = INVALID_ID;
    }
    for (int i=0; i<MAX_DEMANDING_NODES; i++)
    {
        attractionsets[i] = NULL;
    }
    for (int i=0; i<MAX_DEMANDING_NODES; i++)
    {
        attractions[i] = 0.0;
    }

    dampeningset = NULL;
    dampening    = 0.0;

    for (int i=0; i<MAX_CANDIDATE_NODES; i++)
    {
        forcesets[i] = NULL;
    }
    for (int i=0; i<MAX_CANDIDATE_NODES; i++)
    {
        forces[i] = 0.0;
    }
    for (int i=0; i<MAX_CANDIDATE_NODES; i++)
    {
        replications[i] = 0.0;
    }

///////////////////////
	SetupControllers();
///////////////////////

}
////////////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////////////
TReplicationFC::~TReplicationFC()
{
	Fragment = NULL;
    if (C1 != NULL)	delete	C1;    C1 = NULL;
    if (C2 != NULL)	delete	C2;    C2 = NULL;
    if (C3 != NULL)	delete	C3;    C3 = NULL;
}
////////////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////////////
void TReplicationFC::SetupControllers()
{
  // rules predicate (if) side
  int Low		= 0;
  int Medium	= 1;
  int High		= 2;


  // setup "remote node's attraction force on fragmnet" evaluator

  if (C1 != NULL)
  {	// add input and output variables
  	C1->clear_input_variables();         //  name,         unit, begin,  end
	C1->add_input_variable(new TVariable("RemoteDemand",     "",  00.0, 100.0));
	C1->add_input_variable(new TVariable("LocalDemand",      "",  00.0, 100.0));
	C1->output_variable =  new TVariable("Attraction",       "",  00.0, 100.0);

	// add fuzzysets to the input and output variables

    centre = C1Input1Centre; // %
    begin = C1->input_variables[0]->begin;
    end   = C1->input_variables[0]->end;
    ratio = ( centre / ((end-begin) - centre) );
    // areas
    // end-begin = a        + b;
    //           = ratio*b  + b;
    //           = b * (ratio + 1);
    //    b      = (end-begin)/(ratio + 1);

    // width      percent
        b     = (end-begin)/(ratio + 1);
        a     = ratio*b;
    p0 = begin;
    p1 = p0 + a;
    p2 = p1 + b;
  	C1->input_variables[0]->clear_fuzzysets();
    C1->input_variables[0]->add_fuzzyset(new TFuzzySet(C1->input_variables[0], "Low", 	    S_CURVE  , DOWN , p0, -0.0       , (p0+p1)/2.0, -0.0        , p1));
    C1->input_variables[0]->add_fuzzyset(new TFuzzySet(C1->input_variables[0], "Medium", 	PI_CURVE , UP   , p0, (p0+p1)/2.0, p1         , (p1+p2)/2.0 , p2));
    C1->input_variables[0]->add_fuzzyset(new TFuzzySet(C1->input_variables[0], "High",   	S_CURVE  , UP   , p1, -0.0       , (p1+p2)/2.0, -0.0        , p2));


    centre = C1Input2Centre; // %
    begin = C1->input_variables[1]->begin;
    end   = C1->input_variables[1]->end;
    ratio = ( centre / ((end-begin) - centre) );
    // areas
    // end-begin = a        + b;
    //           = ratio*b  + b;
    //           = b * (ratio + 1);
    //    b      = (end-begin)/(ratio + 1);

    // width      percent
        b     = (end-begin)/(ratio + 1);
        a     = ratio*b;
    p0 = begin;
    p1 = p0 + a;
    p2 = p1 + b;
  	C1->input_variables[1]->clear_fuzzysets();
    C1->input_variables[1]->add_fuzzyset(new TFuzzySet(C1->input_variables[1], "Low", 	    S_CURVE  , DOWN , p0, -0.0        , (p0+p1)/2.0, -0.0        , p1));
    C1->input_variables[1]->add_fuzzyset(new TFuzzySet(C1->input_variables[1], "Medium", 	PI_CURVE , UP   , p0, (p0+p1)/2.0 , p1          , (p1+p2)/2.0, p2));
    C1->input_variables[1]->add_fuzzyset(new TFuzzySet(C1->input_variables[1], "High",   	S_CURVE  , UP   , p1, -0.0        , (p1+p2)/2.0, -0.0        , p2));


    centre = C1OutputCentre; // %
    begin = C1->output_variable->begin;
    end   = C1->output_variable->end;
    ratio = ( centre / ((end-begin) - centre) );
    // areas
    // end-begin = a        + b;
    //           = ratio*b  + b;
    //           = b * (ratio + 1);
    //    b      = (end-begin)/(ratio + 1);

    // width      percent
        b     = (end-begin)/(ratio + 1);
        a     = ratio*b;
    p0 = begin;
    p1 = p0 + a;
    p2 = p1 + b;
  	C1->output_variable->clear_fuzzysets();
    C1->output_variable->add_fuzzyset(new TFuzzySet(C1->output_variable, "Low", 	S_CURVE  , DOWN , p0, -0.0        , (p0+p1)/2.0, -0.0        , p1));
    C1->output_variable->add_fuzzyset(new TFuzzySet(C1->output_variable, "Medium", 	PI_CURVE , UP   , p0, (p0+p1)/2.0 , p1          , (p1+p2)/2.0, p2));
    C1->output_variable->add_fuzzyset(new TFuzzySet(C1->output_variable, "High",   	S_CURVE  , UP   , p1, -0.0        , (p1+p2)/2.0, -0.0        , p2));




	// create empty rules    (conditional, alpha,  weight)

    C1->clear_rules();
    C1->add_rule(new TFuzzyRule(true, C1r1alpha, C1r1weight));
    C1->add_rule(new TFuzzyRule(true, C1r2alpha, C1r2weight));
    C1->add_rule(new TFuzzyRule(true, C1r3alpha, C1r3weight));
    C1->add_rule(new TFuzzyRule(true, C1r4alpha, C1r4weight));
    C1->add_rule(new TFuzzyRule(true, C1r5alpha, C1r5weight));
    C1->add_rule(new TFuzzyRule(true, C1r6alpha, C1r6weight));
    C1->add_rule(new TFuzzyRule(true, C1r7alpha, C1r7weight));
    C1->add_rule(new TFuzzyRule(true, C1r8alpha, C1r8weight));
    C1->add_rule(new TFuzzyRule(true, C1r9alpha, C1r9weight));

	// setup rules              remote_demand        local_demand             attraction
    //----------------------------------------------------------------------------------
    //
    if (C1->rules[0] != NULL)
    {
    C1->rules[0]->predicate->clear_premises();
    C1->rules[0]->predicate->clear_operators();
    C1->rules[0]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[Low]));
    C1->rules[0]->predicate->add_operator(AND);
    C1->rules[0]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[Low]));
    C1->rules[0]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r1out]));
    }
    if (C1->rules[1] != NULL)
    {
    C1->rules[1]->predicate->clear_premises();
    C1->rules[1]->predicate->clear_operators();
    C1->rules[1]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[Low]));
    C1->rules[1]->predicate->add_operator(AND);
    C1->rules[1]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[Medium]));
    C1->rules[1]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r2out]));
    }
    if (C1->rules[2] != NULL)
    {
    C1->rules[2]->predicate->clear_premises();
    C1->rules[2]->predicate->clear_operators();
    C1->rules[2]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[Low]));
    C1->rules[2]->predicate->add_operator(AND);
    C1->rules[2]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[High]));
    C1->rules[2]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r3out]));
    }
    if (C1->rules[3] != NULL)
    {
    C1->rules[3]->predicate->clear_premises();
    C1->rules[3]->predicate->clear_operators();
    C1->rules[3]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[Medium]));
    C1->rules[3]->predicate->add_operator(AND);
    C1->rules[3]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[Low]));
    C1->rules[3]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r4out]));
    }
    if (C1->rules[4] != NULL)
    {
    C1->rules[4]->predicate->clear_premises();
    C1->rules[4]->predicate->clear_operators();
    C1->rules[4]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[Medium]));
    C1->rules[4]->predicate->add_operator(AND);
    C1->rules[4]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[Medium]));
    C1->rules[4]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r5out]));
    }
    if (C1->rules[5] != NULL)
    {
    C1->rules[5]->predicate->clear_premises();
    C1->rules[5]->predicate->clear_operators();
    C1->rules[5]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[Medium]));
    C1->rules[5]->predicate->add_operator(AND);
    C1->rules[5]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[High]));
    C1->rules[5]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r6out]));
    }
    if (C1->rules[6] != NULL)
    {
    C1->rules[6]->predicate->clear_premises();
    C1->rules[6]->predicate->clear_operators();
    C1->rules[6]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[High]));
    C1->rules[6]->predicate->add_operator(AND);
    C1->rules[6]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[Low]));
    C1->rules[6]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r7out]));
    }
    if (C1->rules[7] != NULL)
    {
    C1->rules[7]->predicate->clear_premises();
    C1->rules[7]->predicate->clear_operators();
    C1->rules[7]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[High]));
    C1->rules[7]->predicate->add_operator(AND);
    C1->rules[7]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[Medium]));
    C1->rules[7]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r8out]));
    }
    if (C1->rules[8] != NULL)
    {
    C1->rules[8]->predicate->clear_premises();
    C1->rules[8]->predicate->clear_operators();
    C1->rules[8]->predicate->add_premise(new TPremise(C1->input_variables[0]->fuzzysets[High]));
    C1->rules[8]->predicate->add_operator(AND);
    C1->rules[8]->predicate->add_premise(new TPremise(C1->input_variables[1]->fuzzysets[High]));
    C1->rules[8]->consequent->set_premise(new TPremise(C1->output_variable->fuzzysets[C1r9out]));
    }
  }



  // setup "dampening force on fragmnet" evaluator

  if (C2 != NULL)
  {
	// add input and output variables
  	C2->clear_input_variables();         //  name,       unit, begin,  end
	C2->add_input_variable(new TVariable("UpdateLevel",          "",  00.0, 100.0));
	C2->add_input_variable(new TVariable("Availability",     "",  00.0, 100.0));
	C2->output_variable =  new TVariable("Dampening",            "",  00.0, 100.0);


	// add fuzzysets to the input and output variables

    centre = C2Input1Centre; // %
    begin = C2->input_variables[0]->begin;
    end   = C2->input_variables[0]->end;
    ratio = ( centre / ((end-begin) - centre) );
    // areas
    // end-begin = a        + b;
    //           = ratio*b  + b;
    //           = b * (ratio + 1);
    //    b      = (end-begin)/(ratio + 1);

    // width      percent
        b     = (end-begin)/(ratio + 1);
        a     = ratio*b;
    p0 = begin;
    p1 = p0 + a;
    p2 = p1 + b;
  	C2->input_variables[0]->clear_fuzzysets();
    C2->input_variables[0]->add_fuzzyset(new TFuzzySet(C2->input_variables[0], "Low", 	    S_CURVE  , DOWN , p0, -0.0       , (p0+p1)/2.0, -0.0        , p1));
    C2->input_variables[0]->add_fuzzyset(new TFuzzySet(C2->input_variables[0], "Medium", 	PI_CURVE , UP   , p0, (p0+p1)/2.0, p1         , (p1+p2)/2.0 , p2));
    C2->input_variables[0]->add_fuzzyset(new TFuzzySet(C2->input_variables[0], "High",   	S_CURVE  , UP   , p1, -0.0       , (p1+p2)/2.0, -0.0        , p2));


    centre = C2Input2Centre; // %
    begin = C2->input_variables[1]->begin;
    end   = C2->input_variables[1]->end;
    ratio = ( centre / ((end-begin) - centre) );
    // areas
    // end-begin = a        + b;
    //           = ratio*b  + b;
    //           = b * (ratio + 1);
    //    b      = (end-begin)/(ratio + 1);

    // width      percent
        b     = (end-begin)/(ratio + 1);
        a     = ratio*b;
    p0 = begin;
    p1 = p0 + a;
    p2 = p1 + b;
  	C2->input_variables[1]->clear_fuzzysets();
    C2->input_variables[1]->add_fuzzyset(new TFuzzySet(C2->input_variables[1], "Low", 	    S_CURVE  , DOWN , p0, -0.0        , (p0+p1)/2.0, -0.0        , p1));
    C2->input_variables[1]->add_fuzzyset(new TFuzzySet(C2->input_variables[1], "Medium", 	PI_CURVE , UP   , p0, (p0+p1)/2.0 , p1          , (p1+p2)/2.0, p2));
    C2->input_variables[1]->add_fuzzyset(new TFuzzySet(C2->input_variables[1], "High",   	S_CURVE  , UP   , p1, -0.0        , (p1+p2)/2.0, -0.0        , p2));


    centre = C2OutputCentre; // %
    begin = C2->output_variable->begin;
    end   = C2->output_variable->end;
    ratio = ( centre / ((end-begin) - centre) );
    // areas
    // end-begin = a        + b;
    //           = ratio*b  + b;
    //           = b * (ratio + 1);
    //    b      = (end-begin)/(ratio + 1);

    // width      percent
        b     = (end-begin)/(ratio + 1);
        a     = ratio*b;
    p0 = begin;
    p1 = p0 + a;
    p2 = p1 + b;
  	C2->output_variable->clear_fuzzysets();
    C2->output_variable->add_fuzzyset(new TFuzzySet(C2->output_variable, "Low", 	S_CURVE  , DOWN , p0, -0.0        , (p0+p1)/2.0, -0.0        , p1));
    C2->output_variable->add_fuzzyset(new TFuzzySet(C2->output_variable, "Medium", 	PI_CURVE , UP   , p0, (p0+p1)/2.0 , p1          , (p1+p2)/2.0, p2));
    C2->output_variable->add_fuzzyset(new TFuzzySet(C2->output_variable, "High",   	S_CURVE  , UP   , p1, -0.0        , (p1+p2)/2.0, -0.0        , p2));



	// create empty rules    (conditional, alpha,  weight)
    C2->clear_rules();
    C2->add_rule(new TFuzzyRule(true, C2r1alpha, C2r1weight));
    C2->add_rule(new TFuzzyRule(true, C2r2alpha, C2r2weight));
    C2->add_rule(new TFuzzyRule(true, C2r3alpha, C2r3weight));
    C2->add_rule(new TFuzzyRule(true, C2r4alpha, C2r4weight));
    C2->add_rule(new TFuzzyRule(true, C2r5alpha, C2r5weight));
    C2->add_rule(new TFuzzyRule(true, C2r6alpha, C2r6weight));
    C2->add_rule(new TFuzzyRule(true, C2r7alpha, C2r7weight));
    C2->add_rule(new TFuzzyRule(true, C2r8alpha, C2r8weight));
    C2->add_rule(new TFuzzyRule(true, C2r9alpha, C2r9weight));


	// setup rules                update_leve       availability          dampening
    //----------------------------------------------------------------------------------
    //
    if (C2->rules[0] != NULL)
    {
    C2->rules[0]->predicate->clear_premises();
    C2->rules[0]->predicate->clear_operators();
    C2->rules[0]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[Low]));
    C2->rules[0]->predicate->add_operator(AND);
    C2->rules[0]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[Low]));
    C2->rules[0]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r1out]));
    }
    if (C2->rules[1] != NULL)
    {
    C2->rules[1]->predicate->clear_premises();
    C2->rules[1]->predicate->clear_operators();
    C2->rules[1]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[Low]));
    C2->rules[1]->predicate->add_operator(AND);
    C2->rules[1]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[Medium]));
    C2->rules[1]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r2out]));
    }
    if (C2->rules[2] != NULL)
    {
    C2->rules[2]->predicate->clear_premises();
    C2->rules[2]->predicate->clear_operators();
    C2->rules[2]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[Low]));
    C2->rules[2]->predicate->add_operator(AND);
    C2->rules[2]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[High]));
    C2->rules[2]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r3out]));
    }
    if (C2->rules[3] != NULL)
    {
    C2->rules[3]->predicate->clear_premises();
    C2->rules[3]->predicate->clear_operators();
    C2->rules[3]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[Medium]));
    C2->rules[3]->predicate->add_operator(AND);
    C2->rules[3]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[Low]));
    C2->rules[3]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r4out]));
    }
    if (C2->rules[4] != NULL)
    {
    C2->rules[4]->predicate->clear_premises();
    C2->rules[4]->predicate->clear_operators();
    C2->rules[4]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[Medium]));
    C2->rules[4]->predicate->add_operator(AND);
    C2->rules[4]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[Medium]));
    C2->rules[4]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r5out]));
    }
    if (C2->rules[5] != NULL)
    {
    C2->rules[5]->predicate->clear_premises();
    C2->rules[5]->predicate->clear_operators();
    C2->rules[5]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[Medium]));
    C2->rules[5]->predicate->add_operator(AND);
    C2->rules[5]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[High]));
    C2->rules[5]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r6out]));
    }
    if (C2->rules[6] != NULL)
    {
    C2->rules[6]->predicate->clear_premises();
    C2->rules[6]->predicate->clear_operators();
    C2->rules[6]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[High]));
    C2->rules[6]->predicate->add_operator(AND);
    C2->rules[6]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[Low]));
    C2->rules[6]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r7out]));
    }
    if (C2->rules[7] != NULL)
    {
    C2->rules[7]->predicate->clear_premises();
    C2->rules[7]->predicate->clear_operators();
    C2->rules[7]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[High]));
    C2->rules[7]->predicate->add_operator(AND);
    C2->rules[7]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[Medium]));
    C2->rules[7]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r8out]));
    }
    if (C2->rules[8] != NULL)
    {
    C2->rules[8]->predicate->clear_premises();
    C2->rules[8]->predicate->clear_operators();
    C2->rules[8]->predicate->add_premise(new TPremise(C2->input_variables[0]->fuzzysets[High]));
    C2->rules[8]->predicate->add_operator(AND);
    C2->rules[8]->predicate->add_premise(new TPremise(C2->input_variables[1]->fuzzysets[2]));
    C2->rules[8]->consequent->set_premise(new TPremise(C2->output_variable->fuzzysets[C2r9out]));
    }
  }





  // setup "replication force towards remote node" evaluator

  if ( (C1 != NULL) && (C2 != NULL) && (C3 != NULL) )
  {
	// add input and output variables
  	C3->clear_input_variables();         //  name,         unit, begin,  end
	C3->add_input_variable(new TVariable(*(C1->output_variable)));
	C3->add_input_variable(new TVariable(*(C2->output_variable)));
	C3->output_variable =  new TVariable("ReplicationForce",        "", 00.0,   100.0);


	// add fuzzysets to the input and output variables

    centre = C3OutputCentre; // %
    begin = C3->output_variable->begin;
    end   = C3->output_variable->end;
    ratio = ( centre / ((end-begin) - centre) );
    // areas
    // end-begin = a                   + b             + c        + d;
    //           = ratio*ratio*ratio*d + ratio*ratio*d + ratio*d  + d;
    //           = d * (ratio*ratio*ratio + ratio*ratio + ratio  + 1);
    //    d      = (end-begin)/(ratio*ratio*ratio + ratio*ratio + ratio  + 1);

    // width      percent
        d     = (end-begin)/(ratio*ratio*ratio + ratio*ratio + ratio  + 1);
        c     = ratio*d;
        b     = ratio*ratio*d;
        a     = ratio*ratio*ratio*d;

    p0 = begin;
    p1 = p0 + a;
    p2 = p1 + b;
    p3 = p2 + c;
    p4 = p3 + d;
	// add fuzzysets to the output variable
  	C3->output_variable->clear_fuzzysets();
    C3->output_variable->add_fuzzyset(new TFuzzySet(C3->output_variable, "VeryLow", S_CURVE  , DOWN , p0, -0.0       , (p0+p1)/2.0, -0.0       , p1));
    C3->output_variable->add_fuzzyset(new TFuzzySet(C3->output_variable, "Low", 	PI_CURVE , UP   , p0, (p0+p1)/2.0, p1         , (p1+p2)/2.0, p2));
    C3->output_variable->add_fuzzyset(new TFuzzySet(C3->output_variable, "Medium", 	PI_CURVE , UP   , p1, (p1+p2)/2.0, p2         , (p2+p3)/2.0, p3));
    C3->output_variable->add_fuzzyset(new TFuzzySet(C3->output_variable, "High", 	PI_CURVE , UP   , p2, (p2+p3)/2.0, p3         , (p3+p4)/2.0, p4));
    C3->output_variable->add_fuzzyset(new TFuzzySet(C3->output_variable, "VeryHigh",S_CURVE  , UP   , p3, -0.0       , (p3+p4)/2.0, -0.0       , p4));



	// create empty rules    (conditional, alpha,  weight)
    C3->clear_rules();
    C3->add_rule(new TFuzzyRule(true, C3r1alpha, C3r1weight));
    C3->add_rule(new TFuzzyRule(true, C3r2alpha, C3r2weight));
    C3->add_rule(new TFuzzyRule(true, C3r3alpha, C3r3weight));
    C3->add_rule(new TFuzzyRule(true, C3r4alpha, C3r4weight));
    C3->add_rule(new TFuzzyRule(true, C3r5alpha, C3r5weight));
    C3->add_rule(new TFuzzyRule(true, C3r6alpha, C3r6weight));
    C3->add_rule(new TFuzzyRule(true, C3r7alpha, C3r7weight));
    C3->add_rule(new TFuzzyRule(true, C3r8alpha, C3r8weight));
    C3->add_rule(new TFuzzyRule(true, C3r9alpha, C3r9weight));


	// setup rules              attraction        dampening            replication_force
    //----------------------------------------------------------------------------------
    //
    if (C3->rules[0] != NULL)
    {
    C3->rules[0]->predicate->clear_premises();
    C3->rules[0]->predicate->clear_operators();
    C3->rules[0]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[Low]));
    C3->rules[0]->predicate->add_operator(AND);
    C3->rules[0]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[Low]));
    C3->rules[0]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r1out]));
    }
    if (C3->rules[1] != NULL)
    {
    C3->rules[1]->predicate->clear_premises();
    C3->rules[1]->predicate->clear_operators();
    C3->rules[1]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[Low]));
    C3->rules[1]->predicate->add_operator(AND);
    C3->rules[1]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[Medium]));
    C3->rules[1]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r2out]));
    }
    if (C3->rules[2] != NULL)
    {
    C3->rules[2]->predicate->clear_premises();
    C3->rules[2]->predicate->clear_operators();
    C3->rules[2]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[Low]));
    C3->rules[2]->predicate->add_operator(AND);
    C3->rules[2]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[High]));
    C3->rules[2]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r3out]));
    }
    if (C3->rules[3] != NULL)
    {
    C3->rules[3]->predicate->clear_premises();
    C3->rules[3]->predicate->clear_operators();
    C3->rules[3]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[Medium]));
    C3->rules[3]->predicate->add_operator(AND);
    C3->rules[3]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[Low]));
    C3->rules[3]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r4out]));
    }
    if (C3->rules[4] != NULL)
    {
    C3->rules[4]->predicate->clear_premises();
    C3->rules[4]->predicate->clear_operators();
    C3->rules[4]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[Medium]));
    C3->rules[4]->predicate->add_operator(AND);
    C3->rules[4]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[Medium]));
    C3->rules[4]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r5out]));
    }
    if (C3->rules[5] != NULL)
    {
    C3->rules[5]->predicate->clear_premises();
    C3->rules[5]->predicate->clear_operators();
    C3->rules[5]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[Medium]));
    C3->rules[5]->predicate->add_operator(AND);
    C3->rules[5]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[High]));
    C3->rules[5]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r6out]));
    }
    if (C3->rules[6] != NULL)
    {
    C3->rules[6]->predicate->clear_premises();
    C3->rules[6]->predicate->clear_operators();
    C3->rules[6]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[High]));
    C3->rules[6]->predicate->add_operator(AND);
    C3->rules[6]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[Low]));
    C3->rules[6]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r7out]));
    }
    if (C3->rules[7] != NULL)
    {
    C3->rules[7]->predicate->clear_premises();
    C3->rules[7]->predicate->clear_operators();
    C3->rules[7]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[High]));
    C3->rules[7]->predicate->add_operator(AND);
    C3->rules[7]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[Medium]));
    C3->rules[7]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r8out]));
    }
    if (C3->rules[8] != NULL)
    {
    C3->rules[8]->predicate->clear_premises();
    C3->rules[8]->predicate->clear_operators();
    C3->rules[8]->predicate->add_premise(new TPremise(C3->input_variables[0]->fuzzysets[High]));
    C3->rules[8]->predicate->add_operator(AND);
    C3->rules[8]->predicate->add_premise(new TPremise(C3->input_variables[1]->fuzzysets[High]));
    C3->rules[8]->consequent->set_premise(new TPremise(C3->output_variable->fuzzysets[C3r9out]));
    }
  }




  // setup "replication recommendation" evaluator

  if ( (C3 != NULL) )
  {
  // do nothing here
  // A4 must be setup outside and
  // the algorithm is applied in RunA4()
  }

}
////////////////////////////////////////////////////////////////////////////////











// fill demanding_nodes_ids[MAX_DEMANDING_NODES]
void TReplicationFC::IdentifyDemandingNodes() // using Fragment->DemandLevelTable[i]
{
  if (Fragment != NULL)
  {
	num_of_demanding_nodes = Fragment->DemandLevelTable->Count;

    for (int i=0; i<num_of_demanding_nodes; i++)
    {
        demanding_nodes_ids[i] = ((TDemandLevel*)Fragment->DemandLevelTable->Items[i])->NodeId;
    }
  }
}





// for each demanding node in demanding_nodes_ids[MAX_DEMANDING_NODES]
// fill remote_demands[i]
void TReplicationFC::FillRemoteDemands() 		// using Fragment->DemandLevelTable[i]
{
  if (Fragment != NULL)
  {
    for (int i=0; i<num_of_demanding_nodes; i++)
    {
      if (Fragment->Node != NULL)
      {
        if (((TDemandLevel*)Fragment->DemandLevelTable->Items[i])->NodeId != Fragment->Node->get_id())  // remote
        {
        	remote_demands[i] = ((TDemandLevel*)Fragment->DemandLevelTable->Items[i])->Demand;
        }
      }
    }
  }
}


// average info from fragment's statistics in last T (+ forcast)
void TReplicationFC::FillLocalDemand()	// using Fragment->DemandLevelTable[i]
{
  if (Fragment != NULL)
  {
    for (int i=0; i<num_of_demanding_nodes; i++)
    {
      if (Fragment->Node != NULL)
      {
        if (((TDemandLevel*)Fragment->DemandLevelTable->Items[i])->NodeId == Fragment->Node->get_id())  // local
        {
        	local_demand = ((TDemandLevel*)Fragment->DemandLevelTable->Items[i])->Demand;
            break;
        }
      }
    }
  }
}


// for each demanding node, run C1 to get its attraction force and
// fill attractions[MAX_DEMANDING_NODES] with the results
double TReplicationFC::RunC1(int index, double input1, double input2,
						   int AggregationMethod,
                           int CorrelationMethod,
                           int AlphaCutType,
                           int CompositionMethod,
                           int DefuzzificationMethod
					      )
{
  if ( C1 != NULL )
  {
    // apply new inputs
	C1->rules[0]->predicate->premises[0]->set_scalar(input1);
	C1->rules[0]->predicate->premises[1]->set_scalar(input2);
	C1->rules[1]->predicate->premises[0]->set_scalar(input1);
	C1->rules[1]->predicate->premises[1]->set_scalar(input2);
	C1->rules[2]->predicate->premises[0]->set_scalar(input1);
	C1->rules[2]->predicate->premises[1]->set_scalar(input2);
	C1->rules[3]->predicate->premises[0]->set_scalar(input1);
	C1->rules[3]->predicate->premises[1]->set_scalar(input2);
	C1->rules[4]->predicate->premises[0]->set_scalar(input1);
	C1->rules[4]->predicate->premises[1]->set_scalar(input2);
	C1->rules[5]->predicate->premises[0]->set_scalar(input1);
	C1->rules[5]->predicate->premises[1]->set_scalar(input2);
	C1->rules[6]->predicate->premises[0]->set_scalar(input1);
	C1->rules[6]->predicate->premises[1]->set_scalar(input2);
	C1->rules[7]->predicate->premises[0]->set_scalar(input1);
	C1->rules[7]->predicate->premises[1]->set_scalar(input2);
	C1->rules[8]->predicate->premises[0]->set_scalar(input1);
	C1->rules[8]->predicate->premises[1]->set_scalar(input2);

    // run the controller for the new inputs
	attractionsets[index]        = C1->GenerateOutputSet(AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod);
	return attractions[index]    = C1->GenerateOutput   (AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod, DefuzzificationMethod);
  }
  return INVALID_DOMAIN;
}









// fill candidate_nodes_ids[MAX_CANDIDATE_NODES] with ids of
// first nodes with attraction force > CandidacyThreshold, the rest fill with INVALID_ID
void TReplicationFC::SelectCandidates()
{
	// temporary storage areas
    double 		candidate_nodes_remote_demands[MAX_CANDIDATE_NODES];
    double 		candidate_nodes_attractions[MAX_CANDIDATE_NODES];
    TFuzzySet* 	candidate_nodes_attractionsets[MAX_CANDIDATE_NODES];

    int count = 0;
    for (int i=0; i<num_of_demanding_nodes; i++)
    {
//
//PRE-CONDITION:
//
       	if (attractions[i] >= CandidacyThreshold)
        {
	        candidate_nodes_ids[count] 				= demanding_nodes_ids[i];
            candidate_nodes_remote_demands[count] 	= remote_demands[i];
            candidate_nodes_attractions[count]	    = attractions[i];
            candidate_nodes_attractionsets[count]   = new TFuzzySet(*attractionsets[i]);
            count++;
        }
    }
    num_of_candidate_nodes = count;

	// clear-out demandings
    for (int i=0; i<MAX_DEMANDING_NODES; i++)
    {
        remote_demands[i]   = 0.0;
        attractions[i] 	    = 0.0;
        delete attractionsets[i]; attractionsets[i] = NULL;
    }

	// copy-back from candidates
    for (int i=0; i<num_of_candidate_nodes; i++)
    {
        remote_demands[i] 	= candidate_nodes_remote_demands[i];
        attractions[i] 	    = candidate_nodes_attractions[i];
        attractionsets[i]   = candidate_nodes_attractionsets[i];
                              candidate_nodes_attractionsets[i] = NULL;
    }
}









// fill update_level
void TReplicationFC::FillUpdateLevel()          //using Fragment->update_ratio
{
  if (Fragment != NULL)
  {
   	update_level = Fragment->calculate_update_level();
  }
}


// fill availability
void TReplicationFC::FillAvailability()    //using Fragment->calculate_availability()
{
  if (Fragment != NULL)
  {
   	availability = Fragment->calculate_availability();
  }
}



// run C2 to get the dampening force due to upadte-level (synchronization penalty)
// fill dampening with the results
double TReplicationFC::RunC2(double input1, double input2,
						   int AggregationMethod,
                           int CorrelationMethod,
                           int AlphaCutType,
                           int CompositionMethod,
                           int DefuzzificationMethod
					      )
{
  if ( C2 != NULL )
  {
    // apply new inputs
	C2->rules[0]->predicate->premises[0]->set_scalar(input1);
	C2->rules[0]->predicate->premises[1]->set_scalar(input2);
	C2->rules[1]->predicate->premises[0]->set_scalar(input1);
	C2->rules[1]->predicate->premises[1]->set_scalar(input2);
	C2->rules[2]->predicate->premises[0]->set_scalar(input1);
	C2->rules[2]->predicate->premises[1]->set_scalar(input2);
	C2->rules[3]->predicate->premises[0]->set_scalar(input1);
	C2->rules[3]->predicate->premises[1]->set_scalar(input2);
	C2->rules[4]->predicate->premises[0]->set_scalar(input1);
	C2->rules[4]->predicate->premises[1]->set_scalar(input2);
	C2->rules[5]->predicate->premises[0]->set_scalar(input1);
	C2->rules[5]->predicate->premises[1]->set_scalar(input2);
	C2->rules[6]->predicate->premises[0]->set_scalar(input1);
	C2->rules[6]->predicate->premises[1]->set_scalar(input2);
	C2->rules[7]->predicate->premises[0]->set_scalar(input1);
	C2->rules[7]->predicate->premises[1]->set_scalar(input2);
	C2->rules[8]->predicate->premises[0]->set_scalar(input1);
	C2->rules[8]->predicate->premises[1]->set_scalar(input2);

    // run the controller for the new inputs
	dampeningset      = C2->GenerateOutputSet(AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod);
	return dampening  = C2->GenerateOutput   (AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod, DefuzzificationMethod);
  }
  return INVALID_DOMAIN;
}









// for each chosen candidate node, run C3 to get the replication force
// fill forces[MAX_CANDIDATE_NODES] with the results
double TReplicationFC::RunC3(int index, double input1, double input2,
						   int AggregationMethod,
                           int CorrelationMethod,
                           int AlphaCutType,
                           int CompositionMethod,
                           int DefuzzificationMethod
					      )
{
  if ( (C1 != NULL) && (C2 != NULL) && (C3 != NULL) )
  {
    // apply new inputs
	C3->rules[0]->predicate->premises[0]->set_scalar(input1);
	C3->rules[0]->predicate->premises[1]->set_scalar(input2);
	C3->rules[1]->predicate->premises[0]->set_scalar(input1);
	C3->rules[1]->predicate->premises[1]->set_scalar(input2);
	C3->rules[2]->predicate->premises[0]->set_scalar(input1);
	C3->rules[2]->predicate->premises[1]->set_scalar(input2);
	C3->rules[3]->predicate->premises[0]->set_scalar(input1);
	C3->rules[3]->predicate->premises[1]->set_scalar(input2);
	C3->rules[4]->predicate->premises[0]->set_scalar(input1);
	C3->rules[4]->predicate->premises[1]->set_scalar(input2);
	C3->rules[5]->predicate->premises[0]->set_scalar(input1);
	C3->rules[5]->predicate->premises[1]->set_scalar(input2);
	C3->rules[6]->predicate->premises[0]->set_scalar(input1);
	C3->rules[6]->predicate->premises[1]->set_scalar(input2);
	C3->rules[7]->predicate->premises[0]->set_scalar(input1);
	C3->rules[7]->predicate->premises[1]->set_scalar(input2);
	C3->rules[8]->predicate->premises[0]->set_scalar(input1);
	C3->rules[8]->predicate->premises[1]->set_scalar(input2);

    // run the controller for the new inputs
	forcesets[index]        = C3->GenerateOutputSet(AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod);
	return forces[index]    = C3->GenerateOutput   (AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod, DefuzzificationMethod);
  }
  return INVALID_DOMAIN;
}







// fill link_delays[index] with total delay on shortest path
void TReplicationFC::FillLinkDelay(int index)  //using Network->ShortestPathDelay[local_node_index][candidate_node_index]
{
  if (Fragment != NULL)
  {
    int local_node_index     = Fragment->Node->Network->Nodes->IndexOf(Fragment->Node);
    int candidate_node_index = index;

    Fragment->Node->Network->calculate_shortest_paths();
   	link_delays[index] = ((float)(Fragment->Node->Network->delays[local_node_index][candidate_node_index])) / 100.0;
  }
}



// for each chosen candidate node, run A4 to get the replication recommendation
// fill replications[MAX_CANDIDATE_NODES] with the results
double TReplicationFC::RunA4(int index, double input1, double input2)
{
  input2 = 50;
  if ( (C1 != NULL) && (C2 != NULL) && (C3 != NULL) )
  {
    // run the controller for the new inputs
    //                              force                link_delay
	return (replications[index] = (input1/100.0) * pow((input2/100.0), A4)*100.0);
  }
  return INVALID_DOMAIN;
}








// fill replication_node_ids[MAX_CANDIDATE_NODES] with the ids of the nodes with
// (replications[MAX_CANDIDATE_NODES] ?and? ... ) >= ReplicationThreshold
// or fill any replication_node_id with INVALID_ID for no replication
void TReplicationFC::FillBestCandidates(int replication_node_ids[])
{
    double goodness;

    int nodes_count = 0;
    for (int index=0; index<num_of_candidate_nodes; index++)
    {
	    TNode* replication_node = Fragment->Node->Network->GetNode(candidate_nodes_ids[index]);

//
//POST-CONDITION:
//
        // skip if candidate is local node
        if (replication_node->get_id() == Fragment->Node->get_id())
        {
          continue;
        }


//
//POST-CONDITION:
//
        // skip candidate if target is FULL or BUSY or SMALL or SLOW

            int needed_storage_capacity = Fragment->get_size() / MAX_FULL_SIZE_FRAGMENTS_PER_NODE;

		    int weighted_demand             = ( Fragment->get_demand() * Fragment->get_size() ) / 100;
            int needed_processor_capacity   = weighted_demand / MAX_FULLY_DEMANDED_FULL_SIZE_FRAGMNETS_PER_NODE;


            // skip candidate if target is FULL or BUSY
            if ( (replication_node->get_storage_utilization() > 95) || (replication_node->get_processor_utilization() > 95) )
            {
              continue;
            }


            // skip candidate if target is SMALL or SLOW
            if (
                ( replication_node->get_storage_capacity()   < needed_storage_capacity )
              ||
                ( replication_node->get_processor_capacity() < needed_processor_capacity )
               )
            {
              continue;
            }






        if (replication_node != NULL)
        {
            // utilization of first-link on shortest path
            int    remote_node_index = Fragment->Node->Network->Nodes->IndexOf(replication_node);
            int    local_node_index  = Fragment->Node->Network->Nodes->IndexOf(Fragment->Node);
            //double link_load = ((TLink*)(Fragment->Node->Network->shortest_paths[local_node_index][remote_node_index]->Items[0]))->get_utilization();
            double link_delay = Fragment->Node->Network->delays[local_node_index][remote_node_index];

            //goodness    = replication recommendation
            // 	    	  //+ utilization of first-link on shortest path
            // 	    	  + delay of shortest_path to target
            // 	            		  ;
        	goodness      = REPLICATION_RECOMMECNDATION_FACTOR   * replications[index]
           	    		  //+ REPLICATION_LINKS_UTILIZATION_FACTOR * link_delay
           	    		  + REPLICATION_LINKS_DELAY_FACTOR * link_delay
   	            		  ;


            if (goodness >= ReplicationThreshold)
            {
                replication_node_ids[nodes_count] = candidate_nodes_ids[index];
                nodes_count++;
            }
        }
    }
}









// runs all controllers and algorithms and
// fills replication_node_ids[MAX_CANDIDATE_NODES] with the ids of the best candidate nodes
void TReplicationFC::Run(int replication_node_ids[])
{
    IdentifyDemandingNodes();


    FillRemoteDemands();
    FillLocalDemand();
    for (int i=0; i<num_of_demanding_nodes; i++)
    {
	    RunC1(i, remote_demands[i], local_demand,
        //    AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod, DefuzzificationMethod
                   PRODUCT,           TRUNCATE,        STRONG,        BOUNDEDSUM,        NEAR_EDGE        );
    }


	SelectCandidates();


    FillUpdateLevel();
    FillAvailability();
    RunC2(update_level, availability,
    //    AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod, DefuzzificationMethod
               PRODUCT,           TRUNCATE,        STRONG,        BOUNDEDSUM,        NEAR_EDGE        );



    for (int i=0; i<num_of_candidate_nodes; i++)
    {
	    RunC3(i, attractions[i], dampening,
        //    AggregationMethod, CorrelationMethod, AlphaCutType, CompositionMethod, DefuzzificationMethod
                   PRODUCT,           TRUNCATE,        STRONG,        BOUNDEDSUM,        NEAR_EDGE        );
    }


    for (int i=0; i<num_of_candidate_nodes; i++)
    {
    	FillLinkDelay(i);
    }

    for (int i=0; i<num_of_candidate_nodes; i++)
    {
	    RunA4(i, forces[i], link_delays[i]);
    }


	FillBestCandidates(replication_node_ids);
}


